// object definitions for the Dispenser object


#include "ui.H"
#include "dispenser.H"
#include "pose.H"
#include "pbundle.H"
#include "results.H"
#include "sanalyzer.H"
#include "predictors/pholder.H"
#include "predictors/lpoly.H"
#include "predictors/expsmooth.H"

#include <qstatusbar.h>
#include <qprogressdialog.h>
#include <qapplication.h>
#include <qmultilineedit.h>
#include <iostream.h>
#include <sys/time.h>

#define MAX 100
#define SF  4
#define QFACTOR 0.1  // empirical results indicate that orientation noise variance is
                     // smaller by a factor of 10

#define SRATENUM 40
#define PSKIP 6 // could do 2 here
#define MAXLEAD 250
#define MINLEAD 16
#define SAMPLE_RATE 10000 // for noise generation
#define DURATION 24 // 24 seconds of noise
#define ERRSTEP 5 // error step for the error test
#define ERRMAX 100
#define NOISENUM 1000000

// constructor
Dispenser::Dispenser(GlobalWidget *gbw) : _gbw(gbw)
{
 _res = new Results(_gbw);
 reset();
 _bundle = new ParamBundle(); 
 _pholder = new PredictorHolder();  

 // set up noise generators
  struct timeval ts; struct timezone tz;
  gettimeofday(&ts, &tz);
  _posnoise = new GaussWhiteNoise((long)(ts.tv_sec + ts.tv_usec/1e6)); 
  gettimeofday(&ts, &tz);
  _quatnoise = new GaussWhiteNoise((long)(ts.tv_sec + ts.tv_usec/1e6)); 

  // initialize params
  _alphap = _alphaq = 0.0;

}


Dispenser::~Dispenser()
{
   delete _bundle;
   delete _pholder; 
   delete _res;
   delete _posnoise;
   delete _quatnoise;
}


void
Dispenser::reset()
{
  _leadtime = 0;
  _pinterval = 0.0;
  _srate = 0;
  _noisev = 0.0; 
  _ptype = tNone;
}

void
Dispenser::collectTestInfo()
{
  reset();
  _leadtime  =  _gbw->getCPanel()->getLeadTimeInt();
  _pinterval =  _leadtime * (1.0 / _gbw->getCPanel()->getOrigSRate());
  _srate     =  _gbw->getCPanel()->getSRateInt();
  _noisev    =  _gbw->getCPanel()->getNoiseVar();
  _ptype     =  _gbw->getPredType();
  //  delete _res;
  // _res = new Results(_gbw);
  _res->clear(); // get ready for the next run
  _res->reload();
}


void 
Dispenser::fillBundle()
{
  // always going to need this
  _bundle->setLeadTime(_leadtime);
  _bundle->setDegree(_gbw->getLPolyDialog()->getPDegree());
  _bundle->setNumPts(_gbw->getLPolyDialog()->getPWindow());
  _bundle->setLPolyType(_gbw->getLPolyDialog()->getLPType());
  _bundle->setDerFly(_gbw->getLPolyDialog()->getDerFly());
  _bundle->setKFFlag(_gbw->getLPolyDialog()->getKFFlag());
  _bundle->setKFExp(_gbw->getExpSmDialog()->getKFFlag());
  _bundle->setKFGrey(_gbw->getGreyPredDialog()->getKFFlag());
  _bundle->setCurSRate(_gbw->getCPanel()->getCurSRate());
  _bundle->setExpSmoothType(_gbw->getExpSmDialog()->getExpSmoothType());
  _bundle->setAlpha(_gbw->getExpSmDialog()->getAlpha());
  _bundle->setBeta(_gbw->getExpSmDialog()->getBeta());  
  _bundle->setGreyPts(_gbw->getGreyPredDialog()->getNumPts());
  _bundle->setOrigSRate(_gbw->getCPanel()->getOrigSRate());
  _bundle->setKType(_gbw->getKalmanPredDialog()->getKalmanType());
  _bundle->setSpecD(_gbw->getKalmanPredDialog()->getSpectralDensity());
  _bundle->setPosVar(_noisev);
  _bundle->setQuatVar(_noisev*QFACTOR);
  _bundle->setUKFFlag(_gbw->getKalmanPredDialog()->getUKFFlag());
  _bundle->setCISpecD(_gbw->getCIPredDialog()->getCISpectralDensity());
  _bundle->setCIType(_gbw->getCIPredDialog()->getCIType());
  _bundle->setMMAEType(_gbw->getMMAEPredDialog()->getMMAEType());
  _bundle->setCombiner(_gbw->getMMAEPredDialog()->getCombinerType());
  _bundle->setMMAESpecD(_gbw->getMMAEPredDialog()->getMMAESpectralDensity());
  _bundle->setSPSep(_gbw->getMMAEPredDialog()->getMMAESpecDSeparator());
  // set up noise sequences if need be
  createNoise();

}

void
Dispenser::createNoise()
{
  if (_noisev != _posnoise->getVariance()) { // if these variance are different
    cerr << "Generating Noise Sequences..." << endl;
    _posnoise->generateNoise(SAMPLE_RATE, DURATION,_noisev, 0);
    _quatnoise->generateNoise(SAMPLE_RATE, DURATION,_noisev*QFACTOR, 0);
    cerr << "Noise Generation complete!" << endl;
  }

  // no matter what we should randomize the counters
  struct timeval ts; struct timezone tz;
  gettimeofday(&ts, &tz);
  _posnoise->randomizeCounters((long)(ts.tv_sec + ts.tv_usec/1e6)); 
  _quatnoise->randomizeCounters((long)(ts.tv_sec + ts.tv_usec/1e6)); 

}



Pose *
Dispenser::addNoise(Pose * curpos, int srate, double noisev)
{
  // add noise to the current pose
  // first create the noise vector for position
  if (noisev == 0.0)
    return new Pose(curpos->_tstamp, curpos->_pos, curpos->_posvel, curpos->_posacc,
                    curpos->_orient, curpos->_angvel, curpos->_angacc);

    //  get six random numbers from a normal dist.
    double dt = 1.0 / (_gbw->getCPanel()->getOrigSRate() / srate);

    Vec3d noisep(_posnoise->sample(dt, 1), 
                 _posnoise->sample(dt, 2),
                 _posnoise->sample(dt, 3));

    Vec3d noiseaxis(_quatnoise->sample(dt, 1), 
		    _quatnoise->sample(dt, 2),
		    _quatnoise->sample(dt, 3));


    double w = 1.0 - length(noiseaxis);
    Quat noiseq(w, noiseaxis);
 
    return new Pose(curpos->_tstamp, curpos->_pos+noisep, curpos->_posvel, 
                    curpos->_posacc,
                    normalize(curpos->_orient*noiseq), curpos->_angvel, curpos->_angacc);     
}


// simple printing function will take out later
void
printDataAndNoise(Pose *curpos, Pose *npos)
{
 double tmp = curpos->_orient.V()[0];
 double tmp2 = curpos->_orient.V()[1];
 double tmp3 = curpos->_orient.V()[2];
 double tmp4 = curpos->_orient.S();
 cerr << tmp << " " << tmp2 << " " << tmp3 << " " << tmp4;

 tmp = npos->_orient.V()[0];
 tmp2 = npos->_orient.V()[1];
 tmp3 = npos->_orient.V()[2];
 tmp4 = npos->_orient.S();
 cerr << " " << tmp << " " << tmp2 << " " << tmp3 << " " << tmp4;
 cerr << " 1 1 1 1" << endl;

}

// run a prediction test  
// this function is generic and can be used in all of the different testing functions
PredictionRec * 
Dispenser::runTest(int leadtime, double pinterval, int srate, double noisev)
{
  int num = _gbw->getUserRec()->getPList().num();
  UMotionRec *urec = _gbw->getUserRec();
  PredictionRec *prec = new PredictionRec(pinterval, 
                                         _gbw->getCPanel()->getOrigSRate() / srate,
                                          noisev);

  //  _gbw->getCPanel()->getCurSRate()

  // we want to interate through the motion sequence starting at a
  // point slightly into the sequence (this bypasses data with zero velocity and accel)
  // we must end when or leadtime reaches the end of the list and we
  // must iterate through by incrementing by the sample rate integer  
  
  // used in timing predictors
  int nump = 0;
  double total_time = 0.0;
  double startt = 0.0;
  double endt = 0.0;  

  for (int i = SF; i < num - SF - leadtime; i = i + srate) {
    // call the correct predictor
    Pose *curpos = urec->getPList()[i];
    if (getenv("FILTER_DATA")) {
     double tmp = urec->getPList()[i+1]->_orient.V()[0];
     double tmp2 = urec->getPList()[i+1]->_orient.V()[1];
     double tmp3 = urec->getPList()[i+1]->_orient.V()[2];
     double tmp4 = urec->getPList()[i+1]->_orient.S();
     cerr << tmp << " " << tmp2 << " " << tmp3 << " " << tmp4;
    }
    // add noise to the pose
    Pose *npose = addNoise(curpos, srate, noisev);  
    if (getenv("SIGNAL_NOISE_DATA"))
       printDataAndNoise(curpos,npose);	
    
    // timing test
    if (getenv("TIME_TEST")) {
       ++nump;
       struct timeval ts; struct timezone tz;
       gettimeofday(&ts, &tz);
       startt = (ts.tv_sec + ts.tv_usec/1e6); 
    }
    Pose *pos = _pholder->getPredictors()[(int)_ptype-1]->predict(npose,pinterval);
    if (getenv("TIME_TEST")) {
        struct timeval ts; struct timezone tz;
	gettimeofday(&ts, &tz);
        endt = (ts.tv_sec + ts.tv_usec/1e6); 
        total_time += (double)(endt - startt);
    }
    // store the pose and the corresponding true record value
    // if the prediction alg returned a new pos
    if (pos != npose) {
      prec->getPredList() += pos;
      prec->getTrueList() += urec->getPList()[i+_leadtime];
      // if we are computing percent better errors we must have the
      // random walk forecast so lets get it here
      if ((_gbw->getErrorType() == tPctBetter) || 
          (_gbw->getErrorType() == tChooseAll))
        prec->getRWalkList() += _pholder->getPredictors()[0]->predict(npose,pinterval);
    }
    prec->getCorruptList() += npose;
  }

  if (getenv("TIME_TEST")) {
     cerr << "Total Time: " << 1e6*total_time << " Number of preds: " << nump << endl;
     cerr << "Average Running Time: " << (total_time*1e6) / (double)nump << " microseconds" << endl << endl;
  }
 return prec; 
}



void
Dispenser::runSingleTest()
{

  // fill the bundle
  fillBundle();

  // set the bundle
  _pholder->getPredictors()[(int)_ptype-1]->setParamBundle(_bundle);

  // prepare the algorithm
  _pholder->getPredictors()[(int)_ptype-1]->prepare();

  // run a single test 
  PredictionRec *prec = runTest(_leadtime, _pinterval, _srate, _noisev);

  // analyze the results and report analysis to the user
  _res->insertPRec(prec);
  _res->insertURec(_gbw->getUserRec());
  _res->errorAnalysis();
  _gbw->getStatusBar()->message("Results Displayed.");
  qApp->processEvents();
   
}



// run a test over the sampling rate
void
Dispenser::runOverSRate()
{

  fillBundle();
  _res->insertURec(_gbw->getUserRec());

  _gbw->getProgDialog()->setLabelText("Working...");
  _gbw->getProgDialog()->setTotalSteps(SRATENUM);
  for (int i = SRATENUM; i > 0; --i) {
    _gbw->getProgDialog()->setProgress(SRATENUM - i+1);
    qApp->processEvents();
    _srate = i;
    _bundle->setCurSRate(_gbw->getCPanel()->getOrigSRate() / _srate);
    // set the bundle
    _pholder->getPredictors()[(int)_ptype-1]->setParamBundle(_bundle);    
    // prepare the algorithm
    _pholder->getPredictors()[(int)_ptype-1]->prepare();
   
    // run test
    PredictionRec *prec = runTest(_leadtime, _pinterval, _srate, _noisev);
    _res->insertPRec(prec);
    _res->errorAnalysis();
    _res->clearPRec();
  }
    _res->sRateResults(SRATENUM);
     _gbw->getStatusBar()->message("Results Displayed.");
}

// run tests over different prediction intervals
void
Dispenser::runOverPInt()
{
  fillBundle();
  _res->insertURec(_gbw->getUserRec());

  _gbw->getProgDialog()->setLabelText("Working...");
  _gbw->getProgDialog()->setTotalSteps(MAXLEAD);
 
  for (int i = MINLEAD; i <= MAXLEAD; i = i + PSKIP) {
    _gbw->getProgDialog()->setProgress(i);
    qApp->processEvents();
    _leadtime = i;
    _pinterval = i * (1.0 / _gbw->getCPanel()->getOrigSRate());
    _bundle->setLeadTime(_leadtime);
    // set the bundle
    _pholder->getPredictors()[(int)_ptype-1]->setParamBundle(_bundle);    
    // prepare the algorithm
    _pholder->getPredictors()[(int)_ptype-1]->prepare();
    // run test
    PredictionRec *prec = runTest(_leadtime, _pinterval, _srate, _noisev);
    _res->insertPRec(prec);
    _res->errorAnalysis();
    _res->clearPRec();
  } 
  _res->pIntResults(MINLEAD,MAXLEAD,PSKIP);
  _gbw->getStatusBar()->message("Results Displayed.");
}


void
Dispenser::runOverErrVar()
{
 _res->insertURec(_gbw->getUserRec());
 _gbw->getProgDialog()->setLabelText("Working...");
 _gbw->getProgDialog()->setTotalSteps(ERRMAX);

 for (int i = ERRSTEP; i <= ERRMAX; i += ERRSTEP) {
   _gbw->getProgDialog()->setProgress(i);
   qApp->processEvents();
   _noisev = (double)i / (double)NOISENUM;
   cerr << "Noise Variance " << _noisev << endl;
   fillBundle();  // this should create a new noise sequence
   // set the bundle
   _pholder->getPredictors()[(int)_ptype-1]->setParamBundle(_bundle);    
   // prepare the algorithm
   _pholder->getPredictors()[(int)_ptype-1]->prepare();
   // run test
   PredictionRec *prec = runTest(_leadtime, _pinterval, _srate, _noisev);
   _res->insertPRec(prec);
   _res->errorAnalysis();
   _res->clearPRec();
  } 
  _res->errResults(ERRSTEP, ERRMAX, NOISENUM); 
}



// runs various predictor specific tests that vary algorithm params
void
Dispenser::runParamTest()
{
  // go to specific testing routine based on predictor type
  switch (_ptype) {
    case tLocalPoly: 
      localPolyTests();
      break;
    case tGrey:  
      greyParamTest();
      break;
    case tExpSmooth:
      expSmoothParamTest();
      break; 
    case tCI: 
      ciParamTest();
      break;            
    case tKalman: 
      kalmanParamTest();
      break;            
    default: 
      _gbw->getStatusBar()->message("Test not supported for this predictor.",8000);

  }
}

// runs local polynomial parameter tests
void
Dispenser::localPolyTests()
{
  // go to specific local polynomial predictor
  switch(_gbw->getLPolyDialog()->getLPType()) {
   case tlagrangian: {  
      fillBundle();
      _res->insertURec(_gbw->getUserRec());
     int winmax = _gbw->getLPolyDialog()->getPWindow();    
     _gbw->getProgDialog()->setLabelText("Working...");
     _gbw->getProgDialog()->setTotalSteps(winmax);
     for (int i = 1; i <= winmax; ++i) {
       _gbw->getProgDialog()->setProgress(i);
       qApp->processEvents();
       _bundle->setNumPts(i);

       _pholder->getPredictors()[(int)_ptype-1]->setParamBundle(_bundle);
       _pholder->getPredictors()[(int)_ptype-1]->prepare();
       
       PredictionRec *prec = runTest(_leadtime, _pinterval, _srate, _noisev);

       _res->insertPRec(prec);
       _res->errorAnalysis();
       _res->clearPRec();
     }
     _res->displayParamEandG();
     break;

   }
   case tlsquares:  {
      fillBundle();
      _res->insertURec(_gbw->getUserRec());
      int winmax = _gbw->getLPolyDialog()->getPWindow();
      int degree = _gbw->getLPolyDialog()->getPDegree();
      _gbw->getProgDialog()->setLabelText("Working...");
      //    _gbw->getProgDialog()->setTotalSteps((int)((winmax*degree)-(degree*(degree-1)/2)));
      _gbw->getProgDialog()->setTotalSteps((winmax-1)*degree);

      int cnt = 0;
      for(int i = 2; i <= winmax; ++i)
	for(int j = 1; j <= degree; ++j) {
          _gbw->getProgDialog()->setProgress(++cnt);
	  qApp->processEvents();        
          if (j < i) {
	    _bundle->setDegree(j);
	    _bundle->setNumPts(i); 
	    
	    _pholder->getPredictors()[(int)_ptype-1]->setParamBundle(_bundle);
	    _pholder->getPredictors()[(int)_ptype-1]->prepare();
	    
	    PredictionRec *prec = runTest(_leadtime, _pinterval, _srate, _noisev);
	    
	    _res->insertPRec(prec);
	    _res->errorAnalysis();
	    _res->clearPRec();
          }
	}
      _res->displayParamEandG();
      break;
   }
   case tcspline: {   
     fillBundle();
     _res->insertURec(_gbw->getUserRec());
     int winmax = _gbw->getLPolyDialog()->getPWindow();    
     if (_gbw->getLPolyDialog()->getDerFly()) // do not want to include extra
       winmax -= 4;                           // terms needed by der calcs

     _gbw->getProgDialog()->setLabelText("Working...");
     _gbw->getProgDialog()->setTotalSteps(winmax-1);
     for (int i = 2; i <= winmax; ++i) {
       _gbw->getProgDialog()->setProgress(i-1);
       qApp->processEvents();
       if (_gbw->getLPolyDialog()->getDerFly())
	 _bundle->setNumPts(i+4);
       else
         _bundle->setNumPts(i);
       _pholder->getPredictors()[(int)_ptype-1]->setParamBundle(_bundle);
       _pholder->getPredictors()[(int)_ptype-1]->prepare();
       
       PredictionRec *prec = runTest(_leadtime, _pinterval, _srate, _noisev);

       _res->insertPRec(prec);
       _res->errorAnalysis();
       _res->clearPRec();
     }
     _res->displayParamEandG();
     break;
   }
   case ttaylor: { 
      _gbw->getStatusBar()->message("No Parameters to vary. Test Aborted.",8000);
      break;
   }
   default: 
     cerr << "Problem if we get HERE!!!" << endl;
  }
  
  _gbw->getStatusBar()->message("Results Displayed.");
}



// runs the grey system paramter test
void
Dispenser::greyParamTest()
{
   fillBundle();
   _res->insertURec(_gbw->getUserRec());
   int winmax = _gbw->getGreyPredDialog()->getNumPts() - 3; // we start at 4
   _gbw->getProgDialog()->setLabelText("Working...");
   _gbw->getProgDialog()->setTotalSteps(winmax);
     for (int i = 4; i <= winmax+3; ++i) {
       _gbw->getProgDialog()->setProgress(i-3);
       qApp->processEvents();
       _bundle->setGreyPts(i);

       _pholder->getPredictors()[(int)_ptype-1]->setParamBundle(_bundle);
       _pholder->getPredictors()[(int)_ptype-1]->prepare();
       
       PredictionRec *prec = runTest(_leadtime, _pinterval, _srate, _noisev);

       _res->insertPRec(prec);
       _res->errorAnalysis();
       _res->clearPRec();
     }
     _res->displayParamEandG();

} 

// run the exponential smoothing parameter test
void
Dispenser::expSmoothParamTest()
{
   switch(_gbw->getExpSmDialog()->getExpSmoothType()) {
     case tDoubleExp: {  
       fillBundle();
       _res->insertURec(_gbw->getUserRec());
       // we will do 99 interations
       _gbw->getProgDialog()->setLabelText("Working...");
       _gbw->getProgDialog()->setTotalSteps(99);
       double alpha = 0.01;
       while(alpha < 1) {
	 _gbw->getProgDialog()->setProgress((int)(alpha*100));
         qApp->processEvents();
         _bundle->setAlpha(alpha);
	 _pholder->getPredictors()[(int)_ptype-1]->setParamBundle(_bundle);
	 _pholder->getPredictors()[(int)_ptype-1]->prepare();
       
	 PredictionRec *prec = runTest(_leadtime, _pinterval, _srate, _noisev);

	 _res->insertPRec(prec);
	 _res->errorAnalysis();
	 _res->clearPRec();
         alpha += 0.01;
       }
       _res->displayParamEandG();
       break;
     }
   default:
     _gbw->getStatusBar()->message("Test not supported for this predictor.",8000);
   }
}


// runs the kalman paramter test
void
Dispenser::kalmanParamTest()
{
   fillBundle();
   _res->insertURec(_gbw->getUserRec());
   int specd = _gbw->getKalmanPredDialog()->getSpectralDensity();
   _gbw->getProgDialog()->setLabelText("Working...");
   _gbw->getProgDialog()->setTotalSteps(specd);
     for (int i = 1; i <= specd; ++i) {
       _gbw->getProgDialog()->setProgress(i);
       qApp->processEvents();
       _bundle->setSpecD(i);
       _pholder->getPredictors()[(int)_ptype-1]->setParamBundle(_bundle);
       _pholder->getPredictors()[(int)_ptype-1]->prepare();
       
       PredictionRec *prec = runTest(_leadtime, _pinterval, _srate, _noisev);

       _res->insertPRec(prec);
       _res->errorAnalysis();
       _res->clearPRec();
     }
     _res->displayParamEandG();

} 


// run the Covariance intersection parameter test
void
Dispenser::ciParamTest()
{
   fillBundle();
   _res->insertURec(_gbw->getUserRec());
   int specd = _gbw->getCIPredDialog()->getCISpectralDensity();
   _gbw->getProgDialog()->setLabelText("Working...");
   _gbw->getProgDialog()->setTotalSteps(specd);
     for (int i = 1; i <= specd; ++i) {
       _gbw->getProgDialog()->setProgress(i);
       qApp->processEvents();
       _bundle->setSpecD(i);
       _pholder->getPredictors()[(int)_ptype-1]->setParamBundle(_bundle);
       _pholder->getPredictors()[(int)_ptype-1]->prepare();
       
       PredictionRec *prec = runTest(_leadtime, _pinterval, _srate, _noisev);

       _res->insertPRec(prec);
       _res->errorAnalysis();
       _res->clearPRec();
     }
     _res->displayParamEandG();

} 


#define MCSIM 10
#define NUMDATAPTS 32000

// this procedure runs a full test over different sampling rates, pred times,
// vars noise on a particular dataset.  It finds the best param for each data point
// then does a monte carlo simulation and prints out the data point to
// a file

void
Dispenser::runOverAll()
{
   	
  // we do not want to deal with the choose all error type
   if (_gbw->getErrorType() == tChooseAll) {
     _gbw->getStatusBar()->message("Test not supported for the Choose All Metric.",8000);
     return;
   }

  _res->insertURec(_gbw->getUserRec());
  _gbw->getProgDialog()->setLabelText("Working...");
  _gbw->getProgDialog()->setTotalSteps(NUMDATAPTS);

  FILE *fpp, *fpq = NULL;
  QString fstr = _gbw->getCPanel()->getCurFile()->text();
  // open the files for writing
  if ((fpp = fopen(("../data/modeling_data/"+fstr+"_p.dat").latin1(),"w")) == NULL){
     cerr << "Problem opening position datapoint file." << endl;
     return;
  }
  if ((fpq = fopen(("../data/modeling_data/"+fstr+"_q.dat").latin1(), "w")) == NULL) {
     cerr << "Problem opening orientation datapoint file." << endl;
     return;
  }

  int p = 0;    
  // main sim loops
  for (int i = ERRSTEP; i <= ERRMAX; i += ERRSTEP) { // over noise
    _noisev = (double)i / (double)NOISENUM;
    _bundle->setPosVar(_noisev);
    _bundle->setQuatVar(_noisev*QFACTOR);
    createNoise();
    for (int j = SRATENUM; j > 0; --j) { // over sample rates
      _srate = j;
      _bundle->setCurSRate(_gbw->getCPanel()->getOrigSRate() / _srate);
      for (int k = MINLEAD; k <= MAXLEAD; k = k + PSKIP) { // over pred times
	_leadtime = k;
	_pinterval = k * (1.0 / _gbw->getCPanel()->getOrigSRate());
	_bundle->setLeadTime(_leadtime); 
        ++p;
	_gbw->getProgDialog()->setProgress(p);
        qApp->processEvents();  
        // at this point we need to find optimal alg. params
        // to run the Monte Carlo Sims
	findOptimalParams();
        // run the sims and store the data points
	runMCSimulation(fpp, fpq);
	qApp->processEvents();  
      }
    }
  }
  fclose(fpp);
  fclose(fpq);      

  _gbw->getTextBox()->append("Full Test Complete. Data Saved to files.");  
}

// runs through the algorithmic param space to find the best fit for
// the current dataset and system params.  It stores the optimal
// param for the given predictor in the param bundle

void
Dispenser::findOptimalParams()
{
   // we have to switch on predictor type
   switch (_ptype) {
    case tExpSmooth:
      if (_gbw->getExpSmDialog()->getExpSmoothType() == tDoubleExp) {
        _bundle->setExpSmoothType(_gbw->getExpSmDialog()->getExpSmoothType());
	double a = 0.01;
        _alphap = _alphaq = 0.0;
        double errp = 1000.0;
        double errq = 1000.0;
        if (_gbw->getErrorType() == tPctBetter) 
	  errp = errq = 0.0;
    
        while (a < 0.99) {
	  _bundle->setAlpha(a);
	  _pholder->getPredictors()[(int)_ptype-1]->setParamBundle(_bundle);
	  _pholder->getPredictors()[(int)_ptype-1]->prepare();
       
          PredictionRec *prec = runTest(_leadtime, _pinterval, _srate, _noisev);

          // compute error metric
          _res->insertPRec(prec);
	  ErrorNum errnum;
	  if (_gbw->getErrorType() == tPctBetter) 
	     errnum = _res->pctBetter(0);
          else if (_gbw->getErrorType() == tRootMSE)
             errnum = _res->rootMSE(0);
          else if (_gbw->getErrorType() == tLInfinity)
             errnum = _res->lInfinity(0);
	  _res->clearPRec();         

          // find param
          if (_gbw->getErrorType() == tPctBetter) {
	     if (errnum._gpos > errp) {
		errp = errnum._gpos;
		_alphap = a;   
             }
              if (errnum._grot > errq) {
		errq = errnum._grot;
		_alphaq = a;   
             }
          }
          else {
             if (errnum._gpos < errp) {
		errp = errnum._gpos;
		_alphap = a;   
             }
             if (errnum._grot < errq) {
		errq = errnum._grot;
		_alphaq = a;   
	     }
	  }
	  a += 0.01;
	}
      }
      else 
        _gbw->getStatusBar()->message("Test supported for this Predictor.",8000);
      break;
    default:
       _gbw->getStatusBar()->message("Test supported for this Predictor.",8000); 
   }
}


// this function does the actual runs and stores the results in two data files
// one for position and one for orientation

void
Dispenser::runMCSimulation(FILE *fpp, FILE *fpq)
{
  // we have to switch on predictor type
   switch (_ptype) {
    case tExpSmooth:
      if (_gbw->getExpSmDialog()->getExpSmoothType() == tDoubleExp) {
	_bundle->setExpSmoothType(_gbw->getExpSmDialog()->getExpSmoothType());
        double gpos = 0.0;
        double grot = 0.0;
        double pvar = 0.0;
        double qvar = 0.0;
        double psp = 0.0;
        double qsp = 0.0;
        double dpvar = 0.0;
        double dqvar = 0.0;
 
        // first get position
        for (int i = 0; i < MCSIM; ++i) {
          _bundle->setAlpha(_alphap);
          _pholder->getPredictors()[(int)_ptype-1]->setParamBundle(_bundle);
	  _pholder->getPredictors()[(int)_ptype-1]->prepare();
       
          PredictionRec *prec = runTest(_leadtime, _pinterval, _srate, _noisev);

          // compute error metric
          _res->insertPRec(prec);
	  ErrorNum errnum;
	  if (_gbw->getErrorType() == tPctBetter) 
	     errnum = _res->pctBetter(0);
          else if (_gbw->getErrorType() == tRootMSE)
             errnum = _res->rootMSE(0);
          else if (_gbw->getErrorType() == tLInfinity)
             errnum = _res->lInfinity(0);
	  _res->clearPRec();                 
          
          gpos += errnum._gpos;
        }
       
        // get orientation
        for (int i = 0; i < MCSIM; ++i) {
          _bundle->setAlpha(_alphaq);
          _pholder->getPredictors()[(int)_ptype-1]->setParamBundle(_bundle);
	  _pholder->getPredictors()[(int)_ptype-1]->prepare();
       
          PredictionRec *prec = runTest(_leadtime, _pinterval, _srate, _noisev);

          // compute error metric
          _res->insertPRec(prec);
	  ErrorNum errnum;
	  if (_gbw->getErrorType() == tPctBetter) 
	     errnum = _res->pctBetter(0);
          else if (_gbw->getErrorType() == tRootMSE)
             errnum = _res->rootMSE(0);
          else if (_gbw->getErrorType() == tLInfinity)
             errnum = _res->lInfinity(0);              
          
          grot += errnum._grot;
           
          // need to get quantitative signal info here
	  pvar += _gbw->getSigAnD()->getSigAnalyzer()->computePosVar(prec->getCorruptList());
          qvar += _gbw->getSigAnD()->getSigAnalyzer()->computeRotVar(prec->getCorruptList());
          psp += _gbw->getSigAnD()->getSigAnalyzer()->computePosSpd(prec->getCorruptList());
          qsp += _gbw->getSigAnD()->getSigAnalyzer()->computeRotSpd(prec->getCorruptList());          
          dpvar += _gbw->getSigAnD()->getSigAnalyzer()->computeDPosVar(prec->getCorruptList()); 
          dqvar += _gbw->getSigAnD()->getSigAnalyzer()->computeDRotVar(prec->getCorruptList()); 
	  // delete the prec
          _res->clearPRec();     
  
        }  
        // store datapoints in file
        grot /= MCSIM;
        gpos /= MCSIM;
        pvar /= MCSIM;
        pvar *= 12.0;
        qvar /= MCSIM;
        psp  /= MCSIM;
        psp  *= 12.0;
        qsp  /= MCSIM;
        dpvar /= MCSIM;
        dpvar *= 12.0;
        dqvar /= MCSIM;
	if (_gbw->getErrorType() != tPctBetter)  // covert to inches
          gpos *= 12.0;           
      

     
        fprintf(fpp,"%lf %lf %lf %lf %lf %lf %lf %lf\n", gpos, _pinterval*1000, 
		_gbw->getCPanel()->getOrigSRate() / _srate,
                _noisev, _alphap, pvar, psp, dpvar);
        
        fprintf(fpq,"%lf %lf %lf %lf %lf %lf %lf %lf\n", grot, _pinterval*1000, 
		_gbw->getCPanel()->getOrigSRate() / _srate,
                _noisev*QFACTOR, _alphaq, qvar, qsp, dqvar);
	
      }
      else 
        _gbw->getStatusBar()->message("Test supported for this Predictor.",8000);
      break;
    default:
       _gbw->getStatusBar()->message("Test supported for this Predictor.",8000); 
   }
}
