// functions for computing signal analysis and quantification routines
// within the SignalAnalyzer Object


#include "ui.H"
#include "pose.H"
#include "ndifftools.H"
#include "predictors/mmatrix.H"
#include "mollify.H"
#include "ABKP.H"

#include <qmultilineedit.h>
#include <qstring.h>
#include <qprogressdialog.h>
#include <qapplication.h>

#include <iostream.h>
#include <sys/time.h>

#define QFACTOR 0.1 
#define SAMPLE_RATE 10000 // for noise generation
#define DURATION 24 // 24 seconds of noise
#define SF 4
#define NUMD 5

#define DEGREE 2
#define EXTEND 20

// constructor

NumDiff::NumDiff(GlobalWidget *gw) : _gbw(gw), _tmprec(MAXREC),
                                     _smrec(MAXREC), _truerec(MAXREC),
                                     _noisev(0.0), _srate(0.0),
                                     _posewin(50)
{
  // set up noise generators
  struct timeval ts; struct timezone tz;
  gettimeofday(&ts, &tz);
  _posnoise = new GaussWhiteNoise((long)(ts.tv_sec + ts.tv_usec/1e6)); 
  gettimeofday(&ts, &tz);
  _quatnoise = new GaussWhiteNoise((long)(ts.tv_sec + ts.tv_usec/1e6)); 

}


// desctuctor

NumDiff::~NumDiff()
{
  delete _posnoise;
  delete _quatnoise;

  for (int i = 0; i < _tmprec.num(); ++i) {
     delete _tmprec[i];
     delete _truerec[i];
  }
  for (int i = 0; i < _smrec.num(); ++i) 
     delete _smrec[i];

  for (int i = 0; i < _posewin.num(); ++i)
     delete _posewin[i]; 
}


// methods

Pose *
NumDiff::addNoise(Pose *cpos, int srate, double noisev)
{
 // add noise to the current pose
  // first create the noise vector for position
  if (noisev == 0.0)
    return new Pose(cpos->_tstamp, cpos->_pos, cpos->_posvel, cpos->_posacc,
                    cpos->_orient, cpos->_angvel, cpos->_angacc);

  //  get six random numbers from a normal dist.
  double dt = 1.0 / (_gbw->getCPanel()->getOrigSRate() / srate);

  Vec3d noisep(_posnoise->sample(dt, 1), 
	       _posnoise->sample(dt, 2),
	       _posnoise->sample(dt, 3));

  Vec3d noiseaxis(_quatnoise->sample(dt, 1), 
		  _quatnoise->sample(dt, 2),
		  _quatnoise->sample(dt, 3));


  double w = 1.0 - length(noiseaxis);
  Quat noiseq(w, noiseaxis);
 
  return new Pose(cpos->_tstamp, cpos->_pos+noisep, cpos->_posvel, 
		  cpos->_posacc, normalize(cpos->_orient*noiseq), 
                  cpos->_angvel, cpos->_angacc);     

}


void
NumDiff::prepare()
{
   // clear the holding tank 
   for (int i = 0; i < _tmprec.num(); ++i) {
      delete _truerec[i];
      delete _tmprec[i];
   }
   _truerec.clear();
   _tmprec.clear();
   for (int i = 0; i < _smrec.num(); ++i) 
      delete _smrec[i];
   _smrec.clear();

   for (int i = 0; i < _posewin.num(); ++i)
      delete _posewin[i];
   _posewin.clear();

  _srate  =  _gbw->getCPanel()->getSRateInt();
  _noisev =  _gbw->getCPanel()->getNoiseVar();

  // create noise if needed then add it to the data set
   if (_noisev != _posnoise->getVariance()) { // if these variance are different
    cerr << "Generating Noise Sequences..." << endl;
    _posnoise->generateNoise(SAMPLE_RATE, DURATION,_noisev, 0);
    _quatnoise->generateNoise(SAMPLE_RATE, DURATION,_noisev*QFACTOR, 0);
    cerr << "Noise Generation complete!" << endl;
  }

  // no matter what we should randomize the counters
  struct timeval ts; struct timezone tz;
  gettimeofday(&ts, &tz);
  _posnoise->randomizeCounters((long)(ts.tv_sec + ts.tv_usec/1e6)); 
  _quatnoise->randomizeCounters((long)(ts.tv_sec + ts.tv_usec/1e6)); 
 
  int num = _gbw->getUserRec()->getPList().num();
  UMotionRec *urec = _gbw->getUserRec();
  // adding noise, the tmprec will contain poses based on the SRATE
  for (int i = SF; i < num - SF; i = i + _srate) {
     _truerec += new Pose(urec->getPList()[i]);
     _tmprec += addNoise(urec->getPList()[i], _srate, _noisev);    
  }
}

// computational methods

void
NumDiff::computeSimpleND()
{
   for (int i = 0; i < _tmprec.num(); ++i) {
     Pose *tmppos = new Pose(_tmprec[i]);

     // create and update the pose window
      if (_posewin.num() < NUMD) {
	 _posewin += tmppos;
	 continue;
      }
      else {
         Pose *p;
	 for (int i = 1; i < NUMD; ++i) {
            if (i == 1)
	       p = _posewin[0];  
	    _posewin[i-1] = _posewin[i];            
         }
         delete p;
	 _posewin.pop();
	 _posewin += tmppos;
      }
     
      // at this point we should have 5 smoothed points
      // we can now take the numerical derivative for the ith pose  
      Pose *dpose; 
      if (_gbw->getNDiffToolDialog()->getCDiff()) {
	 dpose = centralDiff(_posewin);       
	 dpose->_pos = _posewin[NUMD-3]->_pos;
	 dpose->_orient = _posewin[NUMD-3]->_orient;
      }
      else {
	 dpose = backwardDiff(_posewin);       
	 dpose->_pos = _posewin[NUMD-1]->_pos;
	 dpose->_orient = _posewin[NUMD-1]->_orient;
      }
      _smrec += dpose;     
   }
}


void
NumDiff::computeNDLSOrtho(int win, int deg)
{
   FIGarray<Pose*> smlist;
  
  for (int i = 0; i < _tmprec.num(); ++i) {
     Pose *tmppos = new Pose(_tmprec[i]);
    
     // create and update the pose window
      if (_posewin.num() < win) {
    	 _posewin += tmppos;
	 continue;
      }
      else {
         Pose *p;
	 for (int i = 1; i < win; ++i) {
            if (i == 1)
	       p = _posewin[0];  
	    _posewin[i-1] = _posewin[i];            
         }
         delete p;
	 _posewin.pop();
	 _posewin += tmppos;
      }
       
      // error check
      if ((deg+1 >  win) || (win != _posewin.num())) {
     	 cerr << "Problem, degree higher than num points!!!" << endl;
	 return;
      }

      Pose *dpose;
      if (!_gbw->getNDiffToolDialog()->getDerAvg()) {
	 // get smooth poses from interpolation	
	 for (int i = win-NUMD; i < win; ++i) {
	    Pose *smpose = LSOrtho(win, deg, _posewin[i]->_tstamp, _posewin);
	    smlist += smpose;
	 }

	 // at this point we should have 5 smoothed points
	 // we can now take the numerical derivative for the ith pose   
	 if (_gbw->getNDiffToolDialog()->getCDiff()) {
	    dpose = centralDiff(smlist);       
	    dpose->_pos = smlist[NUMD-3]->_pos;
	    dpose->_orient = smlist[NUMD-3]->_orient;
	 }
	 else {
	    dpose = backwardDiff(smlist);       
	    dpose->_pos = smlist[NUMD-1]->_pos;
	    dpose->_orient = smlist[NUMD-1]->_orient;
	 }
      }
      else { // do derivative averaging
        for (int i = 0; i < win; ++i) {
	    Pose *smpose = LSOrtho(win, deg, _posewin[i]->_tstamp, _posewin);
	    smlist += smpose;
	 }
  
	 if (_gbw->getNDiffToolDialog()->getCDiff()) {
	    dpose = diffAvg(smlist,1);       
	    dpose->_pos = smlist[NUMD-3]->_pos;
	    dpose->_orient = smlist[NUMD-3]->_orient;
	 }
	 else {
	    dpose = diffAvg(smlist,0);       
	    dpose->_pos = smlist[NUMD-1]->_pos;
	    dpose->_orient = smlist[NUMD-1]->_orient;
	 }

      }

      _smrec += dpose;     
     // delete the smlist for good housekeeping
       for (int i = 0; i < smlist.num(); ++i)
	  delete smlist[i];
       smlist.clear();
  }


}


Pose * 
NumDiff::LSOrtho(int win, int deg, double t, FIGarray<Pose*> posewin)
{
 
  // create data structures we need
  double *q     = new double[deg+1]; 
  double *alpha = new double[deg+1];
  double *beta  = new double[deg+1];
  double *gamma = new double[deg+1];
  MMatrix *Qi   = new MMatrix(deg+1, win); // need to store intermediate q's
  Vec3d  *ap    = new Vec3d[deg+1];
  Quat   *aq    = new Quat[deg+1];

  // base case
  q[0] = 1.0;
  int i;
  for(i = 0; i < win; ++i) 
    (*Qi)(0,i) = 1.0;
  
  double ttot = 0.0;
  for(i = 0; i < win; ++i)
   ttot += posewin[i]->_tstamp;

  gamma[0] = win;
  
  if (deg >= 1) {
    q[1] = t - (1.0 / win)*ttot;
    // store q values for ti's
    for(i = 0; i < win; ++i)
      (*Qi)(1,i) = posewin[i]->_tstamp - (1.0 / win)*ttot;
    gamma[1] = 0.0;  
    for (i = 0; i < win; ++i)
      gamma[1] += ((*Qi)(1,i) * (*Qi)(1,i));
    alpha[1] = (1.0 / win)*ttot;
  }
   
  // iteratively solve the recurrence relation
  int j;
  for (j = 1; j <= deg-1; ++j) {
    // compute the jth gamma
    double tmpgamma = 0.0;
    for(i = 0; i < win; ++i)
      tmpgamma += ((*Qi)(j,i)*(*Qi)(j,i));
    gamma[j] = tmpgamma;
    
    // compute the j+1th alpha
    double tmpalpha = 0.0;
    for(i = 0; i < win; ++i)
      tmpalpha += posewin[i]->_tstamp * ((*Qi)(j,i)*(*Qi)(j,i));
    alpha[j+1] = (1.0 / gamma[j]) * tmpalpha;
    
    // compute jth beta
    beta[j] = gamma[j] / gamma[j-1];

    // compute new j+1th q for the time t
    q[j+1] = t * q[j] - (alpha[j+1] * q[j]) -
             (beta[j]*q[j-1]);

    // compute the j+1th q's for the ti's and store in Qi
    for (i = 0; i < win; ++i)
      (*Qi)(j+1,i) = (posewin[i]->_tstamp * (*Qi)(j,i)) - (alpha[j+1] * (*Qi)(j,i)) -
                     (beta[j] * (*Qi)(j-1,i));
  }

  // compute the coefficients (ak's)
  int k;
  for(k = 0; k <= deg; ++k) {
    Vec3d tmppos(0.0,0.0,0.0);
    Quat  tmpqrot(0.0,0.0,0.0,0.0);
    double tmp = 0.0;
    for (i = 0; i < win; ++i) {
      tmppos += (*Qi)(k,i) * posewin[i]->_pos;
      tmpqrot += (*Qi)(k,i) * posewin[i]->_orient;
      tmp += ((*Qi)(k,i) * (*Qi)(k,i));
    }
    gamma[k] = tmp;
    ap[k] = (1.0 / gamma[k]) * tmppos;
    aq[k] = (1.0 / gamma[k]) * tmpqrot;
  }

  Vec3d ppos(0.0,0.0,0.0);
  Quat  prot(0.0,0.0,0.0,0.0);
  // finally compute smoothed pose
  for (k = 0; k <= deg; ++k) {
   ppos += q[k] * ap[k];
   prot += q[k] * aq[k];
  }

  // delete memory
  delete [] q;
  delete [] alpha;
  delete [] beta;
  delete [] gamma;
  delete [] ap;
  delete [] aq;
  Qi->deleteMatrix();
  delete Qi;

  return new Pose(t, ppos, normalize(prot));
}


// Savitsky Golay routines
void
NumDiff::computeNDSavGolay(int left, int right, int deg)
{ 
   // first compute the Savitzky-Golay coefficients
   double *coeff = savGolay(left, right, deg);
   FIGarray<Pose*> smlist;    

   for (int i = 0; i < _tmprec.num()-right; ++i) {
     Pose *tmppos = new Pose(_tmprec[i]);   
 
     // create and update the pose window
      if (_posewin.num() < left+right+1) {
   	 _posewin += tmppos;
	 continue;
      }
      else {
         Pose *p;
	 for (int i = 1; i < left+right+1; ++i) {
            if (i == 1)
	       p = _posewin[0];  
	    _posewin[i-1] = _posewin[i];            
         }
         delete p;
	 _posewin.pop();
	 _posewin += tmppos;
      }
   
      // with the updated window, smooth point i between left and right
      // in the pose win and put it in the smlist
      Vec3d pos(0,0,0);
      Quat rot(0,0,0,0);
      for (int j = -left; j < right+1; ++j) {
	 pos += (coeff[j+left] * _posewin[j+left]->_pos);
	 rot += (coeff[j+left] * _posewin[j+left]->_orient);
      }
      Pose *smpose = new Pose(_posewin[left]->_tstamp, pos, normalize(rot));

      // update the smoothing window
       if (smlist.num() < NUMD) {
	 smlist += smpose;
	 continue;
      }
      else {
         Pose *p;
	 for (int i = 1; i < NUMD; ++i) {
            if (i == 1)
	       p = smlist[0];  
	    smlist[i-1] = smlist[i];            
         }
         delete p;
	 smlist.pop();
	 smlist += smpose;
      }  
      
      // now we differentiate 
       Pose *dpose; 
      if (_gbw->getNDiffToolDialog()->getCDiff()) {
	 dpose = centralDiff(smlist);       
	 dpose->_pos = smlist[NUMD-3]->_pos;
	 dpose->_orient = smlist[NUMD-3]->_orient;
      }
      else {
	 dpose = backwardDiff(smlist);       
	 dpose->_pos = smlist[NUMD-1]->_pos;
	 dpose->_orient = smlist[NUMD-1]->_orient;
      }
      _smrec += dpose;     
   }
   // delete the smlist for good housekeeping
   for (int i = 0; i < smlist.num(); ++i)
      delete smlist[i];
   smlist.clear();
   delete [] coeff;
}


// computes the Savitzky Golay coefficients
double *
NumDiff::savGolay(int left, int right, int deg)
{
   double *coeffs = new double[left+1+right];

   // set up the design matrix
   MMatrix a(left+1+right,deg+1);
   for (int i = -left; i < right+1; ++i)
      for (int j = 0; j < deg+1; ++j)
	 a(i+left,j) = pow(i,j);
 
   // compute the coefficients
   MMatrix a2 = minverse2(transp(a) * a);

   for (int i = -left; i < right+1; ++i) {
     double sum = 0.0;
     for (int m = 0; m < deg+1; ++m) 
	sum += a2(0,m) * pow(i,m);
     coeffs[i+left] = sum;
   }

   return coeffs;
}


void
NumDiff::computeNDWavelet(int left, int right, int deg, char *wavelet)
{
   // matlab must be on to do this technique
  MATLABEngine *eng = _gbw->getMEngine();
  
  if (!eng->getEngine()) {// if its not there do nothing
     _gbw->getTextBox()->append("Turn on Matlab to perform routine.");
    return;
  } 

  FIGarray<Pose*> smlist;    
  // FIGarray<Pose*> smlist2;
  _gbw->getProgDialog()->setLabelText("Working...");
  _gbw->getProgDialog()->setTotalSteps(_tmprec.num()-right-1);
  
   for (int i = 0; i < _tmprec.num()-right; ++i) {
      _gbw->getProgDialog()->setProgress(i);
      qApp->processEvents();
     Pose *tmppos = new Pose(_tmprec[i]);    
     // create and update the pose window
      if (_posewin.num() < left+right+1) {
   	 _posewin += tmppos;
	 continue;
      }
      else {
         Pose *p;
	 for (int i = 1; i < left+right+1; ++i) {
            if (i == 1)
	       p = _posewin[0];  
	    _posewin[i-1] = _posewin[i];            
         }
         delete p;
	 _posewin.pop();
	 _posewin += tmppos;
      }
      // extend the posewin out x units
      FIGarray<Pose*> exlist;
      double s = _gbw->getCPanel()->getCurSRate();
      for (int i = 0; i < _posewin.num(); ++i) 
	 exlist += new Pose(_posewin[i]);
      for (int i = 0; i < EXTEND; ++i) 
         exlist += LSOrtho(_posewin.num(), DEGREE, 
                           _posewin[_posewin.num()-1]->_tstamp + (i+1)*(1.0/s),
                           _posewin);
      
      // use the wavelet denoising algorithm 
      FIGarray<Pose*> dnlist2 = waveletDenoise(exlist,deg,wavelet);
      FIGarray<Pose*> dnlist;
       for (int i = 0; i < _posewin.num(); ++i) 
	 dnlist += dnlist2[i];
    

      //////////////////////////////////////////////////////////////
      // we have denoised the signal, smooth it out more using LSOrtho
      // get smooth poses from interpolation
      //  for (int i = 0; i < dnlist.num(); ++i) {
      //    Pose *smpose = LSOrtho(dnlist.num(), DEGREE, dnlist[i]->_tstamp, 
      //                           dnlist);
      //   smlist2 += smpose;
      //  } 
      ////////////////////////////////////////////////////////////////

      Pose *dpose; 
      // check if we are doing derivative averaging
      if (!_gbw->getNDiffToolDialog()->getDerAvg()) {
	 // now create the smooth list for taking derivatives
	 // we need 5 points, want to try and take them from the center of 
	 // the list 
	 // start by taking the last 5
	 for (int i = left - NUMD +1; i < left+1; ++i) {
	    Pose *tmp = new Pose(dnlist[i]);
	    smlist += tmp;      
	 }
	 // now we differentiate 
	 if (_gbw->getNDiffToolDialog()->getCDiff()) {
	    dpose = centralDiff(smlist);       
	    dpose->_pos = smlist[NUMD-3]->_pos;
	    dpose->_orient = smlist[NUMD-3]->_orient;
	 }
	 else {
	    dpose = backwardDiff(smlist);       
	    dpose->_pos = smlist[NUMD-1]->_pos;
	    dpose->_orient = smlist[NUMD-1]->_orient;
	 }
      }
      else  { // do derivative averaging
	 if (_gbw->getNDiffToolDialog()->getCDiff()) {
	    dpose = diffAvg(dnlist,1);       
	    dpose->_pos = dnlist[dnlist.num()-3]->_pos;
	    dpose->_orient = dnlist[dnlist.num()-3]->_orient;
	 }
	 else {
	    dpose = diffAvg(dnlist,0);       
	    dpose->_pos = dnlist[dnlist.num()-1]->_pos;
	    dpose->_orient = dnlist[dnlist.num()-1]->_orient;
	 }
      }
      _smrec += dpose;    

      for (int i = 0; i < dnlist.num(); ++i)
	 delete dnlist2[i];
      dnlist2.clear(); 
   
      for (int i = 0; i < exlist.num(); ++i)
	 delete exlist[i];
      exlist.clear(); 
      // for (int i = 0; i < smlist2.num(); ++i)
      //   delete smlist2[i];
      // smlist2.clear();

      // delete the smlist for good housekeeping
      for (int i = 0; i < smlist.num(); ++i)
	 delete smlist[i];
      smlist.clear();
   }
}


// routine uses matlab to denoise signals

FIGarray<Pose*> 
NumDiff::waveletDenoise(FIGarray<Pose*> posewin, int deg, char *wavelet)
{
   FIGarray<Pose*> dnlist;
   MATLABEngine *eng = _gbw->getMEngine();

   eng->runCmd(QString("clear"));
 
   // send the data to matlab
   int num = posewin.num();
   double *x  = new double[num];
   double *y  = new double[num];
   double *z  = new double[num];
   double *qx = new double[num];
   double *qy = new double[num];
   double *qz = new double[num];
   double *qw = new double[num];
   
   for (int i = 0; i < num; ++i) {
     x[i]  = posewin[i]->_pos[0];
     y[i]  = posewin[i]->_pos[1];
     z[i]  = posewin[i]->_pos[2];
     qx[i] = posewin[i]->_orient.V()[0];
     qy[i] = posewin[i]->_orient.V()[1];
     qz[i] = posewin[i]->_orient.V()[2];
     qw[i] = posewin[i]->_orient.S();
   }

  mxArray *X = mxCreateDoubleMatrix(1, num, mxREAL);
  mxSetName(X,"XX");
  memcpy((void*)mxGetPr(X), (void*)x, num*sizeof(double));
  eng->sendArray(X); 
  
  mxSetName(X,"YY");  
  memcpy((void*)mxGetPr(X), (void*)y, num*sizeof(double));
  eng->sendArray(X); 

  mxSetName(X,"ZZ");  
  memcpy((void*)mxGetPr(X), (void*)z, num*sizeof(double));
  eng->sendArray(X);
  
  mxSetName(X,"QXX");  
  memcpy((void*)mxGetPr(X), (void*)qx, num*sizeof(double));
  eng->sendArray(X);

  mxSetName(X,"QYY");  
  memcpy((void*)mxGetPr(X), (void*)qy, num*sizeof(double));
  eng->sendArray(X);

  mxSetName(X,"QZZ");  
  memcpy((void*)mxGetPr(X), (void*)qz, num*sizeof(double));
  eng->sendArray(X);
  
  mxSetName(X,"QWW");  
  memcpy((void*)mxGetPr(X), (void*)qw, num*sizeof(double));
  eng->sendArray(X);
  mxDestroyArray(X);

  
  // do the denoising with wden
  QString thresh;
  switch (_gbw->getNDiffToolDialog()->getThreshType()) {
   case tRigsure:
     thresh = QString("'rigrsure',");
     break;
   case tHeursure:
     thresh = QString("'heursure',");
     break;
   case tsqtwolog:
     thresh = QString("'sqtwolog',");
     break;
   case tMinimax:
     thresh = QString("'minimaxi',");
     break;
   case tkillDet:
     thresh = QString("'killdetail'");
     break;
   default:
     cerr << "Problem in wavelet Denoise" << endl;
  }

  QString hsthresh;
  if (_gbw->getNDiffToolDialog()->getHSThresh())
     hsthresh = QString("'s',");
  else
     hsthresh = QString("'h',");

  if (thresh == QString("'killdetail'")) {
  eng->runCmd(QString("x = wdencmp('gbl',XX,")+QString(wavelet)+QString(",")+
              QString::number(deg)+QString(",")+QString::number(1000)+
              QString(",")+hsthresh+QString::number(1)+QString(")"));
  eng->runCmd(QString("y = wdencmp('gbl',YY,")+QString(wavelet)+QString(",")+
              QString::number(deg)+QString(",")+QString::number(1000)+
              QString(",")+hsthresh+QString::number(1)+QString(")"));
  eng->runCmd(QString("z = wdencmp('gbl',ZZ,")+QString(wavelet)+QString(",")+
              QString::number(deg)+QString(",")+QString::number(1000)+
              QString(",")+hsthresh+QString::number(1)+QString(")"));
  eng->runCmd(QString("qx = wdencmp('gbl',QXX,")+QString(wavelet)+QString(",")+
              QString::number(deg)+QString(",")+QString::number(1000)+
              QString(",")+hsthresh+QString::number(1)+QString(")"));
  eng->runCmd(QString("qy = wdencmp('gbl',QYY,")+QString(wavelet)+QString(",")+
              QString::number(deg)+QString(",")+QString::number(1000)+
              QString(",")+hsthresh+QString::number(1)+QString(")"));
  eng->runCmd(QString("qz = wdencmp('gbl',QZZ,")+QString(wavelet)+QString(",")+
              QString::number(deg)+QString(",")+QString::number(1000)+
              QString(",")+hsthresh+QString::number(1)+QString(")"));
  eng->runCmd(QString("qw = wdencmp('gbl',QWW,")+QString(wavelet)+QString(",")+
              QString::number(deg)+QString(",")+QString::number(1000)+
              QString(",")+hsthresh+QString::number(1)+QString(")"));
  }
  else {  
  eng->runCmd(QString("x = wden(XX,")+thresh+hsthresh+QString("'mln',")+
              QString::number(deg)+QString(",")+QString(wavelet)+QString(")"));
  eng->runCmd(QString("y = wden(YY,")+thresh+hsthresh+QString("'mln',")+
              QString::number(deg)+QString(",")+QString(wavelet)+QString(")"));
  eng->runCmd(QString("z = wden(ZZ,")+thresh+hsthresh+QString("'mln',")+
              QString::number(deg)+QString(",")+QString(wavelet)+QString(")"));
  eng->runCmd(QString("qx = wden(QXX,")+thresh+hsthresh+QString("'mln',")+
              QString::number(deg)+QString(",")+QString(wavelet)+QString(")"));
  eng->runCmd(QString("qy = wden(QYY,")+thresh+hsthresh+QString("'mln',")+
              QString::number(deg)+QString(",")+QString(wavelet)+QString(")"));
  eng->runCmd(QString("qz = wden(QZZ,")+thresh+hsthresh+QString("'mln',")+
              QString::number(deg)+QString(",")+QString(wavelet)+QString(")"));
  eng->runCmd(QString("qw = wden(QWW,")+thresh+hsthresh+QString("'mln',")+
              QString::number(deg)+QString(",")+QString(wavelet)+QString(")"));
  }

  // get results back from matlab
  mxArray *XX = eng->getArray("x");
  memcpy((void*)x, (void*)mxGetPr(XX), num*sizeof(double)); 
  XX = eng->getArray("y");
  memcpy((void*)y, (void*)mxGetPr(XX), num*sizeof(double)); 
  XX = eng->getArray("z");
  memcpy((void*)z, (void*)mxGetPr(XX), num*sizeof(double)); 
  XX = eng->getArray("qx");
  memcpy((void*)qx, (void*)mxGetPr(XX), num*sizeof(double)); 
  XX = eng->getArray("qy");
  memcpy((void*)qy, (void*)mxGetPr(XX), num*sizeof(double)); 
  XX = eng->getArray("qz");
  memcpy((void*)qz, (void*)mxGetPr(XX), num*sizeof(double)); 
  XX = eng->getArray("qw");
  memcpy((void*)qw, (void*)mxGetPr(XX), num*sizeof(double)); 
  mxDestroyArray(XX);   

  // put the smoothed data into the dnlist
  for (int i = 0; i < num; ++i) {
     Pose *tmp = new Pose(posewin[i]->_tstamp,
                          Vec3d(x[i],y[i],z[i]),
                          posewin[i]->_posvel,
                          posewin[i]->_posacc,
                          normalize(Quat(qx[i],qy[i],qz[i],qw[i])),
                          posewin[i]->_angvel,
                          posewin[i]->_angacc);
     dnlist += tmp;
  }
  
  delete [] x;
  delete [] y;
  delete [] z;
  delete [] qx;
  delete [] qy;
  delete [] qz;
  delete [] qw;
 
  return dnlist;  
}


void
NumDiff::computeNDMollify(int win)
{
  FIGarray<Pose*> smlist;
  
  for (int i = 0; i < _tmprec.num(); ++i) {
     Pose *tmppos = new Pose(_tmprec[i]);
    
     // create and update the pose window
      if (_posewin.num() < win) {
    	 _posewin += tmppos;
	 continue;
      }
      else {
         Pose *p;
	 for (int i = 1; i < win; ++i) {
            if (i == 1)
	       p = _posewin[0];  
	    _posewin[i-1] = _posewin[i];            
         }
         delete p;
	 _posewin.pop();
	 _posewin += tmppos;
      }
      
       // extend the posewin out x units
      FIGarray<Pose*> exlist;
      double s = _gbw->getCPanel()->getCurSRate();
      for (int i = 0; i < _posewin.num(); ++i) 
	 exlist += new Pose(_posewin[i]);
      for (int i = 0; i < EXTEND; ++i) 
         exlist += LSOrtho(_posewin.num(), DEGREE, 
                           _posewin[_posewin.num()-1]->_tstamp + (i+1)*(1.0/s),
                           _posewin);
     
      FIGarray<Pose*> mollist2 = mollify(exlist, exlist.num()); 
      FIGarray<Pose*> mollist;
       for (int i = 0; i < _posewin.num(); ++i) 
	 mollist += mollist2[i];
      

      Pose *dpose;
      if (!_gbw->getNDiffToolDialog()->getDerAvg()) {
	 for (int i = win-NUMD; i < win; ++i) {
	    Pose *smpose = new Pose(mollist[i]);
	    smlist += smpose;
	 }
     
	 // at this point we should the 5 smoothed points
	 // we can now take the numerical derivative for the ith pose   
	 if (_gbw->getNDiffToolDialog()->getCDiff()) {
	    dpose = centralDiff(smlist);       
	    dpose->_pos = smlist[NUMD-3]->_pos;
	    dpose->_orient = smlist[NUMD-3]->_orient;
	 }
	 else {
	    dpose = backwardDiff(smlist);       
	    dpose->_pos = smlist[NUMD-1]->_pos;
	    dpose->_orient = smlist[NUMD-1]->_orient;
	 }
      }
      else { // do derivative averaging
	 if (_gbw->getNDiffToolDialog()->getCDiff()) {
	    dpose = diffAvg(mollist,1);       
	    dpose->_pos = mollist[NUMD-3]->_pos;
	    dpose->_orient = mollist[NUMD-3]->_orient;
	 }
	 else {
	    dpose = diffAvg(mollist,0);       
	    dpose->_pos = mollist[NUMD-1]->_pos;
	    dpose->_orient = mollist[NUMD-1]->_orient;
	 }

      }

      _smrec += dpose;     
      for (int i = 0; i < mollist2.num(); ++i)
	delete mollist2[i];
      mollist2.clear();
      for (int i = 0; i < exlist.num(); ++i)
	 delete exlist[i];
      exlist.clear(); 
      // delete the smlist for good housekeeping
      for (int i = 0; i < smlist.num(); ++i)
	 delete smlist[i];
      smlist.clear();
  }

}


FIGarray<Pose*>
NumDiff::mollify(FIGarray<Pose*> posewin, int win)
{

 FIGarray<Pose*> mollist;

 int num = posewin.num();
 double *x  = new double[num];
 double *y  = new double[num];
 double *z  = new double[num];
 double *qx = new double[num];
 double *qy = new double[num];
 double *qz = new double[num];
 double *qw = new double[num];

 for (int i = 0; i < num; ++i) {
    x[i]  = posewin[i]->_pos[0];
    y[i]  = posewin[i]->_pos[1];
    z[i]  = posewin[i]->_pos[2];
    qx[i] = posewin[i]->_orient.V()[0];
    qy[i] = posewin[i]->_orient.V()[1];
    qz[i] = posewin[i]->_orient.V()[2];
    qw[i] = posewin[i]->_orient.S();   
 }

 double delta = (((num) / 2.0) - 2.0) / (3.0 * (num));
 // do the mollification
 deltaMol(num-1, delta, x);
 deltaMol(num-1, delta, y);
 deltaMol(num-1, delta, z);
 deltaMol(num-1, delta, qx);
 deltaMol(num-1, delta, qy);
 deltaMol(num-1, delta, qz);
 deltaMol(num-1, delta, qw);


 // put the smoothed data into the mollist
  for (int i = 0; i < num; ++i) {
     Pose *tmp = new Pose(posewin[i]->_tstamp,
                          Vec3d(x[i],y[i],z[i]),
                          posewin[i]->_posvel,
                          posewin[i]->_posacc,
                          normalize(Quat(qx[i],qy[i],qz[i],qw[i])),
                          posewin[i]->_angvel,
                          posewin[i]->_angacc);
     mollist += tmp;
  }
  
  delete [] x;
  delete [] y;
  delete [] z;
  delete [] qx;
  delete [] qy;
  delete [] qz;
  delete [] qw;
 
  return mollist;  
}


void
NumDiff::computeNDABKP(int mode, int win)
{
   FIGarray<Pose*> smlist;
  
   for (int i = 0; i < _tmprec.num(); ++i) {
     Pose *tmppos = new Pose(_tmprec[i]);
    
     // create and update the pose window
      if (_posewin.num() < win) {
    	 _posewin += tmppos;
	 continue;
      }
      else {
         Pose *p;
	 for (int i = 1; i < win; ++i) {
            if (i == 1)
	       p = _posewin[0];  
	    _posewin[i-1] = _posewin[i];            
         }
         delete p;
	 _posewin.pop();
	 _posewin += tmppos;
      }
      
        
      FIGarray<Pose*> abkplist = ABKP(_posewin, win, mode); 

      Pose *dpose;
      if (!_gbw->getNDiffToolDialog()->getDerAvg()) {
	 for (int i = win-NUMD; i < win; ++i) {
	    Pose *smpose = new Pose(abkplist[i]);
	    smlist += smpose;
	 }
     
	 // at this point we should the 5 smoothed points
	 // we can now take the numerical derivative for the ith pose   
	 if (_gbw->getNDiffToolDialog()->getCDiff()) {
	    dpose = centralDiff(smlist);       
	    dpose->_pos = smlist[NUMD-3]->_pos;
	    dpose->_orient = smlist[NUMD-3]->_orient;
	 }
	 else {
	    dpose = backwardDiff(smlist);       
	    dpose->_pos = smlist[NUMD-1]->_pos;
	    dpose->_orient = smlist[NUMD-1]->_orient;
	 }
      }
      else { // do derivative averaging
	 if (_gbw->getNDiffToolDialog()->getCDiff()) {
	    dpose = diffAvg(abkplist,1);       
	    dpose->_pos = abkplist[NUMD-3]->_pos;
	    dpose->_orient = abkplist[NUMD-3]->_orient;
	 }
	 else {
	    dpose = diffAvg(abkplist,0);       
	    dpose->_pos = abkplist[NUMD-1]->_pos;
	    dpose->_orient = abkplist[NUMD-1]->_orient;
	 }

      }

      _smrec += dpose;     
      for (int i = 0; i < abkplist.num(); ++i)
	delete abkplist[i];
      abkplist.clear();
      // delete the smlist for good housekeeping
      for (int i = 0; i < smlist.num(); ++i)
	 delete smlist[i];
      smlist.clear();
  }

}

FIGarray<Pose*>
NumDiff::ABKP(FIGarray<Pose*> posewin, int win, int mode)
{
   FIGarray<Pose*> smlist;

 int num = posewin.num();
 double *x  = new double[num];
 double *y  = new double[num];
 double *z  = new double[num];
 double *qx = new double[num];
 double *qy = new double[num];
 double *qz = new double[num];
 double *qw = new double[num];
 
 double *x2  = new double[num];
 double *y2  = new double[num];
 double *z2  = new double[num];
 double *qx2 = new double[num];
 double *qy2 = new double[num];
 double *qz2 = new double[num];
 double *qw2 = new double[num];

 for (int i = 0; i < num; ++i) {
    x[i]  = posewin[i]->_pos[0];
    y[i]  = posewin[i]->_pos[1];
    z[i]  = posewin[i]->_pos[2];
    qx[i] = posewin[i]->_orient.V()[0];
    qy[i] = posewin[i]->_orient.V()[1];
    qz[i] = posewin[i]->_orient.V()[2];
    qw[i] = posewin[i]->_orient.S();   
 }

 double dt = 1.0 / _gbw->getCPanel()->getCurSRate();
 // do the particular ABKP algorithm depending on mode
 mfilter(mode, num, dt, x, x2);
 mfilter(mode, num, dt, y, y2);
 mfilter(mode, num, dt, z, z2);
 mfilter(mode, num, dt, qx, qx2);
 mfilter(mode, num, dt, qy, qy2);
 mfilter(mode, num, dt, qz, qz2);
 mfilter(mode, num, dt, qw, qw2);


 // put the smoothed data into the mollist
  for (int i = 0; i < num; ++i) {
     Pose *tmp = new Pose(posewin[i]->_tstamp,
                          Vec3d(x2[i],y2[i],z2[i]),
                          posewin[i]->_posvel,
                          posewin[i]->_posacc,
                          normalize(Quat(qx2[i],qy2[i],qz2[i],qw2[i])),
                          posewin[i]->_angvel,
                          posewin[i]->_angacc);
     smlist += tmp;
  }
  
  delete [] x;
  delete [] y;
  delete [] z;
  delete [] qx;
  delete [] qy;
  delete [] qz;
  delete [] qw;
  delete [] x2;
  delete [] y2;
  delete [] z2;
  delete [] qx2;
  delete [] qy2;
  delete [] qz2;
  delete [] qw2;
 
  return smlist;  


}



// differentiation routines
Pose *
NumDiff::diffAvg(FIGarray<Pose*> posewin, int difftype)
{

   //  we compute a set of derivatives and take the average
   FIGarray<Pose*> derlist;
   FIGarray<Pose*> smlist;
   for (int i = 1; i < posewin.num()-1; ++i) {
      Pose *tmppos = new Pose(posewin[i]);
      // create a 5 stencil moving window 
      if (smlist.num() < NUMD) {
	 smlist += tmppos;
	 continue;
      }
      else {
         Pose *p;
	 for (int i = 1; i < NUMD; ++i) {
            if (i == 1)
	       p = smlist[0];  
	    smlist[i-1] = smlist[i];            
         }
         delete p;
	 smlist.pop();
	 smlist += tmppos;
      }  
      // calculate derivative and store in the der list using
      // either a backward or central diff scheme
      if (!difftype)
	 derlist += backwardDiff(smlist);
      else
         derlist += centralDiff(smlist);
   }
   // now we take the averages of the derivatives
   Vec3d pvelavg(0.0,0.0,0.0);
   Vec3d paccavg(0.0,0.0,0.0);
   Quat  avelavg(0,0,0,0);
   Quat  aaccavg(0,0,0,0);
   for (int i = 0; i < derlist.num(); ++i) {
      pvelavg += derlist[i]->_posvel;
      paccavg += derlist[i]->_posacc;
      avelavg += derlist[i]->_angvel;
      aaccavg += derlist[i]->_angacc;
   }
   Pose *p = new Pose();       
   if (!difftype)
      p->_tstamp = posewin[posewin.num()-1]->_tstamp;
   else
      p->_tstamp = posewin[posewin.num()-3]->_tstamp;
   p->_posvel = (1.0 / derlist.num()) * pvelavg;
   p->_posacc = (1.0 / derlist.num()) * paccavg;
   p->_angvel = (1.0 / derlist.num()) * avelavg;
   p->_angacc = (1.0 / derlist.num()) * aaccavg;

   // do some cleanup
   for (int i = 0; i < smlist.num(); ++i)
      delete smlist[i];
   smlist.clear();
   for (int i = 0; i < derlist.num(); ++i)
      delete derlist[i];
   derlist.clear();
   
   return p;
}


Pose *
NumDiff::backwardDiff(FIGarray<Pose*> smlist)
{
   // first compute 1st and 2nd derivative for the position vector
    // using 4th order explicit backward differencing
    Vec3d tmp1 = 25.0 * smlist[4]->_pos;
    Vec3d tmp2 = -48.0 * smlist[3]->_pos;
    Vec3d tmp3 = 36.0 * smlist[2]->_pos;
    Vec3d tmp4 = -16.0 * smlist[1]->_pos;
    Vec3d tmp5 = 3.0 * smlist[0]->_pos;
 
    double srate =  _gbw->getCPanel()->getCurSRate();
    Vec3d posvel = (1.0 / (12.0 * (1.0 / srate))) *
                   (tmp1 + tmp2 + tmp3 + tmp4 + tmp5);   
  
    tmp1 = 3.0 * smlist[4]->_pos;
    tmp2 = -9.0 * smlist[3]->_pos;
    tmp3 = 10.0 * smlist[2]->_pos;
    tmp4 = -5.0 * smlist[1]->_pos;
    tmp5 = smlist[0]->_pos;

    Vec3d posacc = (11.0 / (12.0 * (1.0 / (srate * srate)))) *
                   (tmp1 + tmp2 + tmp3 + tmp4 + tmp5);   

    // now for first and second derivatives for quaternions
    Quat qtmp1 = 25.0 * smlist[4]->_orient;
    Quat qtmp2 = -48.0 * smlist[3]->_orient;
    Quat qtmp3 = 36.0 * smlist[2]->_orient;
    Quat qtmp4 = -16.0 * smlist[1]->_orient;
    Quat qtmp5 = 3.0 * smlist[0]->_orient;

    Quat qtmp6 = (1.0 / (12.0 * (1.0 / srate))) *
             (qtmp1 + qtmp2 + qtmp3 + qtmp4 + qtmp5);   
    
    Quat angvel = 2.0 * inverse(smlist[4]->_orient) * qtmp6;

    qtmp1 = 25.0 * inverse(smlist[4]->_orient);
    qtmp2 = -48.0 * inverse(smlist[3]->_orient);
    qtmp3 = 36.0 * inverse(smlist[2]->_orient);
    qtmp4 = -16.0 * inverse(smlist[1]->_orient);
    qtmp5 = 3.0 * inverse(smlist[0]->_orient);

    Quat qtmp7 = (1.0 / (12.0 * (1.0 / srate))) *
             (qtmp1 + qtmp2 + qtmp3 + qtmp4 + qtmp5);   
    Quat qparta = 2.0 * qtmp7 * qtmp6; //_posewin[_numwin-1]->_orient;
  
    qtmp1 = 3.0 * smlist[4]->_orient;
    qtmp2 = -9.0 * smlist[3]->_orient;
    qtmp3 = 10.0 * smlist[2]->_orient;
    qtmp4 = -5.0 * smlist[1]->_orient;
    qtmp5 = smlist[0]->_orient;

    qtmp6 = (11.0 / (12.0 * (1.0 / (srate * srate)))) *
             (qtmp1 + qtmp2 + qtmp3 + qtmp4 + qtmp5);   
 
    Quat qpartb = 2.0 * inverse(smlist[4]->_orient) * qtmp6;

    Quat angacc = (qparta + qpartb);

    Pose *npos = new Pose();
    npos->_tstamp = smlist[4]->_tstamp;
    npos->_posvel = posvel;
    npos->_posacc = posacc;
    npos->_angvel = angvel;
    npos->_angacc = angacc;
    return npos;
}

Pose *
NumDiff::centralDiff(FIGarray<Pose*> smlist)
{
  double delta =  1.0 / _gbw->getCPanel()->getCurSRate();

   // always assume smlist is has 5 elements
   // compute first and second derivatives using
   // 4th order central differencing scheme
   
   Vec3d diff = (-1.0 * smlist[4]->_pos) + (8.0 * smlist[3]->_pos) - 
                 (8.0 * smlist[1]->_pos) + smlist[0]->_pos;
   Vec3d posvel = (1.0 / (12.0 * delta)) * diff;

   Vec3d diff2 = (-1.0 * smlist[4]->_pos) + (16.0 * smlist[3]->_pos) - 
                 (30.0 * smlist[2]->_pos) +
                 (16.0 * smlist[1]->_pos) - smlist[0]->_pos;
   Vec3d posacc = (1.0 / (12.0 * delta * delta)) * diff2;

   // now for angular velocity and acceleration
   Quat qa = -1.0 * smlist[4]->_orient;
   Quat qb =  8.0 * smlist[3]->_orient;
   Quat qc =  8.0 * smlist[1]->_orient;
   Quat qd = smlist[0]->_orient;
   Quat qe = (1.0 / (12.0*delta)) * (qa + qb - qc + qd);
   Quat qi = inverse(smlist[2]->_orient);
   Quat qfinal = 2.0 * qi * qe;
   Quat angvel = qfinal;

   Quat q1 = -1.0 * inverse(smlist[4]->_orient);
   Quat q2 = 8.0 * inverse(smlist[3]->_orient);
   Quat q3 = 8.0 * inverse(smlist[1]->_orient);
   Quat q4 = inverse(smlist[0]->_orient); 
   Quat q5 = 2.0 * ((1.0 / (12.0 * delta)) * (q1 + q2 - q3 + q4)) *
             smlist[2]->_orient;
   
   Quat q6 = -1.0 * smlist[4]->_orient;
   Quat q7 = 16.0 * smlist[3]->_orient;
   Quat q8 = 30.0 * smlist[2]->_orient;
   Quat q9 = 16.0 * smlist[1]->_orient;
   Quat q10 = smlist[0]->_orient;
  
   Quat q11 = 2.0 * inverse(smlist[2]->_orient) * 
            ((1.0 / (12.0 * delta*delta)) * (q6 +q7 - q8 + q9 - q10));

   Quat angacc = q5 + q11;


    Pose *npos = new Pose();
    npos->_tstamp = smlist[2]->_tstamp;
    npos->_posvel = posvel;
    npos->_posacc = posacc;
    npos->_angvel = angvel;
    npos->_angacc = angacc;
    return npos;

}



ErrorVal
NumDiff::computeNDRMSE(int win)
{
   int ahead = 0;
   if (win > 1)
      ahead = win; //+NUMD;

   ErrorVal err;
   
   // now for component wise rmse
  Vec3d sumpv(0.0);
  Vec3d sumpa(0.0);
  Quat  sumrv(0,0,0,0);
  Quat  sumra(0,0,0,0);
  for (int j = 0; j < _smrec.num(); ++j) {
    Vec3d tmp  = _smrec[j]->_posvel - _truerec[ahead+j]->_posvel;
    sumpv += tmp*tmp;

    Quat tmp2 =  _smrec[j]->_angvel - _truerec[ahead+j]->_angvel;
    sumrv += specialmultiply(tmp2,tmp2);
    
    Vec3d tmp3  = _smrec[j]->_posacc - _truerec[ahead+j]->_posacc;
    sumpa += tmp3*tmp3;
   
    Quat tmp4 =  _smrec[j]->_angacc - _truerec[ahead+j]->_angacc;
    sumra += specialmultiply(tmp4,tmp4);
  }
  err._pvel = mysqrt((1.0 / _smrec.num())* sumpv);
  err._avel = mysqrt((1.0 / _smrec.num())* sumrv);
  err._pacc = mysqrt((1.0 / _smrec.num())* sumpa);
  err._aacc = mysqrt((1.0 / _smrec.num())* sumra);

  return err;
}



// display routines

void
NumDiff::display(const char *title, int win) 
{
   _gbw->getTextBox()->append(" ");
   _gbw->getTextBox()->append(title);
   _gbw->getTextBox()->append("-----------------------------------------------");
   _gbw->getTextBox()->append("File: " + _gbw->getCPanel()->getCurFile()->text());
   _gbw->getTextBox()->append("Sampling Rate: " +  _gbw->getCPanel()->getCurSP()->text() +
                             "      Noise Variance: " +
                             QString::number(_gbw->getCPanel()->getNoiseVar(), 'g',8));
  
   ErrorVal err = computeNDRMSE(win); 

  _gbw->getTextBox()->append(" ");
   char buf[128];
   
   _gbw->getTextBox()->append("       Linear Velocity		Angular Velocity");
    sprintf(buf, "%5.8lf %39.8lf",12.0*err._pvel[0],err._avel.V()[0]);
   _gbw->getTextBox()->append("VX:     " + QString(buf));
   sprintf(buf, "%5.8lf %39.8lf",12.0*err._pvel[1],err._avel.V()[1]);
   _gbw->getTextBox()->append("VY:     " + QString(buf));
   sprintf(buf, "%5.8lf %39.8lf",12.0*err._pvel[2],err._avel.V()[2]);
   _gbw->getTextBox()->append("VZ:     " + QString(buf)); 

   _gbw->getTextBox()->append(" ");
   _gbw->getTextBox()->append("       Linear Acceleration		Angular Acceleration"); 
   
   sprintf(buf, "%5.8lf %39.8lf",12.0*err._pacc[0],err._aacc.V()[0]);
   _gbw->getTextBox()->append("AX:     " + QString(buf));
   sprintf(buf, "%5.8lf %39.8lf",12.0*err._pacc[1],err._aacc.V()[1]);
   _gbw->getTextBox()->append("AY:     " + QString(buf));
   sprintf(buf, "%5.8lf %39.8lf",12.0*err._pacc[2],err._aacc.V()[2]);
   _gbw->getTextBox()->append("AZ:     " + QString(buf)); 

   _gbw->getTextBox()->append(" ");

}


void
NumDiff::drawGraphs()
{
  // lets graph data if matlab is open
  MATLABEngine *eng = _gbw->getMEngine();
  
  if (!eng->getEngine()) {// if its not there do nothing
     _gbw->getTextBox()->append("Turn on Matlab to show graphical results.");
    return;
  } 
  eng->runCmd(QString("clear"));
 
  // store the the timestamp numbers
  int nums = _smrec.num();
  double  *ts = new double[nums];
  for (int i = 0; i < nums; ++i)
    ts[i] = _smrec[i]->_tstamp;
  mxArray *T = mxCreateDoubleMatrix(1, nums, mxREAL);
  mxSetName(T,"TimeS");
  memcpy((void*)mxGetPr(T), (void*)ts, nums*sizeof(double));
  eng->sendArray(T); 
  mxDestroyArray(T);
  delete [] ts;

  int numt = _truerec.num();
  double  *tt = new double[numt];
  for (int i = 0; i < numt; ++i)
    tt[i] = _truerec[i]->_tstamp;
  mxArray *T2 = mxCreateDoubleMatrix(1, numt, mxREAL);
  mxSetName(T2,"TimeT");
  memcpy((void*)mxGetPr(T2), (void*)tt, numt*sizeof(double));
  eng->sendArray(T2); 
  mxDestroyArray(T2);
  delete [] tt;

  // first set of plots will be true vs. corrupt vs. smooth
  sigPlotter(numt,nums);

  // second plot the "true" vs. the numerical 1st derivatives (i.e., velocity)
  velPlotter(numt,nums);
}

void
NumDiff::displaySimpleND()
{

   if (!_gbw->getNDiffToolDialog()->getCDiff()) 
      display("Simple Numerical Differentiation RMS Results (Backward)",NUMD);
   else 
      display("Simple Numerical Differentiation RMS Results (Central)",NUMD-2);
}


void
NumDiff::displayNDSavGolay(int left, int right, int deg)
{

   if (!_gbw->getNDiffToolDialog()->getCDiff()) 
     display("Numerical Differentation RMS Results via Savitsky-Golay (Backward)",
	   left+right+1+(NUMD-right));
   else 
     display("Numerical Differentation RMS Results via Savitsky-Golay (Central)",
	   left+right+1+(NUMD-right)-2); 
   
//     cerr << _smrec.num() << " " << _truerec.num() << endl;
      //   for (int i = 0; i < left+right+NUMD+1; ++i)
//	cerr << i << " " << _smrec[i]->_tstamp << " " << _truerec[i+left+right+1+(NUMD-right)-2]->_tstamp << endl;
  
   // do matlab plotting
   drawGraphs();
 
}


void
NumDiff::displayNDLSOrtho(int win)
{
   // numerical results
   if (!_gbw->getNDiffToolDialog()->getCDiff()) 
      display("Numerical Differentiation RMS Results using LS Ortho Smoothing (Backward)",win);
   else 
      display("Numerical Differentiation RMS Results using LS Ortho Smoothing (Central)",win-2);
            
   // do matlab plotting
   drawGraphs();
}


void
NumDiff::displayNDWavelet(int right, int left, int deg, char *wavelet)
{
   char buf[128];

   if (!_gbw->getNDiffToolDialog()->getCDiff()) {
      sprintf(buf, "Numerical Differentation RMS Results via Wavelet %s Denoising (Backward)", wavelet);
      display(buf,left+right+1);
   }
   else {
      sprintf(buf, "Numerical Differentation RMS Results via Wavelet %s Denoising (Central)", wavelet);
      display(buf,left+right-1); 
   }

   // do matlab plotting
   drawGraphs();

}

void
NumDiff::displayNDMollify(int win)
{
   // numerical results
   if (!_gbw->getNDiffToolDialog()->getCDiff()) 
      display("Numerical Differentiation RMS Results using Mollification (Backward)",win);
   else 
      display("Numerical Differentiation RMS Results using Mollification (Central)",win-2);
            
   // do matlab plotting
   drawGraphs();

}


void
NumDiff::displayNDABKP(int mode, int win)
{

 // numerical results
   if (!_gbw->getNDiffToolDialog()->getCDiff()) 
      switch (mode) {
       case 1:
	 display("Numerical Differentiation RMS Results using Anderssen-Bloomfield (Odd) (Backward)",win);
	 break;
       case 2:
	 display("Numerical Differentiation RMS Results using Anderssen-Bloomfield (No Odd) (Backward)",win);   
	 break;
       case 3:
         display("Numerical Differentiation RMS Results using Kosarev-Pantos (Odd) (Backward)",win);
         break;
       case 4:
         display("Numerical Differentiation RMS Results using Kosarev-Pantos (No Odd) (Backward)",win);
	 break;
       default:
	 cerr << "Problem if we get here" << endl; 
      } 
   else 
     switch (mode) {
       case 1:
	 display("Numerical Differentiation RMS Results using Anderssen-Bloomfield (Odd) (Central)",win-2);
	 break;
       case 2:
	 display("Numerical Differentiation RMS Results using Anderssen-Bloomfield (No Odd) (Central)",win-2);   
	 break;
       case 3:
         display("Numerical Differentiation RMS Results using Kosarev-Pantos (Odd) (Central)",win-2);
         break;
       case 4:
         display("Numerical Differentiation RMS Results using Kosarev-Pantos (No Odd) (Central)",win-2);
	 break;
       default:
	 cerr << "Problem if we get here" << endl; 
      }  
            

   // do matlab plotting
   drawGraphs();
}


// plot routines

void
NumDiff::sigPlotter(int numt, int nums)
{
 
  MATLABEngine *eng = _gbw->getMEngine();	    
  // create large figure and position it
  eng->runCmd(QString("figure(1)"));
  eng->runCmd(QString("set(1, 'Position', [470 -35 800 980])"));
 
   
  // first set of plots will be true vs. corrupt vs. smooth
  // for x
  double *t = new double[numt];
  double *c = new double[numt];
  double *s = new double[nums];
  for (int i = 0; i < numt; ++i) {
     t[i] = _truerec[i]->_pos[0];
     c[i] = _tmprec[i]->_pos[0];
  }
  for (int i = 0; i < nums; ++i) 
     s[i] = _smrec[i]->_pos[0];
  
  plotTCS("subplot(4,2,1)", "title('X: True vs. Corrupt vs. Smoothed')",
          "xlabel('Time')", "ylabel('X')", t,c,s, numt, nums);
  
  // for y

  for (int i = 0; i < numt; ++i) {
     t[i] = _truerec[i]->_pos[1];
     c[i] = _tmprec[i]->_pos[1];
  }
  for (int i = 0; i < nums; ++i) 
     s[i] = _smrec[i]->_pos[1];

  plotTCS("subplot(4,2,3)", "title('Y: True vs. Corrupt vs. Smoothed')",
          "xlabel('Time')", "ylabel('Y')", t,c,s, numt,nums);

  // for z
   for (int i = 0; i < numt; ++i) {
     t[i] = _truerec[i]->_pos[2];
     c[i] = _tmprec[i]->_pos[2];
  }
  for (int i = 0; i < nums; ++i) 
     s[i] = _smrec[i]->_pos[2];
  plotTCS("subplot(4,2,5)", "title('Z: True vs. Corrupt vs. Smoothed')",
          "xlabel('Time')", "ylabel('Z')", t,c,s, numt,nums);

  // for qx
  for (int i = 0; i < numt; ++i) {
     t[i] = _truerec[i]->_orient.V()[0];
     c[i] = _tmprec[i]->_orient.V()[0];
  }
  for (int i = 0; i < nums; ++i)
     s[i] = _smrec[i]->_orient.V()[0];
  
  plotTCS("subplot(4,2,2)", "title('QX: True vs. Corrupt vs. Smoothed')",
          "xlabel('Time')", "ylabel('QX')", t,c,s, numt, nums);

  // for qy
   for (int i = 0; i < numt; ++i) {
     t[i] = _truerec[i]->_orient.V()[1];
     c[i] = _tmprec[i]->_orient.V()[1];
  }
  for (int i = 0; i < nums; ++i)
     s[i] = _smrec[i]->_orient.V()[1];
  plotTCS("subplot(4,2,4)", "title('QY: True vs. Corrupt vs. Smoothed')",
          "xlabel('Time')", "ylabel('QY')", t,c,s, numt, nums);

  // for qz
  for (int i = 0; i < numt; ++i) {
     t[i] = _truerec[i]->_orient.V()[2];
     c[i] = _tmprec[i]->_orient.V()[2];
  }
  for (int i = 0; i < nums; ++i)
     s[i] = _smrec[i]->_orient.V()[2];
  plotTCS("subplot(4,2,6)", "title('QZ: True vs. Corrupt vs. Smoothed')",
          "xlabel('Time')", "ylabel('QZ')", t,c,s, numt, nums);

  // for qw
  for (int i = 0; i < numt; ++i) {
     t[i] = _truerec[i]->_orient.S();
     c[i] = _tmprec[i]->_orient.S();
  }
  for (int i = 0; i < nums; ++i)
     s[i] = _smrec[i]->_orient.S();
  plotTCS("subplot(4,2,8)", "title('QW: True vs. Corrupt vs. Smoothed')",
          "xlabel('Time')", "ylabel('QW')", t,c,s, numt, nums);

  delete [] t;
  delete [] s;
  delete [] c;
}


void
NumDiff::velPlotter(int numt, int nums)
{
  
  MATLABEngine *eng = _gbw->getMEngine();
  eng->runCmd(QString("figure(2)"));
  eng->runCmd(QString("set(2, 'Position', [470 -35 800 980])"));

  double *vt = new double[numt];
  double *vc = new double[nums];

  
  for (int i = 0; i < numt; ++i)
     vt[i] = _truerec[i]->_posvel[0];
  for (int i = 0; i < nums; ++i)
     vc[i] = _smrec[i]->_posvel[0];
 
  plotVels("subplot(3,2,1)", "title('VX: True vs. Computed')",
          "xlabel('Time')", "ylabel('VX')", vt,vc,numt,nums);

  // for vy
  for (int i = 0; i < numt; ++i)
     vt[i] = _truerec[i]->_posvel[1];
  for (int i = 0; i < nums; ++i)
     vc[i] = _smrec[i]->_posvel[1];
  plotVels("subplot(3,2,3)", "title('VY: True vs. Computed')",
          "xlabel('Time')", "ylabel('VY')", vt,vc,numt,nums);

  // for vz
  for (int i = 0; i < numt; ++i)
     vt[i] = _truerec[i]->_posvel[2];
  for (int i = 0; i < nums; ++i)
     vc[i] = _smrec[i]->_posvel[2];
  plotVels("subplot(3,2,5)", "title('VZ: True vs. Computed')",
          "xlabel('Time')", "ylabel('VZ')", vt,vc,numt,nums);

  // for vqx
   
   for (int i = 0; i < numt; ++i) 
     vt[i] = _truerec[i]->_angvel.V()[0];
   for (int i = 0; i < nums; ++i) 
     vc[i] = _smrec[i]->_angvel.V()[0];
   
  plotVels("subplot(3,2,2)", "title('VQX: True vs. Computed')",
          "xlabel('Time')", "ylabel('VQX')", vt,vc,numt,nums);

  // for vqy
   for (int i = 0; i < numt; ++i) 
     vt[i] = _truerec[i]->_angvel.V()[1];
   for (int i = 0; i < nums; ++i) 
     vc[i] = _smrec[i]->_angvel.V()[1];
  plotVels("subplot(3,2,4)", "title('VQY: True vs. Computed')",
	   "xlabel('Time')", "ylabel('VQY')", vt,vc,numt,nums);

  // for vqz
   for (int i = 0; i < numt; ++i) 
     vt[i] = _truerec[i]->_angvel.V()[2];
   for (int i = 0; i < nums; ++i) 
     vc[i] = _smrec[i]->_angvel.V()[2];
  plotVels("subplot(3,2,6)", "title('VQZ: True vs. Computed')",
          "xlabel('Time')", "ylabel('VQZ')", vt,vc,numt,nums);

  delete [] vt;
  delete [] vc;  

}


void 
NumDiff::plotTCS(char *subplot, char *title, char *xlabel, char *ylabel,
                 double *t, double *c, double *s, int numt, int nums)
{

 MATLABEngine *eng = _gbw->getMEngine();
 
 // send the data to matlab
  mxArray *T = mxCreateDoubleMatrix(1, numt, mxREAL);
  mxSetName(T,"T");
  memcpy((void*)mxGetPr(T), (void*)t, numt*sizeof(double));
  eng->sendArray(T); 
  mxDestroyArray(T);
 
  mxArray *C = mxCreateDoubleMatrix(1, numt, mxREAL);
  mxSetName(C,"C");
  memcpy((void*)mxGetPr(C), (void*)c, numt*sizeof(double));
  eng->sendArray(C); 
  mxDestroyArray(C);

  mxArray *S = mxCreateDoubleMatrix(1, nums, mxREAL);
  mxSetName(S,"S");
  memcpy((void*)mxGetPr(S), (void*)s, nums*sizeof(double));
  eng->sendArray(S); 
  mxDestroyArray(S);

  //  char mbuf[2048];
  //  engOutputBuffer(eng->getEngine(), mbuf, 2048);  
  //  cerr << mbuf << endl;

  eng->runCmd(QString(subplot)); 
  eng->runCmd(QString("plot(TimeT,T,'k')"));
  eng->runCmd(QString("hold on"));
  eng->runCmd(QString("plot(TimeT,C,'r')"));
  eng->runCmd(QString("plot(TimeS,S,'b')"));
  eng->runCmd(QString(xlabel));
  eng->runCmd(QString(ylabel));
  eng->runCmd(QString(title));
  eng->runCmd(QString("hold off"));
}


void 
NumDiff::plotVels(char *subplot, char *title, char *xlabel, char *ylabel,
                 double *vt, double *vc, int numt, int nums)
{

 MATLABEngine *eng = _gbw->getMEngine();
 
 // send the data to matlab
  mxArray *VT = mxCreateDoubleMatrix(1, numt, mxREAL);
  mxSetName(VT,"VT");
  memcpy((void*)mxGetPr(VT), (void*)vt, numt*sizeof(double));
  eng->sendArray(VT); 
  mxDestroyArray(VT);
 
  mxArray *VC = mxCreateDoubleMatrix(1, nums, mxREAL);
  mxSetName(VC,"VC");
  memcpy((void*)mxGetPr(VC), (void*)vc, nums*sizeof(double));
  eng->sendArray(VC); 
  mxDestroyArray(VC);

  eng->runCmd(QString(subplot)); 
  eng->runCmd(QString("plot(TimeT,VT,'k')"));
  eng->runCmd(QString("hold on"));
  eng->runCmd(QString("plot(TimeS,VC,'b')"));
  eng->runCmd(QString(xlabel));
  eng->runCmd(QString(ylabel));
  eng->runCmd(QString(title));
  eng->runCmd(QString("hold off"));
}


/////////////////////////////////////////// used in LS Ortho
// Use this code if we are doing one smoothed pose at a time
/*
      // update the smoothing window
       if (smlist.num() < NUMD) {
	 smlist += smpose;
	 continue;
      }
      else {
         Pose *p;
	 for (int i = 1; i < NUMD; ++i) {
            if (i == 1)
	       p = smlist[0];  
	    smlist[i-1] = smlist[i];            
         }
         delete p;
	 smlist.pop();
	 smlist += smpose;
      }  
*/ 
////////////////////////////////////////
