// A simple class that holds a user motion data record.

#ifndef POSE_H
#define POSE_H

#include "math3.H"
#include "list.H"

#define MAXREC 25000

class MATLABEngine;

// holds pose information, everyithing defined as public for simple access.
class Pose {
  public:
    Pose(double t, Vec3d p, Vec3d pv, Vec3d pa, Quat o, Quat av, Quat aa) : _tstamp(t),
                                                                            _pos(p),
                                                                            _posvel(pv),
                                                                            _posacc(pa),
                                                                            _orient(o),
                                                                            _angvel(av),
                                                                            _angacc(aa)
    { }

    Pose() : _tstamp(0.0), _pos(0.0,0.0,0.0), _posvel(0.0,0.0,0.0),
                          _posacc(0.0,0.0,0.0),
                          _orient(0.0,0.0,0.0,1.0),
                          _angvel(0.0,0.0,0.0,1.0),
                          _angacc(0.0,0.0,0.0,1.0)
    { }


    Pose(double t, Vec3d p, Quat o) : _tstamp(t), _pos(p), _orient(o),
                                      _posvel(0.0,0.0,0.0),
                                      _posacc(0.0,0.0,0.0),
                                      _angvel(0.0,0.0,0.0,1.0),
                                      _angacc(0.0,0.0,0.0,1.0)
    { }

    Pose(Pose *p) : _tstamp(p->_tstamp),
                    _pos(p->_pos),
                    _posvel(p->_posvel),
                    _posacc(p->_posacc),
                    _orient(p->_orient),
                    _angvel(p->_angvel),
                    _angacc(p->_angacc)
    { }
    virtual ~Pose() {}
  
  public:
    double _tstamp;  // timestamp
    Vec3d  _pos;     // position
    Vec3d  _posvel;  // linear velocity
    Vec3d  _posacc;  // linear acceleration
    Quat   _orient;  // orientation
    Quat   _angvel;  // angular velocity
    Quat   _angacc;  // angular acceleration

};


// a user motion record class holds a list of all poses for a particular motion record
class UMotionRec {
  public:
    UMotionRec() : _srate(0.0), _poselist(MAXREC) {}

    virtual ~UMotionRec() { // delete entries
      for (int i = 0; i < _poselist.num(); ++i)
        delete _poselist[i];
    }
    
    double getSamplingRate() { return _srate; } 
    FIGarray<Pose*>& getPList() { return _poselist; }

    void loadData(const char *file);        // loads data into the pose list
    void loadSimpleData(char *file);  // loads in data for with no vel and acc info
    void saveData(const char *file);        // saves data to disk   

    void sendToMatlab(MATLABEngine *eng);

  protected:  
    double _srate;  // sampling rate
    FIGarray<Pose*> _poselist;  // holds the collection of poses

};
    

// a prediction record class which holds a list of predicted poses stemming from
// a given prediciton algorthim test
class PredictionRec {
  public:
    PredictionRec(double p, double s, double n) : _predlist(MAXREC), 
                                                  _corruptlist(MAXREC),
                                                  _truelist(MAXREC),
                                                  _rwalklist(MAXREC),
                                                  _pinterval(p), _srate(s),
                                                  _noisev(n)
    { }
    virtual ~PredictionRec() { // better delete elements of the predlist and
                               // corrupt lists
      int i;
      for (i = 0; i < _predlist.num(); ++i)
        delete _predlist[i];
      for (i = 0; i < _corruptlist.num(); ++i)
        delete _corruptlist[i];
      for (i = 0; i < _rwalklist.num(); ++i)
        delete _rwalklist[i];
      
    }

    FIGarray<Pose*>& getPredList()    { return _predlist; }
    FIGarray<Pose*>& getCorruptList() { return _corruptlist; }
    FIGarray<Pose*>& getTrueList()    { return _truelist; }
    FIGarray<Pose*>& getRWalkList()   { return _rwalklist; }

    // acessors
    double getPInterval() { return _pinterval; }
    double getNoiseV()    { return _noisev; }
    double getSRate()     { return _srate; }

  protected:
    FIGarray<Pose*> _predlist;
    FIGarray<Pose*> _corruptlist; // holds the noise corrupted signal
    FIGarray<Pose*> _truelist;  // holds the true poses, used for convienence 
                                // when generating computing error 
 
    FIGarray<Pose*> _rwalklist; // holds the ramdom walk forecast -- needed to
                                // compute the percent better error metric
 
  // need other info here such as
  // the prediction interval for this record, sampling rate used, noise
    double _pinterval;
    double _noisev;
    double _srate;
};


#endif
