#include "ciPVSM.H"
#include "../pose.H"

// constructor, create all variables
CIPVSM::CIPVSM() : CIKalmanBase(), CIExtendedKalmanBase()
{
   // create for position prediction
   _RR = MMatrix(3,3);
   _HH = MMatrix(3,6);
   _QQ = MMatrix(6,6);
   _fundMat = MMatrix(6,6);
   _PP = MMatrix(6,6);
   _state = MMatrix(6,1);
  
   // Using the LogExpMap Rotation Approach
   // create for orientation prediction
   _ER = MMatrix(3,3);
   _EH = MMatrix(3,6);
   _EQ = MMatrix(6,6);
   _EfundMat = MMatrix(6,6);
   _EP = MMatrix(6,6);
   _Estate = MMatrix(6,1);
   _edt = _spd = 0.0;



}

// destructor
CIPVSM::~CIPVSM() 
{
   _RR.deleteMatrix();
   _HH.deleteMatrix();
   _QQ.deleteMatrix();
   _fundMat.deleteMatrix();
   _PP.deleteMatrix();
   _state.deleteMatrix();
   
   _ER.deleteMatrix();
   _EH.deleteMatrix();
   _EQ.deleteMatrix();
   _EfundMat.deleteMatrix();
   _EP.deleteMatrix();
   _Estate.deleteMatrix();
}

// initialize variables except for state vector
void
CIPVSM::initialize(double spd, double dt, double np, double nq)
{
 _ctr = 0;
 
 _HH(0,0) = _HH(1,2) = _HH(2,4) = 1.0;
 _RR(0,0) = _RR(1,1) = _RR(2,2) = np;
 
 _fundMat.setIdentity();
 _fundMat(0,1) = _fundMat(2,3) = _fundMat(4,5) = dt;

 _QQ(0,0) = _QQ(2,2) = _QQ(4,4) = ((dt*dt*dt) / 3.0) * spd;
 _QQ(0,1) = _QQ(1,0) = ((dt*dt) / 2.0) * spd;
 _QQ(2,3) = _QQ(3,2) = ((dt*dt) / 2.0) * spd;
 _QQ(4,5) = _QQ(5,4) = ((dt*dt) / 2.0) * spd;
 _QQ(1,1) = _QQ(3,3) = _QQ(5,5) = dt * spd;

 _PP.setZeros();
 for (int i = 0; i < 6; ++i) {
  _PP(i,i) = INITVAR;
  _state(i,0) = 0.0;
 }

 // initialize for EKF
 _edt = dt;
 _spd = spd;
 // we can set the measurment noise covariance matrix here (at least that can
 // stay constant
 _ER.setZeros();
  
 // H can stay constant too
 _EH(0,0) = _EH(1,1) = _EH(2,2) = 1.0;   
 for (int i = 0; i < 3; ++i)
    _ER(i,i) = np;    // not sure about this
 _EP.setZeros();
 for (int i = 0; i < 6; ++i) {
    _EP(i,i) = INITVAR;
    _Estate(i,0) = 0.0; 
 }
}


// do filter and predicton
Pose *
CIPVSM::filter(Pose *userpos, double pinterval)
{
 if (_ctr <= MAXBEGIN)
    ++_ctr;

 if (_ctr == 1) { // initialize state vector for KF and EKF
   for (int i = 0; i < 3; ++i)
    _state(i*2,0) = userpos->_pos[i];
   //  cerr << " " <<        _state(0,0) << " " << sqrt(_PP(0,0)) << endl;
   
   Vec3d logrot = logMap(userpos->_orient);
   for (int i = 0; i < 3; ++i)
      _Estate(i,0) = logrot[i];
   //    cerr << " " << _Estate(0,0) << " " << sqrt(_EP(0,0)) << endl; 

   return userpos;
 }

 // do the covariance intersection filter
 ciKalmanFilter(userpos, 3, kPVSM);
  
 Vec3d ppos(0.0,0.0,0.0);

 if (_ctr < MAXBEGIN) // let it run a bit before we predict
   ;
 else { // predict
  MMatrix pmat = _fundMat;
  pmat(0,1) = pmat(2,3) = pmat(4,5) = pinterval;
  MMatrix tmp = pmat * _state;
  for(int i = 0; i < 3; ++i)
     ppos[i] = tmp(i*2,0);
 }

 // do covariance intersection filter for orientation
 ciExtendedKalmanFilter(userpos, 3, kPVSM);

 // we have the new state vector for orientation
 if (_ctr < MAXBEGIN) // let it run to stabilize before we predict
   return userpos;

 // predict
 Vec3d tmpv(_Estate(0,0), _Estate(1,0), _Estate(2,0));
 Quat prot(0,0,0,1);
 Quat tq = expMap(tmpv);
 Quat tw(_Estate(3,0), _Estate(4,0), _Estate(5,0), 0);
 prot = tq + (pinterval * 0.5 * (tq * tw));
 
 return new Pose(userpos->_tstamp + pinterval, ppos, normalize(prot)); 
}
