// code for the exponential smoothing predictor

#include "expsmooth.H"
#include "types.H"

#define SPEC_DENSITY 1

// set up storage for simple PVSM KF/EKF
ExpSmooth::ExpSmooth() : KalmanBase(), ExtendedKalmanBase(), _alpha(0.01), _beta(0.01), 
                         _initpos(), _initpos2(), _ctr(0),
                         _flag(0), _plo(0), _phi(1), _initialize(1)
{

 // create for KF
  _RR = MMatrix(1,1);
  _HH = MMatrix(1,2);
  _QQ = MMatrix(2,2);
  _fundMat = MMatrix(2,2);

  for (int i = 0; i < 3; ++i) {
    _PP[i] = MMatrix(2,2); 
    _state[i] = MMatrix(2,1);
  }
  
  // create for EKF
  _ER = MMatrix(4,4);
  _EH = MMatrix(4,7);
  _EQ = MMatrix(7,7);
  _EfundMat = MMatrix(7,7);
  _EP = MMatrix(7,7);
  _Estate = MMatrix(7,1);
  _edt = _spd = 0.0;
}


// desctructor
ExpSmooth::~ExpSmooth()
{
  _RR.deleteMatrix();
  _HH.deleteMatrix();
  _QQ.deleteMatrix();
  _fundMat.deleteMatrix();
  for (int i = 0; i < 3; ++i) {
    _PP[i].deleteMatrix(); 
    _state[i].deleteMatrix();
  }

  _ER.deleteMatrix();
  _EH.deleteMatrix();
  _EQ.deleteMatrix();
  _EfundMat.deleteMatrix();
  _EP.deleteMatrix();
  _Estate.deleteMatrix();

  delete _initpos;
  delete _initpos2;
}


void
ExpSmooth::prepare()
{
  _alpha = _paramb->getAlpha();
  _beta  = _paramb->getBeta();
  _initialize = 1;
  _flag = 1;
  _plo = 0;
  _phi = 1;
  delete _initpos; 
  _initpos = new Pose();
  delete _initpos2;
  _initpos2 = new Pose();
  _ctr = 0;

  // if the Kalman flag is set we will Kalman filter
  if (_paramb->getKFExp()) {
     double dt = 1.0 / _paramb->getCurSRate();
     double np = _paramb->getPosVar();
     double nq = _paramb->getQuatVar();
     double spd = SPEC_DENSITY;
  
     // initialize for KF
     _HH(0,0) = 1.0;
     _RR(0,0) = np;

     _fundMat(0,0) = _fundMat(1,1) = 1.0;
     _fundMat(0,1) = dt; 


     _QQ(0,0) = ((dt*dt*dt) / 3.0) * spd;
     _QQ(0,1) = _QQ(1,0) = ((dt*dt) / 2.0) * spd;
     _QQ(1,1) = dt * spd;

     for (int i = 0; i < 3; ++i) {
	_PP[i].setZeros();
	_PP[i](0,0) = _PP[i](1,1) = INITVAR;
	// zero out the state vectors
	_state[i](0,0) = _state[i](1,0) = 0.0;
     }
     
     // initialize EKF
     _edt = dt;
     _spd = spd;
     // we can set the measurment noise covariance matrix here (at least that can
     // stay constant	
     _ER.setZeros();
   
     for (int i = 0; i < 4; ++i)
      _ER(i,i) = nq;
     _EP.setZeros();
     for (int i = 0; i < 7; ++i) {
	if (i < 4)
	   _EP(i,i) = 1.0;
	else
	   _EP(i,i) = INITVAR;
	_Estate(i,0) = 0.0; 
     }  

  }
}



Pose *
ExpSmooth::predict(Pose *userpos, double pinterval)
{
  Pose *npos = 0;

  // filter if needed	
  if (_paramb->getKFExp()) {
     if (_ctr <= MAXBEGIN-6) 
       ++_ctr;
     if (_ctr == 1) {
	for (int i = 0; i < 3; ++i) 
	   (_state[i])(0,0) = userpos->_pos[i];
	for (int i = 0; i < 3; ++i)
	   _Estate(i,0) = userpos->_orient.V()[i];
	_Estate(3,0) = userpos->_orient.S();
        return userpos;
     }
      // Kalman Filter
      kalmanFilter(userpos, 1, 2, kPVSM);
      extendedKalmanFilter(userpos, 4, 7, kPVSM);
      if (_ctr < MAXBEGIN-6)
        return userpos;
      else { // now start filling the pose records for prediction  
	 Vec3d nv((_state[0])(0,0), (_state[1])(0,0), (_state[2])(0,0));
	 Quat nq(_Estate(0,0), _Estate(1,0), _Estate(2,0), _Estate(3,0)); 
	 npos = new Pose(userpos->_tstamp, nv, normalize(nq));     
      }

  }

  Pose *tmppos;
  if (_paramb->getKFExp())  
    tmppos = new Pose(npos);
  else
    tmppos = new Pose(userpos);  

  // predict
  switch(_paramb->getExpSmoothType()) {
      case tSimpleExp: {    
	              if (_initialize) {
                        _initialize = 0;
	                _initpos->_pos = tmppos->_pos;
                        _initpos->_orient = tmppos->_orient; 
                        delete tmppos;
                        return userpos; 
	              }
                      else {
			 Pose *p = simpleExp(tmppos, pinterval);
                         delete tmppos;
	                 return p;
                      }
      }
      case tDoubleExp:   {
	              if (_initialize) {
	                _initialize = 0;
	                _initpos->_pos = tmppos->_pos;
	                _initpos->_orient = tmppos->_orient; 
                        _initpos2->_pos = tmppos->_pos;
	                _initpos2->_orient = tmppos->_orient; 
                        delete tmppos;
	                return userpos; 
	              }
	              else {
			Pose *p = doubleExp(tmppos, pinterval); 
                        delete tmppos;
	                return p;
                      }
      }
      case tHoltWinters:  { 
             if (_initialize) {
               if (_initialize == 1) {
		 _initpos->_pos = tmppos->_pos;
		 _initpos->_orient = tmppos->_orient; 
		 _initialize = 2;
                 delete tmppos;
                 return userpos;
               }
               else if (_initialize == 2) {
                _initpos2->_pos = tmppos->_pos - _initpos->_pos;
		_initpos2->_orient = tmppos->_orient - _initpos->_orient; 
                _initialize = 0;
                delete tmppos;
                return userpos;
	       }
	     }
	     else {
		Pose *p = holtWinters(tmppos, pinterval);
                delete tmppos;
		 return p;
             }
      }
      default: {
        cerr << "No Predictor Selected!!!" << endl;
        // for testing, just make everythin constant
        Pose *tmp = new Pose(userpos->_tstamp+pinterval, Vec3d(0.0, 0.0, 0.0), Quat(0.0,0.0,0.0,1.0));
        return tmp;
      }
    }
}


// algorithm for simple exponential smoothing

Pose *
ExpSmooth::simpleExp(Pose *userpos, double pinterval)
{
 _initpos->_pos =  (_alpha * userpos->_pos) + (1.0 - _alpha) * _initpos->_pos;
 Quat qtmp =  (_alpha * userpos->_orient) + (1.0 - _alpha) * _initpos->_orient;
 _initpos->_orient = normalize(qtmp);

 return new Pose(userpos->_tstamp + pinterval, _initpos->_pos, _initpos->_orient); 
}




// algorithm for double exponential smoothing
Pose *
ExpSmooth::doubleExp(Pose *userpos, double pinterval)
{
  double csrate = _paramb->getCurSRate();
  double delta = 1.0 / csrate;

  if (_flag) {
    while (delta*(_phi) < pinterval) {
      _plo = _phi; 
      ++_phi;
    }
    _flag = 0;
  }


  // position first
  // get the low prediction and the high prediction
   double t2 = (_phi * (1.0 / csrate));
   double t1 = (_plo * (1.0 / csrate));

  _initpos->_pos = (_alpha * userpos->_pos) + (1.0 - _alpha) * _initpos->_pos;
  _initpos2->_pos = (_alpha * _initpos->_pos) + (1.0 - _alpha) * _initpos2->_pos;

  Vec3d lopos, hipos, actpos;
  if (pinterval == (_phi * delta)) 
    actpos = ((2.0 + (_alpha * _phi) / (1.0 - _alpha)) * _initpos->_pos) -
              ((1.0 + (_alpha * _phi) / (1.0 - _alpha)) * _initpos2->_pos);
  else {
    if (_plo == 0) // just use the latest pose info
      lopos = userpos->_pos;
    else
      lopos = ((2.0 + (_alpha * _plo) / (1.0 - _alpha)) * _initpos->_pos) -
              ((1.0 + (_alpha * _plo) / (1.0 - _alpha)) * _initpos2->_pos);
  
     hipos =  ((2.0 + (_alpha * _phi) / (1.0 - _alpha)) * _initpos->_pos) -
              ((1.0 + (_alpha * _phi) / (1.0 - _alpha)) * _initpos2->_pos);  
     // do interpolation   
     actpos = (((hipos - lopos) / (t2 - t1)) * (pinterval - t1)) + lopos;
  }


  // now for orientation
  Quat qtmp1 =  (_alpha * userpos->_orient) + (1.0 - _alpha) * _initpos->_orient;
  _initpos->_orient = normalize(qtmp1);
  Quat qtmp2 = (_alpha * _initpos->_orient) + (1.0 - _alpha) * _initpos2->_orient;
  _initpos2->_orient = normalize(qtmp2);

  Quat lorot, hirot, actrot;
  if (pinterval == (_phi * delta))
    actrot = normalize(((2.0 + (_alpha * _phi) / (1.0 - _alpha)) * _initpos->_orient) -
             ((1.0 + (_alpha * _phi) / (1.0 - _alpha)) * _initpos2->_orient));
  else {
    if (_plo == 0) // use the latest pose info
      lorot = userpos->_orient;
    else
      lorot = ((2.0 + (_alpha * _plo) / (1.0 - _alpha)) * _initpos->_orient) -
              ((1.0 + (_alpha * _plo) / (1.0 - _alpha)) * _initpos2->_orient);

    hirot = ((2.0 + (_alpha * _phi) / (1.0 - _alpha)) * _initpos->_orient) -
            ((1.0 + (_alpha * _phi) / (1.0 - _alpha)) * _initpos2->_orient); 

    // do need to do quaternion interpolation here
     actrot = slerp(normalize(lorot), normalize(hirot), (pinterval-t1) / (t2 - t1)); 
  
  }

 return new Pose(userpos->_tstamp + pinterval, actpos, normalize(actrot));
}


// algorithm for Holt-Winters exponential smoothing
Pose *
ExpSmooth::holtWinters(Pose *userpos, double pinterval)
{
  double csrate = _paramb->getCurSRate();
  double delta = 1.0 / csrate;

  if (_flag) {
    while (delta*(_phi) < pinterval) {
      _plo = _phi; 
      ++_phi;
    }
    _flag = 0;
  }


  // first position
  // get the low prediction and the high prediction
   double t2 = (_phi * (1.0 / csrate));
   double t1 = (_plo * (1.0 / csrate));

   Vec3d tmpos = _initpos->_pos;
  _initpos->_pos = (_alpha * userpos->_pos) + 
                   ((1.0 - _alpha) * (_initpos->_pos + _initpos2->_pos));
  _initpos2->_pos = _beta * (_initpos->_pos - tmpos) +
                    (1.0 - _beta) * _initpos2->_pos; 

  
  Vec3d lopos, hipos, actpos;
  if (pinterval == (_phi * delta)) 
    actpos = _initpos->_pos + (_phi * _initpos2->_pos);
  else {
    if (_plo == 0) // just use the latest pose info
      lopos = userpos->_pos;
    else
      lopos = _initpos->_pos + (_plo * _initpos2->_pos);
  
    hipos =   _initpos->_pos + (_phi * _initpos2->_pos); 
    // do interpolation   
    actpos = (((hipos - lopos) / (t2 - t1)) * (pinterval - t1)) + lopos;
  }
  


  // orientation comes next
  Quat otmp = _initpos->_orient;
  _initpos->_orient =  (_alpha * userpos->_orient) + 
                   ((1.0 - _alpha) * (_initpos->_orient + _initpos2->_orient));
  _initpos2->_orient =  _beta * (_initpos->_orient - otmp) +
                    (1.0 - _beta) * _initpos2->_orient; 
  
  Quat lorot, hirot, actrot;
  if (pinterval == (_phi * delta))
    actrot = normalize(_initpos->_orient + (_phi * _initpos2->_orient));
  else {
    if (_plo == 0) // use the latest pose info
      lorot = userpos->_orient;
    else
      lorot = _initpos->_orient + (_plo * _initpos2->_orient);

    hirot = _initpos->_orient + (_phi * _initpos2->_orient);

    // do need to do quaternion interpolation here
     actrot = slerp(normalize(lorot), normalize(hirot), (pinterval-t1) / (t2 - t1)); 
  
  }

  return new Pose(userpos->_tstamp + pinterval, actpos, normalize(actrot));
}




