#include "kalmanPVASM.H"
#include "../pose.H"

// functions for the Kalman PVASM model

// constuctor, create all variables according to PVASM specs
KalmanPVASM::KalmanPVASM() : KalmanBase(), ExtendedKalmanBase(),
                             UnscentedKalmanBase(), _ctr(0)
{
 _RR = MMatrix(1,1);
 _QQ = MMatrix(3,3);
 _HH = MMatrix(1,3);
 _fundMat = MMatrix(3,3);

 for (int i = 0; i < 3; ++i) {
   _PP[i] = MMatrix(3,3);
   _state[i] = MMatrix(3,1);
 }

 // create for the EKF
 _ER = MMatrix(4,4);
 _EH = MMatrix(4,10);
 _EQ = MMatrix(10,10);
 _EfundMat = MMatrix(10,10);
 _EP = MMatrix(10,10);
 _Estate = MMatrix(10,1);
 _edt = _spd = 0.0;
  
 // create for the UKF	
 _UR = MMatrix(4,4);
 _UQ = MMatrix(10,10);
 _UP = MMatrix(10,10);
 _Ustate = MMatrix(10,1);
 _udt = _uspd = 0.0;
 _dim = 10;  // 10 dimensional state vector

}

//desctructor
KalmanPVASM::~KalmanPVASM()
{
  _RR.deleteMatrix();
  _HH.deleteMatrix();
  _QQ.deleteMatrix();
  _fundMat.deleteMatrix();
  for (int i = 0; i < 3; ++i) {
    _PP[i].deleteMatrix(); 
    _state[i].deleteMatrix();
  }

  _ER.deleteMatrix();
  _EH.deleteMatrix();
  _EQ.deleteMatrix();
  _EfundMat.deleteMatrix();
  _EP.deleteMatrix();
  _Estate.deleteMatrix();

  _UR.deleteMatrix();
  _UQ.deleteMatrix();
  _UP.deleteMatrix();
  _Ustate.deleteMatrix();
}


// initialize all variables according to PVASM specs
// except for state vectors
void
KalmanPVASM::initialize(double spd, double dt, double np, double nq, int ukfflag) 
{
 _ctr = 0;
 _ukfflag = ukfflag; 

 _HH(0,0) = 1.0;
 _RR(0,0) = np;

 _fundMat(0,0) = _fundMat(1,1) = _fundMat(2,2) = 1.0;
 _fundMat(0,1) = _fundMat(1,2) = dt;
 _fundMat(0,2) = dt*dt * 0.5; 
 
 _QQ(0,0) = (pow(dt,5) / 20.0) * spd;
 _QQ(1,1) = (pow(dt,3) / 3.0) * spd;
 _QQ(2,2) = dt * spd;
 _QQ(0,1) = _QQ(1,0) = (pow(dt,4) / 8.0) * spd;
 _QQ(0,2) = _QQ(2,0) = (pow(dt,3) / 6.0) * spd;
 _QQ(1,2) = _QQ(2,1) = dt*dt*0.5 * spd;

 // initialize P and zero out the state vector
 for (int i = 0; i < 3; ++i) {
   _PP[i].setZeros();
   _PP[i](0,0) = _PP[i](1,1) = _PP[i](2,2) = INITVAR;
   // zero out the state vectors
   _state[i](0,0) = _state[i](1,0) = _state[i](2,0) = 0.0;
 }

 if (!_ukfflag) {
   // initialize the EKF
   _edt = dt;
   _spd = spd;
   // we can set the measurment noise covariance matrix here (at least that can
   // stay constant
   _ER.setZeros();

   for (int i = 0; i < 4; ++i)
     _ER(i,i) = nq;
 
   _EP.setZeros();
   for (int i = 0; i < 10; ++i) {
     if (i < 4)
       _EP(i,i) = 1.0;
     else
       _EP(i,i) = INITVAR;
     _Estate(i,0) = 0.0; 
   }
 }
 else {
   // initialize the UKF
   _udt = dt;
   _uspd = spd;
   _UR.setZeros();
   
   for (int i = 0; i < 4; ++i)
     _UR(i,i) = nq;
   _UP.setZeros();
   for (int i = 0; i < 10; ++i) {
     if (i < 4)
       _UP(i,i) = 1.0;
     else
       _UP(i,i) = INITVAR;
     _Ustate(i,0) = 0.0; 
   }
 }

}


// do kalman filter and prediciton

Pose *
KalmanPVASM::filter(Pose *userpos, double pinterval) 
{
 ++_ctr;

 if (_ctr == 1) { // initialize the state vectors and return the orig pose
                  // for both KF and EKF 
   for (int i = 0; i < 3; ++i) {
    (_state[i])(0,0) = userpos->_pos[i];
    //  if (i == 0)
    //   cerr << " " << _state[i](0,0) << " " << sqrt(_PP[i](0,0)) << endl;
   }

   if (!_ukfflag) { // EKF init
     for (int i = 0; i < 3; ++i)
       _Estate(i,0) = userpos->_orient.V()[i];
     _Estate(3,0) = userpos->_orient.S();
     //  cerr << " " << _Estate(0,0) << " " << sqrt(_EP(0,0)) << endl;  
   }
   else { // UKF init
     for (int i = 0; i < 3; ++i)
       _Ustate(i,0) = userpos->_orient.V()[i];
     _Ustate(3,0) = userpos->_orient.S();
     //  cerr << " " << _Ustate(0,0) << " " << sqrt(_UP(0,0)) << endl;  
   }
  return userpos;
 }

 // do kalman filter for position
 kalmanFilter(userpos, 1, 3, kPVASM); 


 // we have three new state vectors 
 Vec3d ppos(0.0,0.0,0.0);

 if (_ctr < MAXBEGIN) // let it run to stabilize before we predict
   ;
 else {
  // create the prediction matrix
  MMatrix pmat = _fundMat;
  pmat(0,1) = pmat(1,2) = pinterval;
  pmat(0,2) = pinterval*pinterval*0.5;
 
  for (int i = 0; i < 3; ++i) 
     ppos[i] = (pmat *_state[i])(0,0); 
 }

  ////////////////////////////////////////////////////////
  // here we do the EKF or UKF work for orientation

  if (!_ukfflag)
    extendedKalmanFilter(userpos, 4, 10, kPVASM);
  else
    unscentedKalmanFilter(userpos, 6, kPVASM);

   // we have the new state vector for orientation
   if (_ctr < MAXBEGIN) // let it run to stabilize before we predict
    return userpos;
   
   // predict
   if (!_ukfflag) {
     Quat prot(0,0,0,1);
     Quat tq(_Estate(0,0), _Estate(1,0), _Estate(2,0), _Estate(3,0));
     Quat tw(_Estate(4,0), _Estate(5,0), _Estate(6,0), 0);
     Quat ta(_Estate(7,0), _Estate(8,0), _Estate(9,0), 0);
     Quat b = (pinterval * 0.5) * tq * tw;
     Quat c = ((-1.0 / 8.0) * pinterval * pinterval * 
	       tw.length()  *
	       tw.length()) * tq;
     Quat d = (0.25 * pinterval * pinterval) * tq * ta; 
     prot = tq + b + c + d;
     return new Pose(userpos->_tstamp + pinterval, ppos, normalize(prot)); 
   }
   else {
     Quat prot(0,0,0,1);
     Quat tq(_Ustate(0,0), _Ustate(1,0), _Ustate(2,0), _Ustate(3,0));
     Quat tw(_Ustate(4,0), _Ustate(5,0), _Ustate(6,0), 0);
     Quat ta(_Ustate(7,0), _Ustate(8,0), _Ustate(9,0), 0);
     Quat b = (pinterval * 0.5) * tq * tw;
     Quat c = ((-1.0 / 8.0) * pinterval * pinterval * 
	       tw.length()  *
	       tw.length()) * tq;
     Quat d = (0.25 * pinterval * pinterval) * tq * ta; 
     prot = tq + b + c + d;
     return new Pose(userpos->_tstamp + pinterval, ppos, normalize(prot)); 
   }
}



