#include "kalmanUpdate.H"
#include "../math3.H"


double cot(double x)
{ return 1.0 / tan(x); }

// updater functions
MMatrix Updater::RK4(const MMatrix &es, double dt, KALMANMODEL mod)
{


   switch(mod) {
    case kPSM: {
       // this is a random walk, previous estimate is current estimate
       return es;  
    }
    case kPVSM:
    case kPVMMV: {
      // break the state vector out into quaternions
      Quat q(es(0,0), es(1,0), es(2,0), es(3,0));
      Quat w(es(4,0), es(5,0), es(6,0),0);
  
      Quat u1 = q + (0.5 * dt * 0.5 * q * w);
      Quat u2 = q + (0.5 * dt * 0.5 * u1 * w);
      Quat u3 = q + (dt * 0.5 * u2 * w);
      
      Quat u4 = 0.5 * q * w;
      Quat u5 = u1 * w;
      Quat u6 = u2 * w;
      Quat u7 = 0.5 * u3 * w;
  
      Quat u8 = q + ((dt / 6.0) * (u4 + u5 + u6 + u7));

      // put the updated quate	rnion into the a priori state vector and
      // at this point we assume that angular velocity is what it was before

      MMatrix tmp(7,1);
      tmp(0,0) = u8.V()[0];
      tmp(1,0) = u8.V()[1];
      tmp(2,0) = u8.V()[2];
      tmp(3,0) = u8.S();
      tmp(4,0) = w.V()[0];
      tmp(5,0) = w.V()[1];
      tmp(6,0) = w.V()[2];
      return tmp;
    }
    case kPVASM:
    case kPVAMMV:
    case kPVAMMA:
    case kPVAMMVA: {
      // break the state vector out into quaternions
      Quat q(es(0,0), es(1,0), es(2,0), es(3,0));
  
      // first get the angular acceleration and integrate to find
      // the angular velocity	
      Vec3d anga(es(7,0), es(8,0), es(9,0));
      Vec3d angv(es(4,0), es(5,0), es(6,0));

      // integrate to get angular velocity
      Vec3d v1 = angv + (0.5 * dt * anga);
      Vec3d v2 = angv + (0.5 * dt * v1);
      Vec3d v3 = angv + (dt * v2);

      Vec3d v4 = anga;
      Vec3d v5 = 2.0 * v1;
      Vec3d v6 = 2.0 * v2;
      Vec3d v7 = v3;
      Vec3d v8 = angv + ((dt / 6.0) * (v4 + v5 + v6 + v7));
  

      // now that we have angular velocity, integrate to get the new quat
      Quat w(v8[0], v8[1], v8[2],0);
  
      Quat u1 = q + (0.5 * dt * 0.5 * q * w);
      Quat u2 = q + (0.5 * dt * 0.5 * u1 * w);
      Quat u3 = q + (dt * 0.5 * u2 * w);
      
      Quat u4 = 0.5 * q * w;
      Quat u5 = u1 * w;
      Quat u6 = u2 * w;
      Quat u7 = 0.5 * u3 * w;
  
      Quat u8 = q + ((dt / 6.0) * (u4 + u5 + u6 + u7));

      // put the updated quaternion and updated vel. into the a priori state vector and
      // at this point we assume that angular acc is what it was before

      MMatrix tmp(10,1);
      tmp(0,0) = u8.V()[0];
      tmp(1,0) = u8.V()[1];
      tmp(2,0) = u8.V()[2];
      tmp(3,0) = u8.S();
      tmp(4,0) = w.V()[0];
      tmp(5,0) = w.V()[1];
      tmp(6,0) = w.V()[2];
      tmp(7,0) = anga[0];
      tmp(8,0) = anga[1];
      tmp(9,0) = anga[2];
      return tmp;
    }
    default: ;
   }
  
   cerr << "Problem in RK4" << endl;
   return 0; // if we ever get here we have a problem
}


MMatrix Updater::updateFundMat(const MMatrix &es, double dt, KALMANMODEL mod)
{
 double qx = es(0,0);
 double qy = es(1,0);
 double qz = es(2,0);
 double qw = es(3,0);
 double w0,w1,w2;
 if (mod != kPSM) {
  w0 = es(4,0);
  w1 = es(5,0);
  w2 = es(6,0);
 }

 MMatrix tmp;

 switch(mod) {
  case kPSM: {
   MMatrix fund(4,4);
   fund.setIdentity();
   return fund; 
  }
  case kPVSM:
  case kPVMMV:
    tmp = MMatrix(7,7);
    tmp.setIdentity();
    break;
  case kPVASM:
  case kPVAMMV:
  case kPVAMMA:
  case kPVAMMVA:
    tmp = MMatrix(10,10);
    tmp.setIdentity();
    tmp(4,7) = tmp(5,8) = tmp(6,9) = dt;
    break;
  default: ;
 }

 tmp(0,1) = tmp(2,3) = dt * w2 * 0.5;
 tmp(1,0) = -dt*w2*0.5;
 tmp(0,2) = tmp(3,1) = -dt * w1 * 0.5;
 tmp(0,3) = tmp(1,2) = dt * w0 * 0.5;
 tmp(0,4) = tmp(1,5) = tmp(2,6) = dt * qw * 0.5;
 tmp(0,5) = tmp(3,6) = -dt * qz * 0.5;
 tmp(0,6) = dt * qy * 0.5;  
 
 tmp(1,3) = tmp(2,0) = dt * w1 * 0.5;
 tmp(1,4) = dt * qz * 0.5;
 tmp(1,6) = tmp(3,4) = -dt * qx * 0.5;
 
 tmp(2,1) = tmp(3,0) = -dt * w0 * 0.5;
 tmp(2,4) = tmp(3,5) = -dt * qy * 0.5;
 tmp(2,5) = dt * qx * 0.5;
 
 tmp(3,2) = -dt * w2 * 0.5;

 return tmp;
}
 

MMatrix Updater::updateQMat(const MMatrix &es, double dt, double spd, KALMANMODEL mod)
{
 double qx = es(0,0);
 double qy = es(1,0);
 double qz = es(2,0);
 double qw = es(3,0);

 switch(mod) {
   case kPSM: {
     MMatrix tmpq(4,4);
     tmpq.setIdentity();
     return dt * tmpq;
   }
   case kPVSM:
   case kPVMMV: {
     MMatrix tmp(7,7);
     tmp(0,0) = (1.0 / 12.0)* dt*dt*dt * (qw*qw + qy*qy + qz*qz);
     tmp(1,1) = (1.0 / 12.0)* dt*dt*dt * (qw*qw + qx*qx + qz*qz);
     tmp(2,2) = (1.0 / 12.0)* dt*dt*dt * (qw*qw + qx*qx + qy*qy);
     tmp(3,3) = (1.0 / 12.0)* dt*dt*dt * (qx*qx + qy*qy + qz*qz);
     tmp(4,4) = tmp(5,5) = tmp(6,6) = dt;
     
     tmp(0,1) = tmp(1,0) = (-1.0 / 12.0) * dt*dt*dt * qx*qy;
     tmp(0,2) = tmp(2,0) = (-1.0 / 12.0) * dt*dt*dt * qx*qz;
     tmp(0,3) = tmp(3,0) = (-1.0 / 12.0) * dt*dt*dt * qw*qx;
     tmp(1,2) = tmp(2,1) = (-1.0 / 12.0) * dt*dt*dt * qy*qz;
     tmp(1,3) = tmp(3,1) = (-1.0 / 12.0) * dt*dt*dt * qw*qy;
     tmp(2,3) = tmp(3,2) = (-1.0 / 12.0) * dt*dt*dt * qw*qz;

     tmp(0,4) = tmp(4,0) = (dt*dt*qw) / 4.0;
     tmp(0,5) = tmp(5,0) = -(dt*dt*qz) / 4.0;
     tmp(0,6) = tmp(6,0) = (dt*dt*qy) / 4.0;
     tmp(1,4) = tmp(4,1) = (dt*dt*qz) / 4.0;
     tmp(1,5) = tmp(5,1) = (dt*dt*qw) / 4.0;
     tmp(1,6) = tmp(6,1) = -(dt*dt*qx) / 4.0;
     tmp(2,4) = tmp(4,2) = -(dt*dt*qy) / 4.0;
     tmp(2,5) = tmp(5,2) = (dt*dt*qx) / 4.0;
     tmp(2,6) = tmp(6,2) = (dt*dt*qw) / 4.0;
     tmp(3,4) = tmp(4,3) = -(dt*dt*qx) / 4.0;
     tmp(3,5) = tmp(5,3) = -(dt*dt*qy) / 4.0;
     tmp(3,6) = tmp(6,3) = -(dt*dt*qz) / 4.0;
     return spd *tmp;
   }
   case kPVASM:
   case kPVAMMV:
   case kPVAMMA:
   case kPVAMMVA: {
     MMatrix tmp(10,10);
     tmp(4,4) = tmp(5,5) = tmp(6,6) = dt*dt*dt / 3.0;
     tmp(7,7) = tmp(8,8) = tmp(9,9) = dt;
     tmp(4,7) = tmp(5,8) = tmp(6,9) = dt*dt / 2.0; 
     tmp(7,4) = tmp(8,5) = tmp(9,6) = dt*dt / 2.0;
     return spd * tmp;
   }
   default: ;
 }

 cerr << "Problem in updateQMatrix" << endl;
 return 0;
}

MMatrix Updater::updateHMat(const MMatrix &es, KALMANMODEL mod)
{
 double qx = es(0,0);
 double qy = es(1,0);
 double qz = es(2,0);
 double qw = es(3,0);
 double l = qx*qx + qy*qy + qz*qz + qw*qw;
 MMatrix tmp;

 switch(mod) {
   case kPSM:
     tmp = MMatrix(4,4);
     break;
   case kPVSM:
     tmp = MMatrix(4,7);
     break;   
   case kPVASM: 
     tmp = MMatrix(4,10);
     break;
   case kPVMMV: 
   case kPVAMMV:
   case kPVAMMA:
     tmp = MMatrix(7,10);
     break;     
   case kPVAMMVA: 
     tmp = MMatrix(10,10);
     break;
   default: ;
 }

 if (!getenv("JULIER_MATRIX")) {
   tmp(0,0) = qw*qw + qy*qy + qz*qz;
   tmp(1,1) = qw*qw + qx*qx + qz*qz;
   tmp(2,2) = qw*qw + qx*qx + qy*qy;
   tmp(3,3) = qx*qx + qy*qy + qz*qz;
   tmp(0,1) = tmp(1,0) = -qx*qy;
   tmp(0,2) = tmp(2,0) = -qx*qz;
   tmp(0,3) = tmp(3,0) = -qw*qx;
   tmp(1,2) = tmp(2,1) = -qy*qz;
   tmp(1,3) = tmp(3,1) = -qw*qy;
   tmp(2,3) = tmp(3,2) = -qw*qz;
 }
 
 switch(mod) {
   case kPSM:
   case kPVSM:
   case kPVASM: {
      if (getenv("JULIER_MATRIX")) {
         tmp(0,0) = tmp(1,1) = tmp(2,2) = tmp(3,3) = 1.0;
	 return (1.0 / sqrt(l)) * tmp;
      }
      else
        return (1.0 / pow(l, 3.0/2.0)) * tmp;
   }
   case kPVMMV: 
   case kPVAMMV: {
      if (getenv("JULIER_MATRIX")) {
	 tmp(0,0) = tmp(1,1) = tmp(2,2) = tmp(3,3) = 1.0;
         MMatrix tmp3 = (1.0 / sqrt(l)) * tmp;
         tmp3(4,4) = tmp3(5,5) = tmp3(6,6) = 1.0; 
         return tmp3; 
      }
      else {
        MMatrix tmp2 = (1.0 / pow(l, 3.0/2.0)) * tmp; 
        tmp2(4,4) = tmp2(5,5) = tmp2(6,6) = 1.0; 
        return tmp2;
      }
   }
   case kPVAMMA: {
      if (getenv("JULIER_MATRIX")) {
         tmp(0,0) = tmp(1,1) = tmp(2,2) = tmp(3,3) = 1.0;
         MMatrix tmp3 = (1.0 / sqrt(l)) * tmp;  
          tmp3(4,7) = tmp3(5,8) = tmp3(6,9) = 1.0;
	 return tmp3; 
      }
      else {
	 MMatrix tmp2 = (1.0 / pow(l, 3.0/2.0)) * tmp; 
	 tmp2(4,7) = tmp2(5,8) = tmp2(6,9) = 1.0;
	 return tmp2; 
      }
   }
   case kPVAMMVA: {
      if (getenv("JULIER_MATRIX")) {
         tmp(0,0) = tmp(1,1) = tmp(2,2) = tmp(3,3) = 1.0;
         MMatrix tmp3 = (1.0 / sqrt(l)) * tmp;  
         tmp3(4,4) = tmp3(5,5) = tmp3(6,6) = 1.0;
	 tmp3(7,7) = tmp3(8,8) = tmp3(9,9) = 1.0;
	 return tmp3; 
      }
      else {
	 MMatrix tmp2 = (1.0 / pow(l, 3.0/2.0)) * tmp; 
	 tmp2(4,4) = tmp2(5,5) = tmp2(6,6) = 1.0;
	 tmp2(7,7) = tmp2(8,8) = tmp2(9,9) = 1.0;
	 return tmp2; 
      }
   }
  default: ;
 }
 
 cerr << "Problem in updateHMatrix" << endl;
 return 0;
}

/* // used if we are including the second order term of the taylor series exp.
   // for the fundamental matrix calculation
 double b = (1 / 8.0) * (8.0 - dt*dt*(w0*w0 + w1*w1 - w2*w2));
 double c = (1 / 8.0) * (8.0 - dt*dt*(w0*w0 + w1*w1 + w2*w2)); 

 tmp(0,0) = tmp(1,1) = b;
 tmp(2,2) = tmp(3,3) = c;
 tmp(0,1) = tmp(1,0) = w2 * dt * 0.5;
 tmp(0,2) = -dt * w1 * 0.5;
 tmp(0,3) = dt * w0 * 0.5;
 tmp(0,4) = (1.0 / 8.0) * dt * (4.0*qw + dt * (-qx*w0 + qy*w1 + qz*w2));
 tmp(0,5) = -(1.0 / 8.0) * dt * (4.0*qz + dt * (qy*w0 + qx*w1 - qw*w2));
 tmp(0,6) = -(1.0 / 8.0) * dt * (dt * (qz*w0 + qw*w1 + qx*w2) - 4.0 * qy);
 tmp(0,7) = qw * dt*dt * 0.25;
 tmp(0,8) = -qz * dt*dt * 0.25;
 tmp(0,9) = qy * dt*dt * 0.25;
 
 tmp(1,2) = -0.25 * dt * (dt*w1*w2 - 2.0*w0);
 tmp(1,3) = 0.25 * dt * (2.0*w1 + dt*w0*w2);
 tmp(1,4) = -(1.0 / 8.0) * dt * (dt * (qy*w0 + qx*w1 - qw*w2) - 4.0 * qz);
 tmp(1,5) = (1.0 / 8.0) * dt * (4.0*qw + dt * (qx*w0 - qy*w1 - qz*w2));
 tmp(1,6) = (1.0 / 8.0) * dt * (dt * (qw*w0 - qz*w1 + qy*w2) - 4.0 * qx); 
 tmp(1,7) = -tmp(0,8);
 tmp(1,8) = tmp(0,7);
 tmp(1,9) = -qx*dt*dt*0.25;

 tmp(2,0) = -0.25 * dt * (dt*w0*w2 - 2.0*w1);
 tmp(2,1) = -tmp(0,3);
 tmp(2,3) = tmp(0,1);
 tmp(2,4) =  -(1.0 / 8.0) * dt * (4.0*qy + dt * (qz*w0 - qw*w1 + qx*w2));
 tmp(2,5) = -(1.0 / 8.0) * dt * (dt * (qw*w0 + qz*w1 + qy*w2) - 4.0 * qx); 
 tmp(2,6) = (1.0 / 8.0) * dt * (4.0*qw + dt * (qx*w0 + qy*w1 - qz*w2));
 tmp(2,7) = -tmp(0,9);
 tmp(2,8) = -tmp(1,9);
 tmp(2,9) = tmp(0,7);
 
 tmp(3,0) = -0.25 * dt * (2.0*w0 + dt*w1*w2);
 tmp(3,1) = tmp(0,2);
 tmp(3,2) = -tmp(0,1);
 tmp(3,4) = -(1.0 / 8.0) * dt * (4.0*qx + dt * (qw*w0 + qz*w1 - qy*w2));
 tmp(3,5) = (1.0 / 8.0) * dt * (dt * (qz*w0 - qw*w1 - qx*w2) - 4.0 * qy); 
 tmp(3,6) = -(1.0 / 8.0) * dt * (4.0*qz + dt * (qy*w0 - qx*w1 + qw*w2));
 tmp(3,7) = tmp(1,9);
 tmp(3,8) = -tmp(0,9);
 tmp(3,9) = tmp(0,8);

 tmp(4,7) = tmp(5,8) = tmp(6,9) = dt;
 */

/* // used if we are including the 2nd order term in the taylor exp.
   // for updating the Q matrix 
 tmp(0,0) = (1.0 / 80.0)* dt*dt*dt*dt*dt * (qw*qw + qy*qy + qz*qz);
 tmp(1,1) = (1.0 / 80.0)* dt*dt*dt*dt*dt * (qw*qw + qx*qx + qz*qz);
 tmp(2,2) = (1.0 / 80.0)* dt*dt*dt*dt*dt * (qw*qw + qx*qx + qy*qy);
 tmp(3,3) = (1.0 / 80.0)* dt*dt*dt*dt*dt * (qx*qx + qy*qy + qz*qz);

 tmp(0,1) = tmp(1,0) = (-1.0 / 80.0) * dt*dt*dt*dt*dt * qx*qy;
 tmp(0,2) = tmp(2,0) = (-1.0 / 80.0) * dt*dt*dt*dt*dt * qx*qz;
 tmp(0,3) = tmp(3,0) = (-1.0 / 80.0) * dt*dt*dt*dt*dt * qw*qx;
 tmp(1,2) = tmp(2,1) = (-1.0 / 80.0) * dt*dt*dt*dt*dt * qy*qz;
 tmp(1,3) = tmp(3,1) = (-1.0 / 80.0) * dt*dt*dt*dt*dt * qw*qy;
 tmp(2,3) = tmp(3,2) = (-1.0 / 80.0) * dt*dt*dt*dt*dt * qw*qz;

 tmp(0,4) = tmp(4,0) = (dt*dt*dt*dt*qw) / 16.0;
 tmp(0,5) = tmp(5,0) = -(dt*dt*dt*dt*qz) / 16.0;
 tmp(0,6) = tmp(6,0) = (dt*dt*dt*dt*qy) / 16.0;
 tmp(0,7) = tmp(7,0) = (dt*dt*dt*qw) / 12.0;
 tmp(0,8) = tmp(8,0) = -(dt*dt*dt*qz) / 12.0;
 tmp(0,9) = tmp(9,0) = (dt*dt*dt*qy) / 12.0;
 tmp(1,4) = tmp(4,1) = (dt*dt*dt*dt*qz) / 16.0;
 tmp(1,5) = tmp(5,1) = (dt*dt*dt*dt*qw) / 16.0;
 tmp(1,6) = tmp(6,1) = -(dt*dt*dt*dt*qx) / 16.0;
 tmp(1,7) = tmp(7,1) =  (dt*dt*dt*qz) / 12.0;
 tmp(1,8) = tmp(8,1) = (dt*dt*dt*qw) / 12.0;
 tmp(1,9) = tmp(9,1) = -(dt*dt*dt*qx) / 12.0;
 tmp(2,4) = tmp(4,2) = -(dt*dt*dt*dt*qy) / 16.0;
 tmp(2,5) = tmp(5,2) = (dt*dt*dt*dt*qx) / 16.0;
 tmp(2,6) = tmp(6,2) = (dt*dt*dt*dt*qw) / 16.0;
 tmp(2,7) = tmp(7,2) = -(dt*dt*dt*qy) / 12.0;
 tmp(2,8) = tmp(8,2) = (dt*dt*dt*qx) / 12.0;
 tmp(2,9) = tmp(9,2) = (dt*dt*dt*qw) / 12.0;
 tmp(3,4) = tmp(4,3) = -(dt*dt*dt*dt*qx) / 16.0;
 tmp(3,5) = tmp(5,3) = -(dt*dt*dt*dt*qy) / 16.0;
 tmp(3,6) = tmp(6,3) = -(dt*dt*dt*dt*qz) / 16.0;
 tmp(3,7) = tmp(7,3) = -(dt*dt*dt*qx) / 12.0;
 tmp(3,8) = tmp(8,3) = -(dt*dt*dt*qy) / 12.0;
 tmp(3,9) = tmp(9,3) = -(dt*dt*dt*qz) / 12.0;
 */



// updaterLogExp functions

MMatrix 
UpdaterLogExp::RK4(const MMatrix &es, double dt, KALMANMODEL mod, double theta)
{
 double nu, gam;
 // break the state vector out into components
 Vec3d v(es(0,0), es(1,0), es(2,0));
 Vec3d w(es(3,0), es(4,0), es(5,0));

 double rndtheta = fabs(fmod(theta,2.0*M_PI));
 if (rndtheta < EPSILON) {
    nu =  dot(w,v) * (60.0 * theta*theta) / 360.0;
    //gam = 12.0 - (theta*theta);
    gam = (12.0 - (theta*theta)) / 6.0;
 }
 else {
  nu = (dot(w,v) / theta) * (cot(0.5 * theta) - 2.0 / theta);
  gam = theta * cot(0.5 * theta);
 }

  switch(mod) {
    case kPVMMV:
    case kPVSM: { 
      Vec3d u1 = v + (0.5 * dt * 0.5 * (gam * w + cross(v,w) - nu * v));
      Vec3d u2 = v + (0.5 * dt * 0.5 * (gam * w + cross(u1,w) - nu * u1));
      Vec3d u3 = v + (dt * 0.5 * (gam * w + cross(u2,w) - nu * u2));
      
      Vec3d u4 = 0.5 * (gam * w + cross(v,w) - nu * v);
      Vec3d u5 = gam * w + cross(u1, w) - nu * u1;
      Vec3d u6 = gam * w + cross(u2, w) - nu * u2;
      Vec3d u7 = 0.5 * (gam * w + cross(u3,w) - nu * u3);
  
      Vec3d u8 = v + ((dt / 6.0) * (u4 + u5 + u6 + u7));

      // put the updated vector into the a priori state vector and
      // at this point we assume that angular velocity is what it was before

      MMatrix tmp(6,1);
      tmp(0,0) = u8[0];
      tmp(1,0) = u8[1];
      tmp(2,0) = u8[2];
      tmp(3,0) = w[0];
      tmp(4,0) = w[1];
      tmp(5,0) = w[2];
      return tmp;
    }
      default: ;
  }

  cerr << " Problem in RK4 (LogExpMap) " << endl;
  return 0;
}


MMatrix 
UpdaterLogExp::updateFundMat(const MMatrix &es, double dt, KALMANMODEL mod, double theta)
{
   // break it up 
   double v0 = es(0,0);
   double v1 = es(1,0);
   double v2 = es(2,0);
   double w0 = es(3,0);
   double w1 = es(4,0);
   double w2 = es(5,0);
   double th = theta;

   switch (mod) {
    case kPVMMV:
    case kPVSM: {
       MMatrix tmp(6,6);
       tmp.setIdentity();
       double rndtheta = fabs(fmod(theta,2.0*M_PI)); 
       if (rndtheta < EPSILON) {
         tmp(0,0) = 1.0 - (1.0 / 720.0) * dt * (th*th + 60.0) * (2.0*v0*w0+v1*w1+v2*w2);
         tmp(1,1) = 1.0 - (1.0 / 720.0) * dt * (th*th + 60.0) * (v0*w0+2.0*v1*w1+v2*w2);
         tmp(2,2) = 1.0 - (1.0 / 720.0) * dt * (th*th + 60.0) * (v0*w0+v1*w1+2.0*v2*w2);
   
         tmp(1,0) = 0.5 * dt * ((-1.0 / 360.0) * (th*th+60.0)*v1*w0 - w2);
         tmp(2,0) = 0.5 * dt * (w1 - (1.0 / 360.0)*(th*th + 60.0)*v2*w0);
         tmp(0,1) = 0.5 * dt * (w2 - (1.0 / 360.0)*(th*th + 60.0)*v0*w1);
         tmp(2,1) = 0.5 * dt * (-w0 - (1.0 / 360.0)*(th*th + 60.0)*v2*w1);
         tmp(0,2) = 0.5 * dt * (-w1 - (1.0 / 360.0)*(th*th + 60.0)*v0*w2);
         tmp(1,2) = 0.5 * dt * (w0 - (1.0 / 360.0)*(th*th + 60.0)*v1*w2);

         tmp(0,3) = 0.5 * dt * (-(1.0 / 360.0)*(v0*v0 + 60.0)*th*th - (v0*v0/6.0) + 2.0);
	 tmp(1,4) = 0.5 * dt * (-(1.0 / 360.0)*(v1*v1 + 60.0)*th*th - (v1*v1/6.0) + 2.0);
	 tmp(2,5) = 0.5 * dt * (-(1.0 / 360.0)*(v2*v2 + 60.0)*th*th - (v2*v2/6.0) + 2.0);
       
         tmp(1,3) = 0.5 * dt * (v2 - (1.0 / 360.0)*(th*th + 60.0)*v0*v1); 
         tmp(2,3) = 0.5 * dt * (-v1 - (1.0 / 360.0)*(th*th + 60.0)*v0*v2);
         tmp(0,4) = 0.5 * dt * ((-1.0 / 360.0) * (th*th+60.0)*v0*v1 - v2);
         tmp(2,4) = 0.5 * dt * (v0 - (1.0 / 360.0)*(th*th + 60.0)*v1*v2);
         tmp(0,5) = 0.5 * dt * (v1 - (1.0 / 360.0)*(th*th + 60.0)*v0*v2);
         tmp(1,5) = 0.5 * dt * (-v0 - (1.0 / 360.0)*(th*th + 60.0)*v1*v2);
       }
       else {
         double a = th * cot(th / 2.0) - 2.0;
         tmp(0,0) = 1.0 - (dt * (2.0*v0*w0+v1*w1+v2*w2) * a) / (2.0*th*th);
         tmp(1,1) = 1.0 - (dt * (v0*w0+2.0*v1*w1+v2*w2) * a) / (2.0*th*th);
         tmp(2,2) = 1.0 - (dt * (v0*w0+v1*w1+2.0*v2*w2) * a) / (2.0*th*th);

         tmp(1,0) = 0.5 * dt * (-w2 - (v1*w0*a) / (th*th));
         tmp(2,0) = 0.5 * dt * (w1 - (v2*w0*a) / (th*th));
         tmp(0,1) = 0.5 * dt * (w2 - (v0*w1*a) / (th*th));
         tmp(2,1) = 0.5 * dt * (-w0 - (v2*w1*a) / (th*th));
         tmp(0,2) = 0.5 * dt * (-w1 - (v0*w2*a) / (th*th));
         tmp(1,2) = 0.5 * dt * (w0 - (v1*w2*a) / (th*th));

         tmp(0,3) = dt * (2.0*v0*v0 + th* (th*th - v0*v0) *cot(th/2.0)) / (2.0*th*th);
         tmp(1,4) = dt * (2.0*v1*v1 + th* (th*th - v1*v1) *cot(th/2.0)) / (2.0*th*th);
         tmp(2,5) = dt * (2.0*v2*v2 + th* (th*th - v2*v2) *cot(th/2.0)) / (2.0*th*th);
         
         tmp(1,3) = 0.5 * dt * (v2 - (v0*v1*a) / (th*th));
         tmp(2,3) = 0.5 * dt * (-v1 - (v0*v2*a) / (th*th));
         tmp(0,4) = 0.5 * dt * (-v2 - (v0*v1*a) / (th*th));
         tmp(2,4) = 0.5 * dt * (v0 - (v1*v2*a) / (th*th));
         tmp(0,5) = 0.5 * dt * (v1 - (v0*v2*a) / (th*th));
         tmp(1,5) = 0.5 * dt * (-v0 - (v1*v2*a) / (th*th));
       }
       return tmp;
    }
    default: ;
   }
 
   cerr << "Problem in UpdateFund (LogExpMap) " << endl;
   return 0;
}
   

MMatrix 
UpdaterLogExp::updateQMat(const MMatrix &es, double dt, double spd, 
                          KALMANMODEL mod, double theta)
{
   // break it up 	
   double v0 = es(0,0);
   double v1 = es(1,0);
   double v2 = es(2,0);
   double w0 = es(3,0);
   double w1 = es(4,0);
   double w2 = es(5,0);
   double th = theta;
   Vec3d v(es(0,0),es(1,0),es(2,0));

   switch(mod) { 
    case kPVMMV:
    case kPVSM: {
      MMatrix tmp(6,6);
      double rndtheta = fabs(fmod(theta,2.0*M_PI));
      if (rndtheta < EPSILON) {    
	 tmp(3,3) = tmp(4,4) = tmp(5,5) = dt;
         tmp(0,0) = (1.0/12.0)*dt*dt*dt* 
                    (pow((1.0/360.0)*(v0*v0+60.0)*th*th + (v0*v0)/6.0 - 2.0, 2) +
                     pow((1.0/360.0)*(th*th + 60.0)*v0*v1 + v2, 2) +
                     pow(v1 - (1.0/360.0)*(th*th + 60.0)*v0*v2,2));
         tmp(1,1) = (1.0/12.0)*dt*dt*dt* 
                    (pow((1.0/360.0)*(v1*v1+60.0)*th*th + (v1*v1)/6.0 - 2.0, 2) +
                     pow(v2-(1.0/360.0)*(th*th + 60.0)*v0*v1, 2) +
                     pow(v0+(1.0/360.0)*(th*th + 60.0)*v1*v2,2));
         tmp(2,2) = (1.0/12.0)*dt*dt*dt* 
                    (pow((1.0/360.0)*(v2*v2+60.0)*th*th + (v2*v2)/6.0 - 2.0, 2) +
                     pow(v0-(1.0/360.0)*(th*th + 60.0)*v1*v2, 2) +
                     pow(v1+(1.0/360.0)*(th*th + 60.0)*v0*v2,2));

         tmp(1,0) = tmp(0,1) = (dt*dt*dt*v0*v1*((dot(v,v) + 120.0)*pow(th,4) +
                                120*(dot(v,v) + 48.0)*th*th + 3600 *
                                (dot(v,v) - 60.0))) / 1555200.0;
         tmp(2,0) = tmp(0,2) = (dt*dt*dt*v0*v2*((dot(v,v) + 120.0)*pow(th,4) +
                                120*(dot(v,v) + 48.0)*th*th + 3600 *
                                (dot(v,v) - 60.0))) / 1555200.0;
         tmp(3,0) = tmp(0,3) = 0.25 * dt*dt * (-(1.0/360.0)*(v0*v0+60.0)*th*th - 
                               (v0*v0)/6 + 2.0);
         tmp(4,0) = tmp(0,4) = 0.25 * dt*dt * (-(1.0/360.0)*(th*th + 60.0)*v0*v1 - v2);
         tmp(5,0) = tmp(0,5) = 0.25 * dt*dt * (v1 - (1.0/360.0)*(th*th + 60.0)*v0*v2);
           
         tmp(2,1) = tmp(1,2) = (dt*dt*dt*v1*v2*((dot(v,v) + 120.0)*pow(th,4) +
                                120*(dot(v,v) + 48.0)*th*th + 3600 *
                                (dot(v,v) - 60.0))) / 1555200.0;
         tmp(3,1) = tmp(1,3) = 0.25 * dt*dt * (v2 - (1.0/360.0)*(th*th + 60.0)*v0*v1);  
         tmp(4,1) = tmp(1,4) = 0.25 * dt*dt * (-(1.0/360.0)*(v1*v1+60.0)*th*th - 
                               (v1*v1)/6 + 2.0);
         tmp(5,1) = tmp(1,5) = 0.25 * dt*dt * (-v0 - (1.0/360.0)*(th*th + 60.0)*v1*v2);
         tmp(3,2) = tmp(2,3) = 0.25 * dt*dt * (-v1 - (1.0/360.0)*(th*th + 60.0)*v0*v2);  
         tmp(4,2) = tmp(2,4) = 0.25 * dt*dt * (v0 - (1.0/360.0)*(th*th + 60.0)*v1*v2); 
         tmp(5,2) = tmp(2,5) = 0.25 * dt*dt * (-(1.0/360.0)*(v2*v2+60.0)*th*th - 
                               (v2*v2)/6 + 2.0);
      }
      else {
	 tmp(3,3) = tmp(4,4) = tmp(5,5) = dt;
         double a = th * cot(th / 2.0) - 2.0;
         tmp(0,0) = (1.0/12.0)*dt*dt*dt* 
                    (pow(2.0*v0*v0 + th*(th*th-v0*v0)*cot(th/2.0),2)/pow(th,4) +
                    pow(v2 + (v0*v1*a)/(th*th),2) +
                    pow(v1 - (v0*v2*a)/(th*th),2));
         tmp(1,1) = (1.0/12.0)*dt*dt*dt* 
                    (pow(2.0*v1*v1 + th*(th*th-v1*v1)*cot(th/2.0),2)/pow(th,4) +
                    pow(v2 - (v0*v1*a)/(th*th),2) +
                    pow(v0 + (v1*v2*a)/(th*th),2));
         tmp(2,2) = (1.0/12.0)*dt*dt*dt* 
                    (pow(2.0*v2*v2 + th*(th*th-v2*v2)*cot(th/2.0),2)/pow(th,4) +
                    pow(v1 + (v0*v2*a)/(th*th),2) +
                    pow(v0 - (v1*v2*a)/(th*th),2));

         tmp(1,0) = tmp(0,1) = (dt*dt*dt*v0*v1*(-pow(th,4)+(-2.0*th*th+dot(v,v))*
                                pow(cot(th/2.0)*th,2) + 4.0*(th*th-v0*v0-v1*v1-v2*v2)*
                                cot(th/2.0)*th + 4.0*(dot(v,v)))/ (12.0 * pow(th,4)));
         tmp(2,0) = tmp(0,2) = (dt*dt*dt*v0*v2*(-pow(th,4)+(-2.0*th*th+dot(v,v))*
                                pow(cot(th/2.0)*th,2) + 4.0*(th*th-v0*v0-v1*v1-v2*v2)*
                                cot(th/2.0)*th + 4.0*(dot(v,v)))/ (12.0 * pow(th,4)));

         tmp(3,0) = tmp(0,3) = (dt*dt*(2.0*v0*v0+th*(th*th-v0*v0)*cot(th/2.0))/
                               (4.0*th*th));

         tmp(4,0) = tmp(0,4) = 0.25*dt*dt*(-v2-(v0*v1*a)/(th*th));
         tmp(5,0) = tmp(0,5) = 0.25*dt*dt*(v1-(v0*v2*a)/(th*th));

         tmp(2,1) = tmp(1,2) = (dt*dt*dt*v1*v2*(-pow(th,4)+(-2.0*th*th+dot(v,v))*
                                pow(cot(th/2.0)*th,2) + 4.0*(th*th-v0*v0-v1*v1-v2*v2)*
                                cot(th/2.0)*th + 4.0*(dot(v,v)))/ (12.0 * pow(th,4)));
         tmp(3,1) = tmp(1,3) = 0.25*dt*dt*(v2-(v0*v1*a)/(th*th));
         tmp(4,1) = tmp(1,4) = (dt*dt*(2.0*v1*v1+th*(th*th-v1*v1)*cot(th/2.0))/
                               (4.0*th*th));
         tmp(5,1) = tmp(1,5) = 0.25*dt*dt*(-v0-(v1*v2*a)/(th*th));
        
         tmp(3,2) = tmp(2,3) = 0.25*dt*dt*(-v1-(v0*v2*a)/(th*th));
         tmp(4,2) = tmp(2,4) = 0.25*dt*dt*(v0-(v1*v2*a)/(th*th));
         tmp(5,2) = tmp(2,5) = (dt*dt*(2.0*v2*v2+th*(th*th-v2*v2)*cot(th/2.0))/
                               (4.0*th*th));
      }
     return spd * tmp;
    }
    default: ;
   }
   
   cerr << "Problem in UpdateQ (logexp)" << endl;
   return 0;
}
