// MMAE 3 model component wise predictor

#include "mmae3modelc.H"

#define EP 0.01
#define M 1 // length of measurement vector
#define LESSNUM 6

#define X 0
#define Y 1
#define Z 2

MMAE3ModelComp::MMAE3ModelComp() : _ctr(0)
{
   // create the three models
   _kfmod[0] = MMAEKalmanBaseCompWise(kPSM);
   _kfmod[1] = MMAEKalmanBaseCompWise(kPVSM);
   _kfmod[2] = MMAEKalmanBaseCompWise(kPVASM);

   for (int i = 0; i < 3; ++i)
     for (int j = 0; j < 3; ++j)
      _kfprob[i][j] = 0.0;
}


void
MMAE3ModelComp::initialize(double specd, double dt, double pn, 
                       double qn, CombinerType cmb)
{
 _ctr = 0;
 _cmb = cmb;

 // initialize each KF model
 _kfmod[0].initialize(specd, dt, pn);
 _kfmod[1].initialize(specd, dt, pn);
 _kfmod[2].initialize(specd, dt, pn);  

 // initialize the conditional probabilities for the kf models 
 // 3 models, 3 filters in each model
  for (int i = 0; i < 3; ++i)
    for (int j = 0; j < 3; ++j)
       _kfprob[i][j] = (1.0 / 3.0);
  
}


int MMAE3ModelComp::getHighestModel(int i)
{
  int highest = 0;
   
  for (int j = 1; j < 3; ++j)
   if (_kfprob[i][j] > _kfprob[i][highest])
     highest = j;
 
 return highest;
}




Pose * 
MMAE3ModelComp::filter(Pose *userpos, double pinterval)
{
 if (_ctr <= MAXBEGIN-LESSNUM)
   ++_ctr;

 if (_ctr == 1) { // initialize state vectors 
    for (int i = 0; i < 3; ++i) 
      for (int j = 0; j < 3; ++j)
	 _kfmod[i].getStateVector(j)(0,0) = userpos->_pos[j];
    return userpos;
 }  
 
 // run each filter 
 for (int i = 0; i < 3; ++i)
    _kfmod[i].mmaeKalmanFilter(userpos, 1);


 double kfptmp[3][3]; // [comp][model]
 // compute the conditional probabilities
 for (int i = 0; i < 3; ++i) {  // for x, y, and z
    for (int k = 0; k < 3; ++k) { // for models 1 - 3
       // compute the i,kth f function	
    double a = (transp(_kfmod[k].getResidual(i)) * 
                (minverse2(_kfmod[k].getResidualCov(i)) * 
                _kfmod[k].getResidual(i)))(0,0);
    double fk = 1.0 / (pow(2.0*M_PI, M/2.0) * sqrt(det(_kfmod[k].getResidualCov(i)))) *
                 exp(-0.5 * a);
    double num = fk * _kfprob[i][k];
    
    // compute the denominator sum
    double denum = 0.0;
    for (int j = 0; j < 3; ++j) {
     double b = (transp(_kfmod[j].getResidual(i)) * 
                (minverse2(_kfmod[j].getResidualCov(i)) * 
                 _kfmod[j].getResidual(i)))(0,0);
     denum += (1.0 / (pow(2.0*M_PI, M/2.0) * sqrt(det(_kfmod[j].getResidualCov(i)))) *
	       exp(-0.5 * b)) * _kfprob[i][j];
    }
    // compute new conditional probability 
    kfptmp[i][k] = num / denum;
    }
 }
 // make lower bounds on probs
 for ( int i = 0; i < 3; ++i) 
    for (int k = 0; k < 3; ++k) {
       _kfprob[i][k] = kfptmp[i][k];  
       if (_kfprob[i][k] < EP)
	  _kfprob[i][k] = EP;
    }
 cerr << "X: " << kfptmp[X][0] << " " << kfptmp[X][1] << " " << kfptmp[X][2] << endl;
 cerr << "Y: " << kfptmp[Y][0] << " " << kfptmp[Y][1] << " " << kfptmp[Y][2] << endl;
 cerr << "Z: " << kfptmp[Z][0] << " " << kfptmp[Z][1] << " " << kfptmp[Z][2] << endl;
 cerr << endl;  

 if (_ctr < MAXBEGIN-LESSNUM) // let it run for a little bit before we predict
   return userpos;

 // predict and reinitialize if needed
 Vec3d ppos(0.0,0.0,0.0);
 Quat prot(0.0,0.0,0.0,1.0);

 if (_cmb == tMAX) {
    for (int i = 0; i < 3; ++i) {
      int modnum = getHighestModel(i);
      if (modnum == 0) { // PSM model has highest probability for comp. i
         // make predicton	
         ppos[i] = _kfmod[0].getStateVector(i)(0,0);
         // restart other filters
	 if (kfptmp[i][1] < EP) {
	    _kfmod[1].getStateVector(i)(0,0) = ppos[i];
	    _kfmod[1].getCov(i)(0,0) = _kfmod[0].getCov(i)(0,0);
	 }
	 if (kfptmp[i][2] < EP) {
	    _kfmod[2].getStateVector(i)(0,0) = ppos[i];
	    _kfmod[2].getCov(i)(0,0) = _kfmod[0].getCov(i)(0,0);
	 }
      }
      else if (modnum == 1) { 
        // make prediction
        MMatrix pmat = _kfmod[1].getFundMat();
        pmat(0,1) = pinterval;
        ppos[i] = (pmat * _kfmod[1].getStateVector(i))(0,0);
        // restart other filters
        if (kfptmp[i][0] < EP) {
             _kfmod[0].getStateVector(i)(0,0) = _kfmod[1].getStateVector(i)(0,0);
             _kfmod[0].getCov(i)(0,0) = _kfmod[1].getCov(i)(0,0);
        }
        if (kfptmp[i][2] < EP) {
 	   _kfmod[2].getStateVector(i)(0,0) = _kfmod[1].getStateVector(i)(0,0);
           _kfmod[2].getStateVector(i)(1,0) = _kfmod[1].getStateVector(i)(1,0);
           _kfmod[2].getCov(i)(0,0) = _kfmod[1].getCov(i)(0,0);
           _kfmod[2].getCov(i)(1,1) = _kfmod[1].getCov(i)(1,1);
        }
      }
      else if (modnum == 2) {
	 // make prediction
	 MMatrix pmat = _kfmod[2].getFundMat(); 
	 pmat(0,1) = pmat(1,2) = pinterval;
	 pmat(0,2) = pinterval*pinterval*0.5;
	 ppos[i] = (pmat * _kfmod[2].getStateVector(i))(0,0);
	 // restart other filters
	 if (kfptmp[i][0] < EP) {
	    _kfmod[0].getStateVector(i)(0,0) = _kfmod[2].getStateVector(i)(0,0);
	    _kfmod[0].getCov(i)(0,0) = _kfmod[2].getCov(i)(0,0);
	 }
	 if (kfptmp[i][1] < EP) {
	    _kfmod[1].getStateVector(i)(0,0) = _kfmod[2].getStateVector(i)(0,0); 
	    _kfmod[1].getStateVector(i)(1,0) = _kfmod[2].getStateVector(i)(1,0); 
	    _kfmod[1].getCov(i)(0,0) = _kfmod[2].getCov(i)(0,0);
	    _kfmod[1].getCov(i)(1,1) = _kfmod[2].getCov(i)(1,1);
	 }
      }
    }
 }
 else if (_cmb == tSUM) {
    // add up position and velocities
    Vec3d pp1(0.0,0.0,0.0);
    Vec3d pp2(0.0,0.0,0.0);
    Vec3d pp3(0.0,0.0,0.0);

    Vec3d v1(0.0,0.0,0.0);
    Vec3d v2(0.0,0.0,0.0);

    Vec3d a1(0.0,0.0,0.0);

    for (int i = 0; i < 3; ++i) {
      pp1[i] = _kfmod[0].getStateVector(i)(0,0);
      pp2[i] = _kfmod[1].getStateVector(i)(0,0);
      pp3[i] = _kfmod[2].getStateVector(i)(0,0);
      v1[i]  = _kfmod[1].getStateVector(i)(1,0);
      v2[i]  = _kfmod[2].getStateVector(i)(1,0); 
      a1[i]  = _kfmod[2].getStateVector(i)(2,0);
    }

    // blend
    Vec3d pp4(0.0,0.0,0.0);
    Vec3d v3(0.0,0.0,0.0);

    for (int i = 0; i < 3; ++i) {
       pp4[i] = kfptmp[i][0] * pp1[i] + kfptmp[i][1] * pp2[i] + kfptmp[i][2] * pp3[i];
       v3[i] = (kfptmp[i][1] / (kfptmp[i][1] + kfptmp[i][2])) * v1[i] + 
               (kfptmp[i][2] / (kfptmp[i][1] + kfptmp[i][2])) * v2[i];
       a1[i] = kfptmp[i][2] * a1[i];
    }
   

    // restart if needed
    for (int i = 0; i < 3; ++i) {
      if (kfptmp[i][0] < EP)
	 _kfmod[0].getStateVector(i)(0,0) = pp4[i];
      if (kfptmp[i][1] < EP) {
	 _kfmod[1].getStateVector(i)(0,0) = pp4[i];
         _kfmod[1].getStateVector(i)(1,0) = v3[i];
      }
      if (kfptmp[i][2] < EP) {
	 _kfmod[2].getStateVector(i)(0,0) = pp4[i];
	 _kfmod[2].getStateVector(i)(1,0) = v3[i];
         _kfmod[2].getStateVector(i)(2,0) = 0.0;
      }
    }
  
    // predict with blended data
     MMatrix nstate[3];
     for (int i = 0; i < 3; ++i)
       nstate[i] = MMatrix(3,1);
    for (int i = 0; i < 3; ++i) {
      nstate[i](0,0) = pp4[i];
      nstate[i](1,0) = v3[i];
      nstate[i](2,0) = a1[i];
    }

    MMatrix pmat = _kfmod[2].getFundMat();
    pmat(0,1) = pmat(1,2) = pinterval;
    pmat(0,2) = pinterval*pinterval*0.5;
    for(int i = 0; i < 3; ++i)
       ppos[i] = (pmat * nstate[i])(0,0); 
 }

 return new Pose(userpos->_tstamp + pinterval, ppos, normalize(prot)); 
}


