// functions for computing signal analysis and quantification routines
// within the SignalAnalyzer Object


#include "ui.H"
#include "pose.H"
#include "sanalyzer.H"

#include <qmultilineedit.h>
#include <qstring.h>

#include <iostream.h>
#include <sys/time.h>

#define QFACTOR 0.1 
#define SAMPLE_RATE 10000 // for noise generation
#define DURATION 24 // 24 seconds of noise
#define SF 4

// constructor

SignalAnalyzer::SignalAnalyzer(GlobalWidget *gbw) : _gbw(gbw), _srate(0), 
                                                    _noisev(0.0), _tmprec(MAXREC),
                                                    _acf(2500), _acflists(2500),
                                                    _tpower(), _apower(),
                                                    _var(), _rms(),
                                                    _avgabsvel(), _velvar()

{
  // set up noise generators
  struct timeval ts; struct timezone tz;
  gettimeofday(&ts, &tz);
  _posnoise = new GaussWhiteNoise((long)(ts.tv_sec + ts.tv_usec/1e6)); 
  gettimeofday(&ts, &tz);
  _quatnoise = new GaussWhiteNoise((long)(ts.tv_sec + ts.tv_usec/1e6));    
}


// destructor
SignalAnalyzer::~SignalAnalyzer()
{
  delete _posnoise;
  delete _quatnoise;

  for (int i = 0; i < _tmprec.num(); ++i)
     delete _tmprec[i];
}

// methods

Pose *
SignalAnalyzer::addNoise(Pose *cpos, int srate, double noisev)
{
 // add noise to the current pose
  // first create the noise vector for position
  if (noisev == 0.0)
    return new Pose(cpos->_tstamp, cpos->_pos, cpos->_posvel, cpos->_posacc,
                    cpos->_orient, cpos->_angvel, cpos->_angacc);

  //  get six random numbers from a normal dist.
  double dt = 1.0 / (_gbw->getCPanel()->getOrigSRate() / srate);

  Vec3d noisep(_posnoise->sample(dt, 1), 
	       _posnoise->sample(dt, 2),
	       _posnoise->sample(dt, 3));

  Vec3d noiseaxis(_quatnoise->sample(dt, 1), 
		  _quatnoise->sample(dt, 2),
		  _quatnoise->sample(dt, 3));


  double w = 1.0 - length(noiseaxis);
  Quat noiseq(w, noiseaxis);
 
  return new Pose(cpos->_tstamp, cpos->_pos+noisep, cpos->_posvel, 
		  cpos->_posacc, normalize(cpos->_orient*noiseq), 
                  cpos->_angvel, cpos->_angacc);     

}


void
SignalAnalyzer::prepare()
{
   // clear the holding tank 
   for (int i = 0; i < _tmprec.num(); ++i)
     delete _tmprec[i];
   _tmprec.clear();

   // clear the acf 
    for (int i = 0; i < _acf.num(); ++i)
      delete _acf[i];
    _acf.clear();

   // clear the list of acfs
    for (int i = 0; i < _acflists.num(); ++i) {
       for (int j = 0; j < _acflists[i]->_ptlist.num(); ++j)
	  delete _acflists[i]->_ptlist[j];
       _acflists[i]->_ptlist.clear();
       delete _acflists[i];
    }
    _acflists.clear();

    // zero out the signal metrics
    _tpower.setZeros();
    _apower.setZeros();
    _var.setZeros();
    _rms.setZeros();
    _avgabsvel.setZeros();
    _velvar.setZeros();

  _srate  =  _gbw->getCPanel()->getSRateInt();
  _noisev =  _gbw->getCPanel()->getNoiseVar();

  // create noise if needed then add it to the data set
   if (_noisev != _posnoise->getVariance()) { // if these variance are different
    cerr << "Generating Noise Sequences..." << endl;
    _posnoise->generateNoise(SAMPLE_RATE, DURATION,_noisev, 0);
    _quatnoise->generateNoise(SAMPLE_RATE, DURATION,_noisev*QFACTOR, 0);
    cerr << "Noise Generation complete!" << endl;
  }

  // no matter what we should randomize the counters
  struct timeval ts; struct timezone tz;
  gettimeofday(&ts, &tz);
  _posnoise->randomizeCounters((long)(ts.tv_sec + ts.tv_usec/1e6)); 
  _quatnoise->randomizeCounters((long)(ts.tv_sec + ts.tv_usec/1e6)); 
 
  int num = _gbw->getUserRec()->getPList().num();
  UMotionRec *urec = _gbw->getUserRec();
  // adding noise, the tmprec will contain poses based on the SRATE
  for (int i = SF; i < num - SF; i = i + _srate) 
    _tmprec += addNoise(urec->getPList()[i], _srate, _noisev);    

}


void
SignalAnalyzer::computeBasicSignalMetrics()
{
   // first compute total power
   for (int i = 0; i < _tmprec.num(); ++i) {
      _tpower._pos += _tmprec[i]->_pos * _tmprec[i]->_pos;
      _tpower._orient += specialmultiply(_tmprec[i]->_orient,
                                         _tmprec[i]->_orient);
   } 
     
   // avg power
   _apower._pos = _tpower._pos / _tmprec.num();
   _apower._orient = _tpower._orient / _tmprec.num();

   // rms
   _rms._pos = mysqrt(_apower._pos);
   _rms._orient = mysqrt(_apower._orient);

   // variance
   // compute averages	
   Vec3d posavg(0.0,0.0,0.0);
   Quat  qavg(0.0,0.0,0.0,0.0);
   for (int i = 0; i < _tmprec.num(); ++i) {
      posavg += _tmprec[i]->_pos;
      qavg += _tmprec[i]->_orient;  
   }
   posavg = posavg / _tmprec.num();
   qavg = qavg / _tmprec.num();

   for (int i = 0; i < _tmprec.num(); ++i) {
      _var._pos += (_tmprec[i]->_pos - posavg) * (_tmprec[i]->_pos - posavg);
      _var._orient += specialmultiply((_tmprec[i]->_orient - qavg),
				      (_tmprec[i]->_orient - qavg));   
   }
   _var._pos = _var._pos / _tmprec.num();
   _var._orient = _var._orient / _tmprec.num();

}

void 
SignalAnalyzer::computeBasicVelMetrics()
{

   // compute average absolute velocity
   Vec3d pvavg(0.0,0.0,0.0);
   Quat  qvavg(0.0,0.0,0.0,0.0);
   for (int i = 0; i < _tmprec.num(); ++i) {
      pvavg += Vec3d(fabs(_tmprec[i]->_posvel[0]),
                     fabs(_tmprec[i]->_posvel[1]), 
                     fabs(_tmprec[i]->_posvel[2]));

      qvavg += Quat(fabs(_tmprec[i]->_angvel.V()[0]),
                    fabs(_tmprec[i]->_angvel.V()[1]),
                    fabs(_tmprec[i]->_angvel.V()[2]),
                    0.0);  
   }
   _avgabsvel._pos = pvavg / _tmprec.num();
   _avgabsvel._orient = qvavg / _tmprec.num();

   // compute velocity variance
   pvavg = Vec3d(0,0,0);
   qvavg = Quat(0,0,0,0);
   for (int i = 0; i < _tmprec.num(); ++i) {
      pvavg += _tmprec[i]->_posvel;
      qvavg += _tmprec[i]->_angvel;  
   }  

   pvavg = pvavg / _tmprec.num();
   qvavg = qvavg / _tmprec.num();

   for (int i = 0; i < _tmprec.num(); ++i) {
      _velvar._pos += (_tmprec[i]->_posvel - pvavg) * 
                      (_tmprec[i]->_posvel - pvavg);
      _velvar._orient += specialmultiply((_tmprec[i]->_angvel - qvavg),
				      (_tmprec[i]->_angvel - qvavg));   
   }
   _velvar._pos = _velvar._pos / _tmprec.num();
   _velvar._orient = _velvar._orient / _tmprec.num();
}



void
SignalAnalyzer::computeSACF(int lag, int maxlag)
{
  // tmprec is used to compute the SACF and stored in _acf
   if (lag > _tmprec.num()) {
      cerr << "Lag exceeds dataset size" << ", Lag: " << lag << 
              " Dataset: " << _tmprec.num() << endl;
      return;
   }
    

   // compute averages
   Vec3d posavg(0.0,0.0,0.0);
   Quat  qavg(0.0,0.0,0.0,0.0);
   for (int i = 0; i < _tmprec.num(); ++i) {
      posavg += _tmprec[i]->_pos;
      qavg += _tmprec[i]->_orient;  
   }
   posavg = posavg / _tmprec.num();
   qavg = qavg / _tmprec.num();

   // compute sample autocorrelation points
   Vec3d maxp(0.0,0.0,0.0);
   Quat maxq(0.0,0.0,0.0,0.0);

   if (maxlag)  // we want a full SACF
      lag = _tmprec.num()-1;

   for (int k = 0; k <= lag; ++k) {
      // compute denominator
      Vec3d denump(0.0,0.0,0.0);
      Quat denumq(0.0,0.0,0.0,0.0);
      for (int i = 0; i < _tmprec.num(); ++i) {
        denump += (_tmprec[i]->_pos - posavg) * (_tmprec[i]->_pos - posavg);
        denumq += specialmultiply((_tmprec[i]->_orient - qavg),
                                  (_tmprec[i]->_orient - qavg));   
      }
      // compute the numerator
      Vec3d nump(0.0,0.0,0.0);
      Quat  numq(0.0,0.0,0.0,0.0);
      for (int i = 0; i < _tmprec.num() - k; ++i) {
        nump += (_tmprec[i]->_pos - posavg) * (_tmprec[i+k]->_pos - posavg);
        numq += specialmultiply((_tmprec[i]->_orient - qavg),
                                (_tmprec[i+k]->_orient - qavg)); 
      }

      // compute the acf for each point and find max
      ACFPts *tmpacf = new ACFPts();
      for (int i = 0; i < 3; ++i) {
        tmpacf->_pos[i] = nump[i] / denump[i];
        if (tmpacf->_pos[i] > maxp[i])
          maxp[i] = tmpacf->_pos[i];
        tmpacf->_orient.V()[i] = numq.V()[i] / denumq.V()[i];
        if (tmpacf->_orient.V()[i] > maxq.V()[i])
          maxq.V()[i] = tmpacf->_orient.V()[i]; 
      }
      tmpacf->_orient.S() = numq.S() / denumq.S();
      if (tmpacf->_orient.S() > maxq.S())
        maxq.S() = tmpacf->_orient.S();
      
      _acf += tmpacf;
   }

   // normalize the acf points so they lie between -1 and 1
   for (int i = 0; i < _acf.num(); ++i) {
      for (int j = 0; j < 3; ++j) {
	 _acf[i]->_pos[j] /= maxp[j];
	 _acf[i]->_orient.V()[j] /= maxq.V()[j];
      }  
      _acf[i]->_orient.S() /= maxq.S();
   }

}


// computes a set of autocorrelactions based on a given window size

void
SignalAnalyzer::computeACFWindows(int initwin)
{
  
   // only interested in x right now
   for (int i = 0; i < _tmprec.num(); ++i) {
      if (i < initwin)
        ;
      else {
	 // at each i we use the initwin to
	 // do our little acf
	 Vec3d posavg(0.0,0.0,0.0);
         Quat qavg(0.0,0.0,0.0,0.0);
	 for (int j = i; j >= i - initwin; --j) {
	    posavg += _tmprec[j]->_pos;
            qavg += _tmprec[j]->_orient;
         }
         posavg = posavg / (initwin+1);
         qavg = qavg / (initwin+1);
	 
	 Vec3d maxp(0.0,0.0,0.0);
	 Quat maxq(0.0,0.0,0.0,0.0);

	 ACFPtList *nlist = new ACFPtList(initwin+1);
	 for (int k = 0; k <= initwin; ++k) {
	    Vec3d denump(0.0,0.0,0.0); 
	    Quat denumq(0.0,0.0,0.0,0.0);
            for (int p = i; p >= i - initwin; --p) {
	       denump += (_tmprec[p]->_pos - posavg) * (_tmprec[p]->_pos - posavg);
	       denumq += specialmultiply((_tmprec[p]->_orient - qavg),
                                  (_tmprec[p]->_orient - qavg));   
	    }
	    Vec3d nump(0.0,0.0,0.0);
	    Quat  numq(0.0,0.0,0.0,0.0);
            for (int p = i; p >= i - initwin + k; --p) {
	       nump += (_tmprec[p]->_pos - posavg) * (_tmprec[p-k]->_pos - posavg);
	       numq += specialmultiply((_tmprec[p]->_orient - qavg),
                                (_tmprec[p-k]->_orient - qavg)); 
	    }
            
            // compute acf and find max
            Vec3d acfp(0.0,0.0,0.0);   
            Quat acfq(0.0,0.0,0.0,0.0);        
	    for (int i = 0; i < 3; ++i) {
	       acfp[i] = nump[i] / denump[i];
	       if (acfp[i] > maxp[i])
		  maxp[i] = acfp[i];
	       acfq.V()[i] = numq.V()[i] / denumq.V()[i];
	       if (acfq.V()[i] > maxq.V()[i])
		  maxq.V()[i] = acfq.V()[i]; 
	    }
	    acfq.S() = numq.S() / denumq.S();
	    if (acfq.S() > maxq.S())
	       maxq.S() = acfq.S();          
            
            ACFPts *tmp = new ACFPts(acfp,acfq);
            nlist->_ptlist += tmp;            
	 }

	 for (int i = 0; i < nlist->_ptlist.num(); ++i) {
	    for (int j = 0; j < 3; ++j) {
	       nlist->_ptlist[i]->_pos[j] /= maxp[j];
	       nlist->_ptlist[i]->_orient.V()[j] /= maxq.V()[j];
	    }  
	    nlist->_ptlist[i]->_orient.S() /= maxq.S();
	 }
         // add it the global list
         _acflists += nlist;
      
      }
   }
}


// compute and display results from a critical point estimation 
// the routine counts the number zero axis crossings found in the first
// derivatives of the motion signal -- Also has to incorporate
// peak sizes and distances between max pean and min peaks


void
SignalAnalyzer::computeAndDisplayCrPt()
{
   double x = 0.0;
   double prevx = 0.0;

   FIGarray<Pose*> plist;     

   for (int i = 0; i < _tmprec.num(); ++i) {
      if (i == 0) {
	 if (_tmprec[i]->_posacc[0] > 0.0)
	    x = 1.0;
	 else if (_tmprec[i]->_posacc[0] < 0.0)
	    x = -1.0;
	 else 
	    x = 0;
         Pose *tmp = new Pose(_tmprec[i]->_tstamp,Vec3d(0,0,0),Quat(0,0,0,0));
         plist += tmp;
      }
      else {
	 prevx = x;
	 if (_tmprec[i]->_posacc[0] > 0.0)
	    x = 1.0;
	 else if (_tmprec[i]->_posacc[0] <= 0.0)
	    x = -1.0;
	 if (x + prevx == 0) {
            Pose *tmp = new Pose(_tmprec[i]->_tstamp,Vec3d(1,0,0),Quat(0,0,0,0));
	    plist += tmp;
         }
         else {
            Pose *tmp = new Pose(_tmprec[i]->_tstamp,Vec3d(0,0,0),Quat(0,0,0,0));
	    plist += tmp;   
	 }
     
      }
   }

   // ones mean we have a min or max peak in the velocity data
  

   _gbw->getTextBox()->append(" ");
   _gbw->getTextBox()->append("Critical Point/Peak/Distance (CPPDP) Metric Computation");
    _gbw->getTextBox()->append("Under Construction. Come back later.");
   
}




// compute and display the PSD, computations done in Matlab

void
SignalAnalyzer::computeAndDisplayPSD()
{
 // some simple stuff to the text display
   _gbw->getTextBox()->append(" ");
   _gbw->getTextBox()->append("Power Spectral Density Computation");
   _gbw->getTextBox()->append("-------------------------------------------");
   _gbw->getTextBox()->append("File: " + _gbw->getCPanel()->getCurFile()->text());
   _gbw->getTextBox()->append("Sampling Rate: " +  _gbw->getCPanel()->getCurSP()->text() +
                             "      Noise Variance: " +
                             QString::number(_gbw->getCPanel()->getNoiseVar(), 'g',8));

  // lets graph data if matlab is open
  MATLABEngine *eng = _gbw->getMEngine();

 
  if (!eng->getEngine()) {// if its not there do nothing
     _gbw->getTextBox()->append("Matlab required to compute PSD.");
    return;
  } 

  _gbw->getTextBox()->append("Graphs plotted in Matlab.");
  eng->runCmd(QString("clear"));	    
  // create large figure and position it
  eng->runCmd(QString("figure(1)"));
  eng->runCmd(QString("set(1, 'Position', [470 -35 800 980])"));
   
  int num = _tmprec.num();
  double  *data = new double[num];
 
  // X
  for (int i = 0; i < num; ++i)
    data[i] = _tmprec[i]->_pos[0];
  mxArray *DATA = mxCreateDoubleMatrix(1, num, mxREAL);
  mxSetName(DATA,"DATA");
  memcpy((void*)mxGetPr(DATA), (void*)data, num*sizeof(double));
  eng->sendArray(DATA); 
  mxDestroyArray(DATA);

  eng->runCmd(QString("subplot(4,2,1)"));
  eng->runCmd(QString("psd(DATA,[],")+
              QString::number(_gbw->getCPanel()->getCurSRate())+QString(",[])"));
  eng->runCmd(QString("title('PSD for X')"));

  // Y
  for (int i = 0; i < num; ++i)
    data[i] = _tmprec[i]->_pos[1];
  DATA = mxCreateDoubleMatrix(1, num, mxREAL);
  mxSetName(DATA,"DATA");
  memcpy((void*)mxGetPr(DATA), (void*)data, num*sizeof(double));
  eng->sendArray(DATA); 
  mxDestroyArray(DATA);

  eng->runCmd(QString("subplot(4,2,3)"));
  eng->runCmd(QString("psd(DATA,[],")+
              QString::number(_gbw->getCPanel()->getCurSRate())+QString(",[])"));
  eng->runCmd(QString("title('PSD for Y')"));

  // Z
  for (int i = 0; i < num; ++i)
    data[i] = _tmprec[i]->_pos[2];
  DATA = mxCreateDoubleMatrix(1, num, mxREAL);
  mxSetName(DATA,"DATA");
  memcpy((void*)mxGetPr(DATA), (void*)data, num*sizeof(double));
  eng->sendArray(DATA); 
  mxDestroyArray(DATA);

  eng->runCmd(QString("subplot(4,2,5)"));
  eng->runCmd(QString("psd(DATA,[],")+
              QString::number(_gbw->getCPanel()->getCurSRate())+QString(",[])"));
  eng->runCmd(QString("title('PSD for Z')"));

  // QX
  for (int i = 0; i < num; ++i)
    data[i] = _tmprec[i]->_orient.V()[0];
  DATA = mxCreateDoubleMatrix(1, num, mxREAL);
  mxSetName(DATA,"DATA");
  memcpy((void*)mxGetPr(DATA), (void*)data, num*sizeof(double));
  eng->sendArray(DATA); 
  mxDestroyArray(DATA);

  eng->runCmd(QString("subplot(4,2,2)"));
  eng->runCmd(QString("psd(DATA,[],")+
              QString::number(_gbw->getCPanel()->getCurSRate())+QString(",[])"));
  eng->runCmd(QString("title('PSD for QX')"));

  // QY
   for (int i = 0; i < num; ++i)
    data[i] = _tmprec[i]->_orient.V()[1];
  DATA = mxCreateDoubleMatrix(1, num, mxREAL);
  mxSetName(DATA,"DATA");
  memcpy((void*)mxGetPr(DATA), (void*)data, num*sizeof(double));
  eng->sendArray(DATA); 
  mxDestroyArray(DATA);

  eng->runCmd(QString("subplot(4,2,4)"));
  eng->runCmd(QString("psd(DATA,[],")+
              QString::number(_gbw->getCPanel()->getCurSRate())+QString(",[])"));
  eng->runCmd(QString("title('PSD for QY')"));

  // QZ
  for (int i = 0; i < num; ++i)
   data[i] = _tmprec[i]->_orient.V()[2];
  DATA = mxCreateDoubleMatrix(1, num, mxREAL);
  mxSetName(DATA,"DATA");
  memcpy((void*)mxGetPr(DATA), (void*)data, num*sizeof(double));
  eng->sendArray(DATA); 
  mxDestroyArray(DATA);

  eng->runCmd(QString("subplot(4,2,6)"));
  eng->runCmd(QString("psd(DATA,[],")+
              QString::number(_gbw->getCPanel()->getCurSRate())+QString(",[])"));
  eng->runCmd(QString("title('PSD for QZ')"));

  // QW
   for (int i = 0; i < num; ++i)
    data[i] = _tmprec[i]->_orient.S();
  DATA = mxCreateDoubleMatrix(1, num, mxREAL);
  mxSetName(DATA,"DATA");
  memcpy((void*)mxGetPr(DATA), (void*)data, num*sizeof(double));
  eng->sendArray(DATA); 
  mxDestroyArray(DATA);

  eng->runCmd(QString("subplot(4,2,8)"));
  eng->runCmd(QString("psd(DATA,[],")+
              QString::number(_gbw->getCPanel()->getCurSRate())+QString(",[])"));
  eng->runCmd(QString("title('PSD for QW')"));
  
  delete [] data;
}




// display functions

void
SignalAnalyzer::displaySACFResults()
{
   // some simple stuff to the text display
   _gbw->getTextBox()->append(" ");
   _gbw->getTextBox()->append("Sample Autocorrelation Function Computation");
   _gbw->getTextBox()->append("-------------------------------------------");
   _gbw->getTextBox()->append("File: " + _gbw->getCPanel()->getCurFile()->text());
   _gbw->getTextBox()->append("Sampling Rate: " +  _gbw->getCPanel()->getCurSP()->text() +
                             "      Noise Variance: " +
                             QString::number(_gbw->getCPanel()->getNoiseVar(), 'g',8));
   _gbw->getTextBox()->append("Lag: " + QString::number(_acf.num()-1));
   _gbw->getTextBox()->append("Total Number of Samples: " + QString::number(_tmprec.num()));

   // lets graph data if matlab is open
  MATLABEngine *eng = _gbw->getMEngine();

  if (!eng->getEngine()) {// if its not there do nothing
     _gbw->getTextBox()->append("Turn on Matlab to show results.");
    return;
  } 

  _gbw->getTextBox()->append("Graphs plotted in Matlab.");
  eng->runCmd(QString("clear"));	    
  // create large figure and position it
  eng->runCmd(QString("figure(1)"));
  eng->runCmd(QString("set(1, 'Position', [470 -35 800 980])"));

  // store the lag numbers
  int num = _acf.num();
  double  *lag = new double[num];
  for (int i = 0; i < num; ++i)
    lag[i] = i;
  mxArray *LAG = mxCreateDoubleMatrix(1, num, mxREAL);
  mxSetName(LAG,"LAG");
  memcpy((void*)mxGetPr(LAG), (void*)lag, num*sizeof(double));
  eng->sendArray(LAG); 
  mxDestroyArray(LAG);
  delete [] lag;

  // plot the data
  double *data = new double[num];
  // X
  for (int i = 0; i < num; ++i)
    data[i] = _acf[i]->_pos[0];
  plot("X", "subplot(4,2,1)", "plot(LAG,X)", num, data,   
       "title('Sample ACF for X')",
               "xlabel('LAG')", "ylabel('Autocorrelation')");
  // Y
  for (int i = 0; i < num; ++i)
     data[i] = _acf[i]->_pos[1];
  plot("Y", "subplot(4,2,3)", "plot(LAG,Y)", num, data,   
       "title('Sample ACF for Y')",
               "xlabel('LAG')", "ylabel('Autocorrelation')");
  // Z
  for (int i = 0; i < num; ++i)
     data[i] = _acf[i]->_pos[2];
  plot("Z", "subplot(4,2,5)", "plot(LAG,Z)", num, data,   
       "title('Sample ACF for Z')",
               "xlabel('LAG')", "ylabel('Autocorrelation')");
  // QX
  for (int i = 0; i < num; ++i)
     data[i] = _acf[i]->_orient.V()[0];
  plot("QX", "subplot(4,2,2)", "plot(LAG,QX)", num, data,   
       "title('Sample ACF for QX')",
               "xlabel('LAG')", "ylabel('Autocorrelation')");
  // QY
  for (int i = 0; i < num; ++i)
    data[i] = _acf[i]->_orient.V()[1];
  plot("QY", "subplot(4,2,4)", "plot(LAG,QY)", num, data,   
       "title('Sample ACF for QY')",
               "xlabel('LAG')", "ylabel('Autocorrelation')");
  // QZ
   for (int i = 0; i < num; ++i)
     data[i] = _acf[i]->_orient.V()[2];
  plot("QZ", "subplot(4,2,6)", "plot(LAG,QZ)", num, data,   
       "title('Sample ACF for QZ')",
               "xlabel('LAG')", "ylabel('Autocorrelation')");
  // QW
   for (int i = 0; i < num; ++i)
     data[i] = _acf[i]->_orient.S();
  plot("QW", "subplot(4,2,8)", "plot(LAG,QW)", num, data,   
       "title('Sample ACF for QW')",
               "xlabel('LAG')", "ylabel('Autocorrelation')");

  delete [] data;      
}


void
SignalAnalyzer::displayWinACFResults()
{
  // some simple stuff to the text display
   _gbw->getTextBox()->append(" ");
   _gbw->getTextBox()->append("Window-Based Sample Autocorrelation Function Computation");
   _gbw->getTextBox()->append("--------------------------------------------------------");
   _gbw->getTextBox()->append("File: " + _gbw->getCPanel()->getCurFile()->text());
   _gbw->getTextBox()->append("Sampling Rate: " +  _gbw->getCPanel()->getCurSP()->text() +
                             "      Noise Variance: " +
                             QString::number(_gbw->getCPanel()->getNoiseVar(), 'g',8));
   _gbw->getTextBox()->append("Window Size: " + QString::number(_acflists[0]->_ptlist.num()));
   _gbw->getTextBox()->append("Total Number of ACFs: " + QString::number(_acflists.num()));


 // lets graph data if matlab is open
  MATLABEngine *eng = _gbw->getMEngine();

  if (!eng->getEngine()) {// if its not there do nothing
     _gbw->getTextBox()->append("Turn on Matlab to show results.");
    return;
  } 

  _gbw->getTextBox()->append("Graphs plotted in Matlab.");
  eng->runCmd(QString("clear"));	    
  // create large figure and position it
  eng->runCmd(QString("figure(1)"));
  eng->runCmd(QString("set(1, 'Position', [470 -35 800 980])"));

   // store the lag numbers
  int num = _acflists[0]->_ptlist.num();
  double  *lag = new double[num];
  for (int i = 0; i < num; ++i)
    lag[i] = i;
  mxArray *LAG = mxCreateDoubleMatrix(1, num, mxREAL);
  mxSetName(LAG,"LAG");
  memcpy((void*)mxGetPr(LAG), (void*)lag, num*sizeof(double));
  eng->sendArray(LAG); 
  mxDestroyArray(LAG);
  delete [] lag;

  // plots the ACFs
  
  // X
  eng->runCmd(QString("subplot(3,1,1)"));
  for (int i = 0; i < _acflists.num(); ++i) {
     double *alist = new double[num];
     for (int j = 0; j < num; ++j) 
	alist[j] = _acflists[i]->_ptlist[j]->_pos[0];
     plotWin(alist, num); 
     if (i == 0) {
	eng->runCmd(QString("hold on"));
	eng->runCmd(QString("grid"));
     }
     delete [] alist;
  }          
  plotBounds("title('Window-Based ACF for X')");

  // Y
  eng->runCmd(QString("subplot(3,1,2)"));
  for (int i = 0; i < _acflists.num(); ++i) {
     double *alist = new double[num];
     for (int j = 0; j < num; ++j) 
	alist[j] = _acflists[i]->_ptlist[j]->_pos[1];
     plotWin(alist, num); 
     if (i == 0) {
	eng->runCmd(QString("hold on"));
	eng->runCmd(QString("grid"));
     }
     delete [] alist;
  }          
  plotBounds("title('Window-Based ACF for Y')");     

  // Z
  eng->runCmd(QString("subplot(3,1,3)"));
  for (int i = 0; i < _acflists.num(); ++i) {
     double *alist = new double[num];
     for (int j = 0; j < num; ++j) 
	alist[j] = _acflists[i]->_ptlist[j]->_pos[2];
     plotWin(alist, num); 
     if (i == 0) {
	eng->runCmd(QString("hold on"));
	eng->runCmd(QString("grid"));
     }
     delete [] alist;
  }          
  plotBounds("title('Window-Based ACF for Z')");


  eng->runCmd(QString("figure(2)"));
  eng->runCmd(QString("set(2, 'Position', [470 -35 800 980])"));

  // QX
  eng->runCmd(QString("subplot(4,1,1)"));
  for (int i = 0; i < _acflists.num(); ++i) {
     double *alist = new double[num];
     for (int j = 0; j < num; ++j) 
	alist[j] = _acflists[i]->_ptlist[j]->_orient.V()[0];
     plotWin(alist, num); 
     if (i == 0) {
	eng->runCmd(QString("hold on"));
	eng->runCmd(QString("grid"));
     }
     delete [] alist;
  }          
  plotBounds("title('Window-Based ACF for QX')");


  // QY
  eng->runCmd(QString("subplot(4,1,2)"));
  for (int i = 0; i < _acflists.num(); ++i) {
     double *alist = new double[num];
     for (int j = 0; j < num; ++j) 
	alist[j] = _acflists[i]->_ptlist[j]->_orient.V()[1];
     plotWin(alist, num); 
     if (i == 0) {
	eng->runCmd(QString("hold on"));
	eng->runCmd(QString("grid"));
     }
     delete [] alist;
  }          
  plotBounds("title('Window-Based ACF for QY')");

  // QZ
  eng->runCmd(QString("subplot(4,1,3)"));
  for (int i = 0; i < _acflists.num(); ++i) {
     double *alist = new double[num];
     for (int j = 0; j < num; ++j) 
	alist[j] = _acflists[i]->_ptlist[j]->_orient.V()[2];
     plotWin(alist, num); 
     if (i == 0) {
	eng->runCmd(QString("hold on"));
	eng->runCmd(QString("grid"));
     }
     delete [] alist;
  }          
  plotBounds("title('Window-Based ACF for QZ')");

  // QW
  eng->runCmd(QString("subplot(4,1,4)"));
  for (int i = 0; i < _acflists.num(); ++i) {
     double *alist = new double[num];
     for (int j = 0; j < num; ++j) 
	alist[j] = _acflists[i]->_ptlist[j]->_orient.S();
     plotWin(alist, num); 
     if (i == 0) {
	eng->runCmd(QString("hold on"));
	eng->runCmd(QString("grid"));
     }
     delete [] alist;
  }          
  plotBounds("title('Window-Based ACF for QW')");
  

}


// Using a Watts to dbW conversion, Power (in Watts) = 10*log(P) (in decibels) 
void
SignalAnalyzer::displayBasicSignalMetrics()
{
   _gbw->getTextBox()->append(" ");
   _gbw->getTextBox()->append("Basic Signal Metric Results");
   _gbw->getTextBox()->append("---------------------------");
   _gbw->getTextBox()->append("File: " + _gbw->getCPanel()->getCurFile()->text());
   _gbw->getTextBox()->append("Sampling Rate: " +  _gbw->getCPanel()->getCurSP()->text() +
                             "      Noise Variance: " +
                             QString::number(_gbw->getCPanel()->getNoiseVar(), 'g',8));

   _gbw->getTextBox()->append("Power measured in decibels using dBW scale.");
   _gbw->getTextBox()->append("       Total Power			Average Power");
   _gbw->getTextBox()->append("X:     " + QString::number(10.0*log(_tpower._pos[0]),'g',8)+"			"+QString::number(10.0*log(_apower._pos[0]),'g',8));
   _gbw->getTextBox()->append("Y:     " + QString::number(10.0*log(_tpower._pos[1]),'g',8)+ "			"+ QString::number(10.0*log(_apower._pos[1]),'g',8));
   _gbw->getTextBox()->append("Z:     " + QString::number(10.0*log(_tpower._pos[2]),'g',8)+ "			"+ QString::number(10.0*log(_apower._pos[2]),'g',8)); 
   
   _gbw->getTextBox()->append("QX:   " + QString::number(10.0*log(_tpower._orient.V()[0]),'g',8)+ "			" + QString::number(10.0*log(_apower._orient.V()[0]),'g',8));
   _gbw->getTextBox()->append("QY:   " + QString::number(10.0*log(_tpower._orient.V()[1]),'g',8)+ "			" + QString::number(10.0*log(_apower._orient.V()[1]),'g',8));
   _gbw->getTextBox()->append("QZ:   " + QString::number(10.0*log(_tpower._orient.V()[2]),'g',8)+ "			" + QString::number(10.0*log(_apower._orient.V()[2]),'g',8));
   _gbw->getTextBox()->append("QW:  " + QString::number(10.0*log(_tpower._orient.S()),'g',8) + "			" + QString::number(10.0*log(_apower._orient.S()),'g',8));

   _gbw->getTextBox()->append(" ");
   char buf[128];

   _gbw->getTextBox()->append("       Signal Variance		Signal Root Mean Square");
    sprintf(buf, "%5.8lf %39.8lf",_var._pos[0],_rms._pos[0]);
   _gbw->getTextBox()->append("X:     " + QString(buf));
    sprintf(buf, "%5.8lf %39.8lf",_var._pos[1],_rms._pos[1]);
   _gbw->getTextBox()->append("Y:     " + QString(buf));
    sprintf(buf, "%5.8lf %39.8lf",_var._pos[2],_rms._pos[2]);
   _gbw->getTextBox()->append("Z:     " + QString(buf)); 
   sprintf(buf, "%5.8lf %39.8lf",_var._orient.V()[0],_rms._orient.V()[0]);
   _gbw->getTextBox()->append("QX:   " + QString(buf));
   sprintf(buf, "%5.8lf %39.8lf",_var._orient.V()[1],_rms._orient.V()[1]);
   _gbw->getTextBox()->append("QY:   " + QString(buf));
   sprintf(buf, "%5.8lf %39.8lf",_var._orient.V()[2],_rms._orient.V()[2]);
   _gbw->getTextBox()->append("QZ:   " + QString(buf));
   sprintf(buf, "%5.8lf %39.8lf",_var._orient.S(),_rms._orient.S());
   _gbw->getTextBox()->append("QW:  " + QString(buf));

    _gbw->getTextBox()->append(" ");
}

void
SignalAnalyzer::displayBasicVelMetrics()
{
  _gbw->getTextBox()->append(" ");
  _gbw->getTextBox()->append("Basic Velocity-Based Metric Results");
  _gbw->getTextBox()->append("---------------------------");
  _gbw->getTextBox()->append("File: " + _gbw->getCPanel()->getCurFile()->text());
  _gbw->getTextBox()->append("Sampling Rate: " +  _gbw->getCPanel()->getCurSP()->text() +
                             "      Noise Variance: " +
                             QString::number(_gbw->getCPanel()->getNoiseVar(), 'g',8));

  _gbw->getTextBox()->append(" ");
  char buf[128];

   _gbw->getTextBox()->append("         AVG Absolute Speed	Velocity Variance");
    sprintf(buf, "   %5.8lf %36.8lf",_avgabsvel._pos[0],_velvar._pos[0]);
   _gbw->getTextBox()->append("X:     " + QString(buf));
    sprintf(buf, "   %5.8lf %36.8lf",_avgabsvel._pos[1],_velvar._pos[1]);
   _gbw->getTextBox()->append("Y:     " + QString(buf));
    sprintf(buf, "   %5.8lf %36.8lf",_avgabsvel._pos[2],_velvar._pos[2]);
   _gbw->getTextBox()->append("Z:     " + QString(buf)); 
   sprintf(buf, "%5.8lf %36.8lf",_avgabsvel._orient.V()[0],_velvar._orient.V()[0]);
   _gbw->getTextBox()->append("AVX:   " + QString(buf));
   sprintf(buf, "%5.8lf %36.8lf",_avgabsvel._orient.V()[1],_velvar._orient.V()[1]);
   _gbw->getTextBox()->append("AVY:   " + QString(buf));
   sprintf(buf, "%5.8lf %36.8lf",_avgabsvel._orient.V()[2],_velvar._orient.V()[2]);
   _gbw->getTextBox()->append("AVZ:   " + QString(buf));

    _gbw->getTextBox()->append(" ");
}


void 
SignalAnalyzer::plot(const char *varname, const char *subplot, 
                     const char *plotcmd, int numl, double *data,
                     const char *title, const char *xlabel, 
                     const char *ylabel)
{

  MATLABEngine *eng = _gbw->getMEngine();
  mxArray *DAT =  mxCreateDoubleMatrix(1, numl, mxREAL);
  mxSetName(DAT,varname);
  memcpy((void*)mxGetPr(DAT), (void*)data, numl*sizeof(double));
  eng->sendArray(DAT); 
  mxDestroyArray(DAT);

  eng->runCmd(QString(subplot));
  eng->runCmd(QString(plotcmd));
  eng->runCmd(QString("hold on"));
  eng->runCmd(QString("plot(LAG,zeros(1,size(LAG,2)),'g')"));
  double denum = sqrt(_tmprec.num());
  eng->runCmd(QString("bound = 1.96 /")+QString::number(denum));;
  eng->runCmd(QString("plot(LAG, bound*ones(1,size(LAG,2)),'r--')"));
  eng->runCmd(QString("plot(LAG, -bound*ones(1,size(LAG,2)),'r--')"));
  eng->runCmd(QString(title));
  eng->runCmd(QString(xlabel));
  eng->runCmd(QString(ylabel));
  eng->runCmd(QString("hold off"));

}

void
SignalAnalyzer::plotWin(double *data, int num)
{
    //    eng->runCmd(QString("axis([0 size(LAG,2) 0 ")+
   //     QString::number(5*_acflists.num())+QString(" -1 1])"));
   MATLABEngine *eng = _gbw->getMEngine(); 
   mxArray *ALIST = mxCreateDoubleMatrix(1, num, mxREAL);
   mxSetName(ALIST,"ALIST");
   memcpy((void*)mxGetPr(ALIST), (void*)data, num*sizeof(double));
   eng->sendArray(ALIST); 
   mxDestroyArray(ALIST);
   // do the plot	
   // eng->runCmd(QString("plot3(LAG,")+QString::number(5*(i+1))+
   // QString("*ones(1,size(LAG,2)),ALIST,'EraseMode','Background')"));
   eng->runCmd(QString("plot(LAG,ALIST,'EraseMode','Background')"));

}

void SignalAnalyzer::plotBounds(const char *title)
{
   MATLABEngine *eng = _gbw->getMEngine();
   eng->runCmd(QString("plot(LAG,zeros(1,size(LAG,2)),'g')"));
   eng->runCmd(QString("bound = 1.96 / sqrt(1+size(LAG,2))"));
   eng->runCmd(QString("plot(LAG, bound*ones(1,size(LAG,2)),'r--')"));
   eng->runCmd(QString("plot(LAG, -bound*ones(1,size(LAG,2)),'r--')"));
   eng->runCmd(QString(title));
   eng->runCmd(QString("xlabel('LAG')"));
   eng->runCmd(QString("ylabel('Autocorrelation')"));
   eng->runCmd(QString("hold off"));
}



// analysis routines used in collecting modeling data points


double SignalAnalyzer::computePosVar(FIGarray<Pose*> plist)
{
   Vec3d posavg(0.0,0.0,0.0);
   Vec3d posvar(0.0,0.0,0.0);
   for (int i = 0; i < plist.num(); ++i){
      posavg += plist[i]->_pos;
   }
   posavg = posavg / plist.num();

   for (int i = 0; i < plist.num(); ++i) 
      posvar += (plist[i]->_pos - posavg) * (plist[i]->_pos - posavg);

   posvar = posvar / plist.num();
   return length(posvar);

}

double SignalAnalyzer::computeRotVar(FIGarray<Pose*> &plist)
{
   Quat qavg(0.0,0.0,0.0,0.0);
   Quat qvar(0.0,0.0,0.0,0.0);
   for (int i = 0; i < plist.num(); ++i) 
      qavg += plist[i]->_orient;
   qavg = qavg / plist.num();

   for (int i = 0; i < plist.num(); ++i) 
      qvar += specialmultiply((plist[i]->_orient - qavg),
			      (plist[i]->_orient - qavg));   

   qvar = qvar / plist.num();
   return qvar.length();

}

double SignalAnalyzer::computePosSpd(FIGarray<Pose*> &plist)
{
   // compute average absolute velocity
   Vec3d pvavg(0.0,0.0,0.0);
   for (int i = 0; i < plist.num(); ++i) {
      pvavg += Vec3d(fabs(plist[i]->_posvel[0]),
                     fabs(plist[i]->_posvel[1]), 
                     fabs(plist[i]->_posvel[2]));
   }

   pvavg = pvavg / plist.num();
   return length(pvavg);
}

double SignalAnalyzer::computeRotSpd(FIGarray<Pose*> &plist)
{
  // compute average absolute velocity
   Quat  qvavg(0.0,0.0,0.0,0.0);
   for (int i = 0; i < plist.num(); ++i) {
      qvavg += Quat(fabs(plist[i]->_angvel.V()[0]),
                    fabs(plist[i]->_angvel.V()[1]),
                    fabs(plist[i]->_angvel.V()[2]),
                    0.0);  
   }

   qvavg = qvavg / plist.num();
   return qvavg.length();


}

double SignalAnalyzer::computeDPosVar(FIGarray<Pose*> &plist)
{
  // compute velocity variance
   Vec3d pvavg = Vec3d(0,0,0);
   Vec3d pvvar = Vec3d(0,0,0);

   for (int i = 0; i < plist.num(); ++i) {
      pvavg += plist[i]->_posvel;
   }  

   pvavg = pvavg / plist.num();

   for (int i = 0; i < plist.num(); ++i) {
      pvvar += (plist[i]->_posvel - pvavg) * 
               (plist[i]->_posvel - pvavg);

   }
   pvvar = pvvar / plist.num();
   return length(pvvar);

}


double SignalAnalyzer::computeDRotVar(FIGarray<Pose*> &plist)
{
  
   Quat qvavg = Quat(0,0,0,0);
   Quat qvvar = Quat(0,0,0,0);
   for (int i = 0; i < plist.num(); ++i) {
      qvavg += plist[i]->_angvel;  
   }  

   qvavg = qvavg / plist.num();

   for (int i = 0; i < plist.num(); ++i) {
      qvvar += specialmultiply((plist[i]->_angvel - qvavg),
			       (plist[i]->_angvel - qvavg));   
   }

   qvvar  = qvvar / plist.num();
   return qvvar.length();

}
